;------------------------------------------------------------------------------
;
; FPGA Nexys2 version: from badaloc_reloaded V0.99b 14/01/2009, which
;             in turn was from cpld-based badaloc bootrom V. 4.94c 27/01/2007
;
;
;
; Version 1.01 28/08/2011
;
; ULA+ management: the palette is initialized to a dimmed-low preset.
; sd-card and rs232 snapshot are now able to save and restore che ULA+
; state and palette.
;
;
; Version 1.00 08/01/2011
;
; Modified the name printed on screen.
;
; 13/02/2011: The 'E' command (ROM receive and jump) now (properly) resets
;             the 'bootrom' flag in the sh_flags register.
;             When the bootrom is interrupted by an NMI, the border turns
;             CYAN in order to easily detect it (was blue before).
;
;
;
;
; As in the badaloc_reloaded, this NEXYS2 implementation relies only on the
; RAM chip (cellularRAM with asynchronous emulation at 70ns).
;
; The size is 8Mx16; we currently only use the lower byte so we have
; 8Mx8 addressing space. As in the original project, the ram is divided into
; 1MByte blocks where the lower 512K are system RAM (residos-compatible)
; and the upper 512K are mapped when ROM access is requested (0-3FFF).
;
; This program initializes the upper half of the first megabyte with SPI FLASH
; content: sinclair roms. As before, the SPI layout is as follows:
;
; 000000 - 0EFFFF = reserved for FPGA configuration: NOT USED HERE!
; 0F0000 - 0FFFFF = 64K for bootrom (only first 16K are currently used)
; 100000 - 1FFFFF = available for Sinclair (or whatever) ROM storage
;
; Unfortunately, the NEXYS2 does not have an onboard M25P16 SPI flash, which
; has been added externally (using CS2 in the SD-CARD board). Note that CS2 was
; already the SPI Flash chip select on the original project, so no change
; is needed in order to access it.
;
;
;
; --- badaloc_reloaded V0.99b original documentation:
;
; Modified for execution on FPGA:
;
; STACK moved down to $6000 so it resides within blockram area at 4000-5FFF.
; TEMP_RAM is right next the variables area (starting at $5B00) for the same
; reason. Warning: stack space is small.
;
;
;
; Version 0.99b 14/01/2009
;
; New ResiDOS 2.08 installer tested: the signature is now properly placed
; at beginning of RAM_BANK 7, which is normally shaded-out by 8KB of video
; blockram. The detect routine has been updated accordingly.
; On previous (CPLD-BADALOC only) version, the signature was at address $3D00
; of same bank (which is now provided by DDR and not blockram, so it was
; exposed to accidental erasure. On CPLD badaloc, the entire RAM_BANK 7
; is normally provided by dualport video ram, freeing the bank on the 512K chip
; for exclusive bootrom use (it is the bank forced by context_switch on CPLD).
; The CPLD-Based clone will be updated accordingly: now the signature will
; overwrite the beginning of bootrom code.
; Smaller changes in timings for autorepeat functions
;
;
;
; Version 0.99 06/01/2009
;
; If an NMI is received while already servicing an NMI, a warm_restart is
; performed, entering the bootrom main menu' without reloading the DDR bank
; from SPI Flash
;
;
;
; Version 0.98e 05/01/2009
;
; Solved a bug in the sd-card snapshot: offset +60 in the data record was used
; for DDR_MSB register $64DF, but this byte has been allocated on ZXMMC+ for
; it's fastpage register. Now the DDR_MSB is saved at offset +61 and '00'
; is loaded at +60, which is an acceptable default for zxmmc+ to load the snap.
;
; Small delay (1 sec @3.5MHz) after sd-card snapshot restore if SW2 is ON
; allows any screen to be visible before program starts
;
;
; Version 0.98d 31/12/2008
;
; Joystick programming is now enabled by writing 'B' on D7:D4 of OUT_PORT
; register ($1F, the SPI chip select register). This will map 32 bytes of
; programmable joystick keymapping into Z80 memory.
; Should be used with FPGA version 1.13 or higher.
;
;
; Version 0.98c 29/11/2008
;
; INI unrolling code can now be used, thanks to the patched T80 core which now
; (properly) takes 16T instead of 15T for INI execution.
; mmc_read_multidata uses INI instead of INIR if USE_INI = 1 (define).
; NOTE: starting from FPGA V1.10, the 9572_CLK has been raised from 21 to 28MHz.
; Since the mmc read/write routines do not exceed 21MHz on T80 (to avoid
; screen ratio switch), INI could be used at 21MHz anyway (even with a 15T INI).
;
;
; Version 0.98b 29/10/2008
;
; Programmable joystick routine can be quit by LEFT joystick or holding CAPS
; for more than 1,5 seconds.
; 512K ram restore now reboots the machine.
; 512K ram save/restore progress shown on border.
;
;
; Version 0.98 26/10/2008
;
; Programmable joystick now supported (requires FPGA 1.08 or higher)
; Fixed a bug in the rs-232 snapshot upload which missed to reset the new
; 'bootrom is active' flag introduced in version 0.97, thanks to Victor Trucco
;
;
; Version 0.97 18/10/2008
;
; Random cleanup
;
; Fixed a bug in the 'M' rs-232 command (SPI Flash dump)
;
; Added RAM Backup snap/restore (512K)
;
; New flag indicates if NMI interrupted the bootrom itself (instead of relying
; on ROM Bank bits)
;
; Added BBC Basic ROM (32K)
;
; Added ResiDOS detection / restart. Due to volatile memory, this wil only
; work until the machine is powered off. It is however possible to perform
; a full 512K ram backup into SD-CARD.
;
; Added $54DF and $64DF registers support on RS-232 commands
;
; $64DF register is now saved and restored by rs-232 and sd-card snapshots
;
;
;
; Version 0.96 14/10/2008
;
; 28 and 42MHz options now available in the [W] edit snapshot parameters
;
;
;
; Version 0.95 12/10/2008
;
; Restored the 'Flash -> Get rom bank', which now reads the SPI Flash chip
; RTC time is preset at a fixed value; time/date can be set by Tools -> SET
; Time/Date. The new value will be used in new snapshots.
;
;
; Version 0.94 09/10/2008
;
; Restored the 'Fastpage dump of entire RAM chip'; added 'Fastpage dump of
; entire ROM chip'
;
;
;
; Version 0.93 08/10/2008
;
; Restored the 'send ROM' function. This option let the user send a 16/32 (and
; now 64) K rom under testing, which will be executed in the upper half of the
; selected megabyte (rom space)
;
;
;
; Version 0.92 04/10/2008
;
; Fixed a bug in the sd-card detection and initialization for proprietary
; snapshots TMP_RAM and STACK moved within $5FFF (now DDR starts at $6000)
;
;
;
; Version 0.91 02/10/2008
;
; The 'F' command (FLASH) now deals with the M25P16 2MB SPI Flash on the xilinx
; evakit board. This rom provides 128 banks of 16KBytes each. The second half
; (+1MB offset) is used for Sinclair ROM storage, copied to proper DDR banks at
; power on.
;
; The 64K block preceeding this area stores the present BOOTROM firmware, at
; offset $0F0000. The firmware always runs into a dedicated 16K blockram area
; (for reliability and speed reason). The first 16K of this SPI area is copied
; to blockram at power on by a bootloader included in the FPGA logic.
; The NMI handler resides here; the blockram is switched-in automatically by
; FPGA when the CPU performs an opcode fetch at $66 (NMI vector). This way,
; any program can be halted for snapshots.
;
; The first half of the SPI Flash (except last 64K) is preserved for FPGA
; configuration, which works perfectly. The bootloader let program a Xilinx .bit
; file into the SPI using the Win32 "ZX-com" program and RS-232. This feature
; may be ported to this bootrom in the future. The almost full megabyte is
; enough even for the largest FPGA of the 3E family (the 1600).
;
; SPI flash layout:
;
; 000000 - 0EFFFF = reserved for FPGA configuration
; 0F0000 - 0FFFFF = 64K for bootrom (only first 16K are currently used)
; 100000 - 1FFFFF = available for Sinclair (or whatever) ROM storage
;
; See the 'spi2ddr_copy' routine and it's associated 'copytable' to see how to
; map a SPI flash area into DDR ram at power on.
;
;
;
; Version 0.90  30/09/2008
;
; Sinclair ROM are copied from SPI FLASH into proper DDR banks for rom
; emulation. New fastpage capability on C000-FFFF is used for this purpose.
;
; Original history: in bottom of file.
;
;
;
; TABELLA REGISTRI S128KREG, CLKREG e PLUS3REG:
;
; S128KREG ($7FFD): standard nello spectrum 128K
; D0 = D0 RAM_BANK
; D1 = D1 RAM_BANK
; D2 = D2 RAM_BANK
; D3 = Normal (0) or Shadow (1) video bank is shown by ULA (CPLD)
; D4 = LSB (A14) ROM_BANK (selects ROM 0 or 1)
; D5 = Register LOCK: when set, prevents any further write until system reset.
; D6, D7 = unused
;
; CLKREG ($24DF): registro proprietario, non standard ed inesistente su spectrum
; D1 - D0 = clock selector: 00 = 3.5MHz, 01 = 7MHz, 10 = 21MHz
; D2 = clock/video multiplier: 0 = normal; 1 = clock *2, video *2 (100Hz screen)
; D3 = ROM_BANK A15
; D4 = ROM_BANK A16
; D5 = ROM_BANK A17
; D6 = Shadow rom write: non implementato
; D7 = Shadow rom read: non implementato
;
; PLUS3REG ($1FFD): registro standard nello spectrum 128K +3
; D0 = selezione modalita` standard / speciale di indirizzamento
; D1 = mod. standard = ignorato; mod. speciale seleziona tabella (vedi sotto)
; D2 = in modalita` standard e` il bit A15 ROM_BANK (seleziona fra ROM 2 o 4);
;      in modalita` speciale seleziona una tabella (vedi sotto)
;
; In modalita` di indirizzamento speciale, D2:D1 selezionano una fra 4 possibili
; modalita`: ...
;
;
;---------------------------------------------------------------------------------------------
;
; DESCRIZIONE IPOTESI DI FILESYSTEM PER SALVATAGGIO DI SNAPSHOTS:
;
; Clusters = 128KBytes: possono contenere uno snapshot da 16/48/128K. Ciascun
; file ha lunghezza fissa pari ad 1 cluster.
;
; Primi 512KBytes contengono la FAT, con 64 bytes per entry, max 8192 files,
; secondo il seguente formato:
;
; [0-1]         N. della prima ENTRY LIBERA (valido solo nel PRIMO FILE, negli altri
;               e` inutilizzato)
; [2]           Codice del contenuto: 0 = cluster libero; 1 = SNAP16K, 2 = SNAP48K,
;               8 = SNAP128K, altri da definire
; [3-34]	32 bytes Filename
; [35-41]	7 bytes Snapshot date and time from RTC chip
; [42-43]	STACK POINTER
; [44]		I Register
; [45]		$7FFD CPLD register
; [46]          F register (INT enable, IM MODE, SPECT_TYPE, DOUBLE_CLOCK BIT)
; [47]		$24DF CPLD register
; [48-57]	10 bytes joystick programming
; [58]		$1FFD CPLD register
; [59]		$54DF CPLD register
; [60]          ZXMMC+ $7F register (not used here)
; [61]          $64DF CPLD register
; [62-63]       spare
;
;
; Codici utilizzati correntemente per il byte [2]:
;
MMC_FREECLUSTER	equ	0
MMC_SNAP16	equ	1	; il cluster contiene uno snapshot da 16K
MMC_SNAP48	equ	2	; snapshot 48K
MMC_SNAP128	equ	8	; snapshot 128K
MMC_RAWDATA     equ     3       ; 128K raw data, readable and writable on
                                ; 512-bytes basis (1 byte offset, 1 byte length)
MMC_RAMBACKUP   equ     4       ; first of 4 blocks for 512KB backup (subsequent
                                ; 3 blocks will be RAWDATA type)
MMC_FLASHBACKUP equ     5       ; first of 4 blocks for 512KB FLASH backup

;
; Bitmap del BYTE [46]:
; D1-D0 = IM MODE;
; D2    = INT ENABLE (1 = enabled);
; D3	= SPECTRUM TYPE (0 = 16/48K, 1 = 128K)
; D4    = DOUBLE_INT (0 = normal; 1 = int frequency *2 (100 or 200 Hz))
;         (default = 0, only set by user menu`)
; D5	= CONTEXT_SWITCH DISABLE BIT
; D6	= '1' = All non-standard register are locked (to be implemented)
; D7    = '1' = ULA+ colors in use
;
;
; Le dimensioni della FAT possono essere cambiate mediante modifica di una EQU,
; anche se poi alcune routines sono ottimizzate per questa dimensione e quindi
; la cosa non e` cosi` immediata. Ogni entry punta ad un cluster 128K FISSO,
; calcolato in base alle dimensioni della FAT stessa (in EQU) piu` 128K per
; ciascuna entry che precede.
;
; Per accedere ai dati di una entry occorre leggere tutto il blocco da 512 bytes, che ne contiene 8.
; Questo e` vero soprattutto se tali dati devono essere modificati (riscrittura del blocco).
;
; L'indirizzo di lettura/scrittura nella SD/MMC e` costituito da un puntatore a 32bit, che indirizza 4GB.
; Di questi, i primi 9 bit indirizzano i singoli bytes di un BLOCCO, che e` sempre 512 bytes, e sono quindi
; sempre a zero quando la card viene interrogata.
;
; I restanti 7 bit relativi alla WORD LSB indirizzano uno fra 128 possibili blocchi, per un totale di 64KBytes
; (infatti si tratta di un puntatore a 16 bit, che indirizza quindi 64KBytes). Per indirizzare un cluster e`
; quindi necessario utilizzare anche D0 della word MSB.
;
; Posizione del cluster HARD-CODED con il numero dell'entry:
;
; L'indirizzo del cluster relativo ad una certa entry corrisponde al numero della entry stessa riportato a
; partire da D1 della WORD MSB, in modo che il suo valore unitario corrisponda a 128KBytes di indirizzamento
; (D0 della stessa word + altri 16 bit della word LSB). A questo si aggiunge l'offset relativo allo spazio
; utilizzato dalla FAT.
; Anche se i nostri cluster sono da 128K, i 16 bit di indirizzo del cluster (parametro 3-4) vengono copiati
; pari-pari nella WORD MSB del puntatore fornito alla SD, permettendo potenzialmente di indirizzare con una
; risoluzione di 64KBytes. Questo e` fatto per semplicita` nelle operazioni; il bit D0 di tale puntatore
; dovra` essere sempre BASSO. Con una SD da 1GB, i due bit MSB di questo indirizzo non vengono utilizzati
; in quanto la capacita` complessiva di indirizzamento di 32bit e` pari a 4GB.
;
;
; NOTA IMPORTANTE relativa a prove effettuate con SD/MMC da 1GB prodotte da Kingston e eXagerate (!)
; La capacita` disponibile non e` davvero pari ad 1GB (inteso come 1024*1024*1024 = 1073741824).
;
; kINGSTON:  MAX snapshot = 7855; su 7856 non riesce. (7856*128)+512K = 1006080KB (-41,5MB)
; eXagerate: MAX snapshot = 7647 (7648*128)+512 = 979456KB = (-67,5MB)
;
; Nel caso venisse aggiunta una tabella "bitmap" in cui ciascun bit contiene lo stato occupato/libero
; di un cluster da 128KB (1 byte = 1MB), occorrerebbe 1KB per memorizzare lo stato di una SD da 1GB.
; Questo KB puo` essere sottratto alla FAT attualmente implementata (64 bytes per entry, per 1GB occupa
; 512KB) perche` comporta la perdita di 16 entries, che non possono ugualmente essere utilizzate
; (nel caso migliore, ovvero il kingston, manca la memoria equivalente a 336 snapshots).
; Con una simile tabella e la definizione di 2 bytes nell'entry che indichi il cluster associato, sarebbe
; possibile implementare una versione con entry/cluster non "hard-coded". Il vantaggio potrebbe essere
; quello di implementare le directory: una directory e` una entry con un flag specifico, il cui cluster
; da 128K contiene un nuovo blocco di entry come quelle base (128K = 2048 sotto-entry, sembra abbastanza).
; Ovviamente la tabella bitmap relativa allo stato libero/occupato dei clusters resterebbe una sola.
;


;FLASH	equ	0		; Nessuna FLASH a bordo
;FLASH	equ	1		; FLASHROM WINBOND 2Mbit W29C020
FLASH	equ	2		; FLASHROM ST 1Mbit M29F010B


WRSNAP_21MHz	equ	0	; '1' = Utilizza 21MHz durante scrittura snapshot, altrimenti 14MHz

RDSNAP_SPEED    equ     3       ; Frequenza T80 ripristino snapshot (3 = 21MHz)
;RDSNAP_SPEED   equ     6       ; 28MHz

RESISPEED_DEFAULT equ	3	; value for "residos_speed" initialization (maximum allowable speed, written
				; by residos to the CLKREG register for MMC transfer or something).
				; also used for maximum speed selection in this program


USE_INI		equ	1	; Routines di lettura SD/MMC: 0 = INIR (21T-states/byte), 1 = INI*16 (16T-states/byte)


TMP_DDRMSB	equ	1	;MSB address (1 = +1MB offset) of DDR used as scratchpad (rom flash, etc)
TMP_FASTPAGE	equ	$C0	;FASTPAGE used with the above TMP_DDRMSB. C0 = bank 0, rd/wr enabled.

PJOY_SIZE       equ     10      ;programmable joystick inputs


; UART/JOYSTICK REGISTERS
KEMPSTON	equ $1F
TXREG   	equ $7F		; registro trasmissione
RXREG   	equ $7F		; registro ricezione
STATUS  	equ $5F		; STATUS register
NMIREG		equ $5F		; registro abilitazione pin DATA_AVAILABLE --> NMI Z80

; VECCHI (kempston e SPI sono rimasti al loro posto)
;TXREG   	equ $0F		; registro trasmissione
;RXREG   	equ $0F		; registro ricezione
;STATUS  	equ $2F		; STATUS register
;NMIREG		equ $2F		; registro abilitazione pin DATA_AVAILABLE --> NMI Z80


PLUS3REG equ	$1FFD		; registro standard nello spectrum 128K +3
CLKREG	equ	$24DF		; clock select / rom-bank register / shadow rom control register
SPECTR_MODE equ $34DF		; D0 = 0 = 48K, 1 = 128K; D1 = '1' = DOUBLE_INT; D2,D3 = set/reset context;
				; D4 = flash rom write enable; D5 = context_switch disable; D6 = non-standard regs LOCK
UNLOCKREG equ	$44DF		; scrivendo $AA su questo registro, si sblocca D4 $7FFD (lock bit)
FASTPAGE equ	$54DF		; RAM/ROM fast paging register
DDR_MSB equ	$64DF
S128KREG equ	$7FFD		; registro standard gestione banchi dello spectrum 128K

ULAP_CTRL equ   $BF3B           ; Ula+ control register
ULAP_DATA equ   $FF3B           ; Ula+ data register

BORDER	equ $FE			; PORT setup border sui bit D2 - D0 (gli altri riguardano SPKR/TAPE)
BORDCR	equ 23624		; variabile SPECTRUM contenente il colore del border *8

CHAR_X	equ 6			; larghezza dei caratteri nel font (che e` cmq 8x8)


; SD/MMC RELATED DEFINES:

MMC_GO_IDLE_STATE	equ	$40
MMC_SEND_OP_COND	equ	$41
MMC_READ_CID		equ	$4A
MMC_SET_BLOCK_SIZE	equ	$50
MMC_READ_SINGLE_BLOCK	equ	$51
MMC_READ_MULTIPLE_BLOCK	equ	$52
MCC_TERMINATE_MULTI_READ equ	$4C
MMC_WRITE_SINGLE_BLOCK	equ	$58
MMC_WRITE_MULTIPLE_BLOCK equ	$59
MMC_STOP_TRAN		equ	$FD


OUT_PORT	equ	$1F	; port for CS control (D1:D0) and software RS-232 (D2)
SPI_PORT	equ 	$3F


BLOCKSIZE	equ	512	; SD/MMC block size (bytes)

FATSIZE		equ	8	; dimensioni FAT in termini di blocchi da 64KBytes (in pratica e` l'offset
				; da caricare nella word MSB dell'indirizzo della SD per accedere al primo
				; cluster di dati, subito oltre la FAT). 8 = 512KBytes (8192 entries da 64 bytes)

MMCSIZE		equ	8192	; dimensioni della MMC in termini di blocchi da 128K: 8192 = 1GB
				; Prima di prendere/ripristinare uno snapshot, il programmma verifica che				; l'ubicazione dei dati non sia oltre il limite fisico, tenendo anche conto dello				; spazio occupato dalla FAT. Con una FATSIZE = 8 (4 blocchi da 128K) e una MMC da 1GB				; (8192 blocchi) sara` possibile gestire 8192-4 = 8188 snapshots.

OFFSET_GRAN	equ	16	; sd_check ora cerca l'header snapshot a piu' offset, spaziati di OFFSET_GRAN.
				; con 16 avremo salti di 256MB (per cui controlla a 0, +256, +512, etc

MAXOFFSET	equ	112	; limite ricerca (compreso): con 48 cerca fino a +768MB, con 96 fino a +1,5GB





; Area RAM variabili utilizzate dal menu` di BOOT (possono risiedere in ram
; spectrum in quanto non contengono informazioni che debbano sopravvivere al
; lancio di una ROM, che azzera tutta la memoria).
; Notare che in quest'area non devono esserci variabili usate da routines
; dell'NMI handler, per non rischiare di corrompere il funzionamento.

	org $5B00

x_pos		equ	$	; coordinata X attuale posizione di scrittura (IN PIXEL, 0 - 255)
	org $+1
y_pos		equ	$	; coordinata Y attuale posizione di scrittura (in caratteri, 0 - 23)
	org $+1
wr_video	equ	$	; puntatore ram video (2 bytes)
	org $+2
attribute	equ	$	; attributi usati dalle routines di scrittura a video
	org $+1
frames		equ    	$	; loop counter per ritardi vari (4 bytes)
	org $+4
flags		equ	$	; flags vari
	org $+1
flags2		equ	$	; flags vari. questo byte DEVE RISIEDERE IN RAM SPECTRUM, non in SHADOW_VARIABLES,
				; in quanto il bit ROM_SIZE (utilizzato durante la ricezione di una ROM per la
				; sua valutazione in SHADOW_ROM) deve essere accessibile anche se, una volta
				; disattivato il context, passiamo sul banco SHADOW 0 (che non contiene le
				; variabili del firmware BOOTROM).
	org $+1
sd_entry	equ	$	; numero entry snapshot visualizzata per prima nel menu` (2 bytes)
	org $+2
tmp_str		equ	$	; spazio per una stringa da 64 bytes (scrtittura riga su video?)
	org $+64
wr_attribute	equ	$	; puntatore all'area attributi dello schermo (routines di scrittura) (2 bytes)
	org $+2
firstfree	equ	$	; numero della prima entry libera per snapshot (2 bytes)
	org $+2
rewrite_timer	equ 	$	; contatore ritardo per aggiornamento FIRSTFREE in SD/MMC
	org $+1
start_item	equ	$	; usato per la selezione di una voce di menu`
	org $+1
repeat_timer	equ 	$	; contatore autorepeat generico
	org $+1
saved_attrib	equ	$
	org $+1
joystick_item	equ	$	; usato per la selezione di uno degli ingressi (programmazione)
	org $+1
joystick_pnt	equ	$	; address of joystick key assignment buffer
	org $+2


;NOTA IMPORTANTE:
;L'area 6000-7FFF e' occupata da DDR (per sdoppiare la blockram in due banchi video da 8K)
;quindi TEMP_RAM e lo stack vengono ora tenuti entro la blockram per questioni di sicurezza

TEMP_RAM	equ	$	; scratchpad ram

STACK		equ	$6000	; attenzione all'utilizzo di TEMP_RAM, la quantita' totale di ram
				; a partire da $5B00 e` 1280 bytes!!


;
; SHADOW_ROM RD/WR SPACE
;
; It's purpose is to provide a RD/WR scratchpad area, very useful when no spectrum memory can be
; altered (for example during SD/MMC snapshot operations).
;
; When a NMI is detected and the CPLD switches the bootrom in, the last KB ($3C00 - $3FFF) will access
; the SHADOW_ROM RAM chip in RD/WR mode. This area can then be used for firmware variables.
; However, the first page ($3Cxx) is not used. ULA3 switches the entire KB in, because this requires
; less logic than working on page basis.
;


SHADOW_VARS	equ	$3D00	; SHADOW_ROM space used as scratchpad rd/wr RAM.
				; This is also the amout of bytes copied from EPROM to SHADOW_ROM
				; when making a copy of BOOTROM firmware prior to switch
				; to SHADOW_ROM execution for any reason, in order to preserve
				; variables content

SHADOW_RAM	equ	$3E00	; 512 bytes scratchpad for snapshot SD/MMC operations
SHADOW_STACK	equ	$3E00	; stack in shadow_rom. SP will start writing at $3DFF


; The following variables are allocated in SHADOW_ROM space and are used by SD/MMC snapshot
; subroutines in order to leave all the spectrum memory untouched. THIS 256 BYTES PAGE IS SHARED
; WITH SHADOW_STACK.
;
; ANY VARIABLE MODIFIED BY BASIC-ACCESSIBLE ROUTINES (USR 37) SHOULD BE PLACED HERE!

	org SHADOW_VARS

start_residos   equ     $
residos_sign    equ     $       ; two bytes ResiDOS signature (should be 'R', 'D')
	org $+2
residos_version	equ	$	; two bytes ResiDOS version (BCD)
	org $+2
residos_bank	equ	$	; ResiDOS fastpage RAM BANK to jump into (JP 0)
	org $+1
residos_speed	equ	$	; maximum allowable clock speed ($3D05)
	org $+1
end_residos     equ     $

sd_firstentry	equ	$	; first free entry in SD/MMC
	org $+2
card_offset	equ	$	; bits 31-24 (4o MSB byte) of 32 bit absolute base card address for snapshots.
	org $+1			; 0 = begin of card; 16 = +256MB offset; 32 = +512MB offset; 48 = +768MB offset
real_sp		equ	$	; the real SP save place
	org $+2
entry_addr	equ	$	; offset of entry in 512 bytes read block
	org $+2
snap_torestore equ	$	; number of snapshot to be restored (16 bit)
	org $+2
im_mode		equ	$	; used to detect the IM MODE currently selected in the processor
	org $+1
prev_clock	equ	$	; previous clock speed backup
	org $+1
nmi_count	equ	$	; used to detect nested NMI calls
	org $+1
rtc_data	equ	$	; RTC date and time local variables (read from RTC by rtc_enquire)
	org $+7
sh_flags	equ	$	; various flags in SHADOW_RAM
	org $+1
i_content	equ	$	; I register backup for INT status detection
	org $+1
snap_1ffd	equ	$	; $1FFD content, saved when taking an sd/mmc snapshot
	org $+1
cs_outreg	equ	$	; value output on OUTREG PORT by CS_LOW routine (MMC card 0/1 select)
	org $+1
snap_7ffd	equ	$	; $7FFD content, saved when taking an sd/mmc snapshot or rs-232 snapshot
	org $+1
snap_34df	equ	$
	org $+1
fastp_save	equ	$
	org $+1
spectrm_save	equ	$
	org $+1
ddrmsb_save     equ     $
        org $+1
flash_bank	equ	$	; FLASH ROM BANK to be programmed
	org $+1
flash_towr	equ	$	; nbytes to be flashed
	org $+2
tmp_reg	equ	$
	org $+10

;
; Definizione di 'flags'
;
mask_int	equ 0		; maskable interrupt
eccede		equ 1		; usato dalla routine di scrittura a video
residos_present	equ 2		; 1 = ResiDOS was detected in nonvolatile RAM
sp_pressed	equ 3		; usato per autorepeat cambio n. entry in cui effettuare il prossimo snapshot
key_pressed	equ 4		; gestione autorepeat generico
restore_attrib	equ 5		; usato per la selezione di un item nel menu principale
save_attrib	equ 6		; usato per la selezione di un item nel menu principale
update_mmc	equ 7		; usato nelle routines di setup parametri snapshot

;
; Definizione di 'flags2'
;
need_reloc	equ 0		; usato da LOADTHIS per stabilire se occorre rilocare il codice in RAM
force_points	equ 1		; build_timestring will display the ':' symbol regardless of the 1Hz line
rom_size	equ 2		; dimensioni ROM (16 o 32KB) che sta per essere ricevuta dall'host
suppress_leadz	equ 3		; conv_4cifre will suppress leading zeroes
weare_shadow 	equ 4		; 1 = stiamo girando in shadow_rom

;
; Definizione di 'sh_flags' (byte di flag salvato in SHADOW_RAM)
;
prev_1Hz 	equ	0	; previous status of the 1Hz signal from RTC chip
reload_rtc 	equ	1	; the NMI handler sets this flag so that if bootrom is running, RTC will be read again
sd_present 	equ	2	; set = SD/MMC is present
sd_initsnap 	equ	3	; set = SD/MMC is initialized for proprietary snapshots filesystem
redraw_screen	equ	4	; NMI handler sets this flag to tell the main bootrom loop to redraw all menus
bootrom         equ     5       ; '1' = bootrom is active (for NMI purposes)
ulap_in_use     equ     6       ; '1' = ULA+ colors detected (during snapshots)


        org 0                   ; Begin of ROM space (in BLOCKRAM)

	di
	jp reset_vector

	org $8
	ret

	org $10
	ret

	org $14			; $14: 4 locazioni libere su tutte le rom (128/48K)
	ret			; istruzioni di uscita dallo stato di CONTEXT ATTIVO
	retn

	org $18
	ret

	org $20
	ret


	org $25			; $25: 3 istruzioni libere su tutte le rom
	jp bootrom_services	; l'esecuzione su $25 provoca il CONTEXT MODE per accedere alla BOOTROM


	org $28
	ret

	org $30
	ret


        org $38         	; MASKABLE INTERRUPT HANDLER: non deve e non puo`
        ei              	; fare niente, potremmo appena aver abilitato gli INT
        reti            	; dopo un ripristino SNAPSHOT


	org $66
	jp nmi_handler		; NMI handler


	; The IM_2 vector table (points to $202 regardless of the value fetched from databus)

	org $100
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202, $202
	dw	$202

	push af			; This instruction is located at address $202
	ld a,2			; If called, this routine loads "2" in IM_MODE and turns
	ld (im_mode),a		; magenta the border. This will only happen if IM_2
	inc a			; was selected in the processor by the host program
	out (BORDER),a
	pop af
	reti			; exits without re-enabling interrupts

;	org $ff			; (OLD) IM_MODE TEST interrupt vector $00FF
;	dw	vectint_m2	; address of service routine


;
;---------------------------------------------------------------------------------------------
; Get IM MODE Subroutine.
;
; Call with D = previous I state, that will be restored on exit.
;
; Loads IM_MODE with 1 or 2 depending on the actual IM_MODE of the processor. IM0 is detected
; as IM1. REQUIRES THE SHADOW_ROM TO BE ENABLED BEFORE CALLING.
;
; When IM_MODE 2 is detected, the border is turned MAGENTA.
;
; Destroys AF
;---------------------------------------------------------------------------------------------
get_immode

	ld a,1
	ld i,a			; restart vector for IM2 = $01xx
	ld (im_mode),a		; default to IM1
	ei
	halt			; wait at least 1 INT
	di			; NOTE: INT will need to be re-enabled if IFF2 was SET
	ld a,d
	ld i,a			; restores I content
	ret

;
;---------------------------------------------------------------------------------------------
;--- NMI HANDLER. La CPLD ULA3 attiva questa ROM (allocata sul quarto banco, 0xC000)
;--- quando rileva un ciclo M1 all'indirizzo $66. In questo modo la rom originale non deve
;--- essere modificata: si salta direttamente in questa. Il ROM_BANK viene ovviamente
;--- preservato e rimesso in linea quando viene eseguito un ciclo M1 all'indirizzo $14 o $16.
;--- L'ultimo KB non accede alla ROM, ma bensi` alla SHADOW_ROM in modo RD/WR.
;
;--- Le locazioni $14 e $15 (che sono libere sia nella rom 48K che nella 128K) dovranno
;--- contenere gli opcode delle istruzioni RET (1 byte) e RETN (2 bytes) rispettivamente.
;
;--- Al termine delle nostre operazioni, invece di effettuare un RETN eseguiamo un JP $15.
;--- Per uscire dal menu` di boot, invece, si usera` JP $14 (RET).
;
; Quando chiamata, controlla se un byte e` ricevuto. Se l'UART non ha bytes pronti, assumiamo
; che sia stato premuto il pulsante il quale permette di scegliere la frequenza operativa
; mediante la pressione dei tasti da 1 a 5 oppure eseguire uno snapshot su SD/MMC con Q,W o E.
;
; Se un byte e` presente ed e` il primo della sequenza di entrata nel loop di attesa comandi,
; disabilita ulteriore generazione di NMI ed attende il resto.
;
; Codici di risposta da parte dell'handler NMI:
;
; '*' = OK in risposta (varie situazioni)
;
; '!' = handler non attivo; il byte ricevuto ha generato NMI ma non e` $55; l'handler e`
;	uscito lasciando attiva la generazione di ulteriori NMI, in attesa di un $55.
;
; '?' = handler attivo; il byte ricevuto non e` un comando riconosciuto. si resta in loop
;	in attesa di un comando valido. NMI disattivati (gestione in polling)
;
; '#' = handler attivo; in seguito ad errore di ricezione, siamo fuori sincrono in una
;	sequenza di ricezione. Trasmettere ulteriori bytes (ad esempio 0x00) fino ad ottenere
;	una risposta '?', oppure trasmettere ulteriori ESC (27) fino ad ottenere '!'.
;	NMI disattivati (gestione in polling)
;
; Nel momento in cui viene rilevato un errore (OVERRUN o FRAMING) viene trasmesso il contenuto
; dello STATUS con D0 = reset (D0 = data_available) e D7 set (ad indicare che e` un codice di errore)
; seguito dal codice '#' sopradescritto.
;
;
; Codici dei comandi riconosciuti:
;
; ESC (27) = uscita dall'handler con riattivazione generazione di NMI per successivi bytes
; ESC+1 (28) = come sopra, ma con preventiva esecuzione di EI. Serve perche` un eventuale snapshot
; avra` sovrascritto lo stato del flag IFF2 (durante la routine di rilevamento dello stato IM_MODE
; del processore) e quindi gli INT non saranno riabilitati all'uscita dell'handler NMI. Il programma
; VC++ ne tiene conto ed invia ESC+1 quando e se l'utente preme Log-OUT.
;
; 'R' = Ricezione di un blocco di memoria
;
; 'T' = Trasmissione di un blocco di memoria
;
; 'S' = Trasmissione dell'attuale contenuto dei registri 7FFD, 24DF, 34DF, 1FFD,
;       54DF e 64DF
;
; 'G' = Ricezione nuovo valore da caricare in 7FFD
;
; 'C' = Ricezione nuovo valore da caricare in 24DF (shadow_rd/wr, rom msb, cpu clock)
;
; 'D' = Ricezione nuovo valore da caricare in 34DF
;
; 'J' = Salto ad indirizzo (nn)
;
; NOTA: In realta` il programma lanciato potrebbe poi rientrare nel punto di interruzione NMI
; eseguendo la seguente sequenza di uscita: pop IY, BC', DE, BC, HL, AF, [ret].
; Vedi programma di esempio "REENTER.ASM".
;
;
; 'Q' = Trasmissione del contenuto dei registri non salvati sullo STACK (ad uso SNAPSHOT)
;
; 'W' = Ricezione del contenuto dei registri non salvati sullo stack (ad uso SNAPSHOT)
;
; 'P' = Ricezione stato INTERRUPT MASCHERABILI e $24DF CPLD (ad uso snapshot)
;
; 'Z' = Trasmissione di un blocco (512 bytes) di memoria SD/MMC
;
; 'X' = Ricezione di un blocco (512 bytes) di memoria SD/MMC
;
; 'A' = Trasmissione del contenuto dell'RTC (per l'occasione viene re-interrogato il PIC)
;
; 'B' = Ricezione di data/ora e trasmissione al pic, che aggiornera` l'RTC
;
; 'E' = Ricezione di una ROM (16/32KB), che viene poi copiata in shadow_rom e lanciata
;
; 'F' = Ricezione di un blocco rom da 16K e scrittura IN FLASH EPROM
;
; 'H' = Ricezione nuovo valore da caricare in 1FFD
;
; 'K' = Ricezione nuovo valore da caricare in 54DF, 64DF
;
; 'I' = Dump di un banco da 16K selezionato mediante FASTPAGE register
;
; 'M' = Dump di un banco 16K dalla SPI Flash (copiato in DDR e poi trasmesso)
;
;---------------------------------------------------------------------------------------------

nmi_handler

	push af

	ld a,(nmi_count)	; detects nested NMI calls
	cp 0
	jr z,ok_nmi
	ld a,2			; error: NMI routine was already active
	out ($FE),a
;        pop af
;        ret                     ; it does not perform RETN nor disables context_mode: we still are in NMI handler

        ld (nmi_count),a         ;NEW: signals that nested NMI occurred
        jp warm_reset            ;then restarts the program (quick restart)

ok_nmi
	ld a,1
	ld (nmi_count),a
	out ($FE),a

	ld a,(sh_flags)		; in case this NMI interrupted the bootrom firmware, we signal that the RTC
	set reload_rtc,a	; chip should be re-read because local variables increment will be halted
	ld (sh_flags),a		; for a while


;	ld a,$87	; TEMPORANEO SANDY: disabilita l'I/O interno all'interfaccia (D7 SET)
;	out ($FC),a	; poniamo i 3 bit LSB = 7 in quanto a questo OUT risponde anche l'ULA su $FE!
			; serve solo in rarissimi casi (snap del programma FORMAT SANDY), normalmente
			; l'I/O sandy e` gia` disabilitato



;	xor a		; TEMPORANEO PER ENTRARE IN CASO DI LOOP SPI:
;	out (NMIREG),a	; disabilita ulteriori NMI in caso di ricezione da seriale
;	push bc
;	ld bc,0
;	ld a,3
;dldl
;	dec c		; esegue un loop lungo. PREMERE il bottone di NMI E POI fare LOGIN
;	jr nz,dldl	; su zx-com.
;	dec b
;	jr nz,dldl
;	dec a
;	jr nz,dldl
;	pop bc


	in a,(STATUS)	; legge lo status register
        bit 0,a         ; c'e' un byte? (l'NMI potrebbe essere generato da un
        jp nz,qualcosa  ; pulsante)


; GESTIONE COMANDI SU SINGOLO TASTO DOPO NMI: SCELTA CLOCK, SNAPSHOT SU SD/MMC


	push hl		; la strana sequenza di salvataggio e` fatta per ricalcare quella
	push bc		; usata nelle routine di comunicazione seriale, per avere un certo
	push de		; grado di equivalenza negli snapshot via RS-232 e su SD/MMC
w_keyclock
	call key_scan
	jr nz,w_keyclock; piu` di 2 tasti: ignora
	ld a,e
	cp $FF
	jr z,w_keyclock	; no keys pressed

	call key_number	; A = numero del tasto premuto (1 - 40)

	cp 9
	jr nc,no_newclock
	ld l,a

	ld d,a		; DE = delay loop counter ($100 for A = 1)
	ld e,0

	dec l		; 0-7 value for D2:D1:D0 setting

	ld bc,CLKREG
	in a,(c)
	and $F8
	or l
	out (c),a

l_ritard2
	xor a
l_ritardo
	dec a
	jr nz,l_ritardo
	dec de
	ld a,d
	or e
	jr nz,l_ritard2
exit_nmi
	pop de
	pop bc
	pop hl
        jp nocode_exit


;
;********************************************************************************************
; SNAPSHOT 16/48K/128K
;********************************************************************************************
;
; Nuova versione che sblocca i due bit di LOCK (D5 $7FFD per i registri $7FFD e $1FFD,
; D6 $34DF per tutti i registri non-standard) mediante scrittura di $AA su $44DF (UNLOCKREG).
;
; Prima di fare questo, salva il contenuto di $7FFD, $34DF, $1FFD in variabili shadow_ram.
; Per i primi due il motivo e` il ripristino al termine dello snapshot; per il $1FFD invece
; si tratta di una temporanea disabilitazione di eventuale modalita` speciale di indirizzamento,
; che non consentirebbe di snappare la ram in modo corretto. Se D0 = '1' (special addressing
; mode), esso viene forzato a '0' ed il tipo di snapshot posto a 128K. In questo modo, il
; contenuto della ram verra` correttamente salvato mediante tradizionale indirizzamento
; su $7FFD (RAM_BANK), leggendo nell'area $C000 - $FFFF.
;
; In uscita, i registri vengono rimessi a posto.
;
;
; STACK CONSIDERATIONS during SnapShot:
;
; These considerations are valid where no SHADOW_ROM dedicated ram is available. From version
; 3.00 and later, stack is no longer a problem since she SP points to SHADOW_ROM space.
;
; The original description follows:
;
; When the system is freezed into a snapshot, the stack contains host-program data, then
; return address (saved on NMI) then complete register set (for the rs-232 version, some of
; them are saved on the PC instead).
;
; When the roll-back takes place, FIRST OF ALL the SP should point at the same address where it
; was when the last register was pushed, and the rollback routine should terminate by poping
; all registers in the same order, then performing a RETN (that will actually return control
; to the host program, as nothing happened in the meantime).
;
; While creatinging/restoring a 128K snapshot, as it takes a couple of seconds per 16K block, the
; BORDER color is loaded with the number of the ram_bank being accessed.
;
; The make/get snapshot routines, however, will need the stack for their operations.
;
; STACK ISSUES while TAKING a snapshot:
; After all registers has been saved on the stack, the whole memory content is snapped into
; nonvolatile memory (either SD/MMC or by rs-232 transmission). Then, CPLD registers and SP
; value (along with some other infos) are saved in the entry header (in the FAT for SD/MMC
; or sent to the host for rs-232 snapshot). In the meantime, the stack can be used FREELY to save
; call's return-addresses and registers for take_snapshot routines.
; This will just alter ram locations BELOW the vital informations (registers and RETN address)
; and, provided that the stack space is large enough, this is a ram region that can be written
; to without problems.
; THE ONLY CONCERN THAT MUST BE OBSERVED is that, when saved to header informations, the SP
; SHOULD point to the first of the registers to be recovered during snapshot rollback, without
; regard to any other temporary data that may be on the stack for routines operations.
; If no other infos are on the stack at SP save time, then the SP can be stored as it is.
; If, for example, we know that a 16 bit register was pushed and not poped yet, then the saved
; SP value should be ADDED BY +2: this way, at restore time we will just forget about this
; additional data on the stack (which can be considered as garbage).
;
; STACK ISSUES while RESTORING a snapshot:
; The restore routine works in the opposite way: it first restores the SP register content, then
; reads back the whole memory, then restores CPLD registers, pops all registers and performs RETN.
; The reason for immediately having the SP in the right place is simple: while restoring
; memory content, the routines will need to save data and return addresses on the stack.
; If the stack points to the stack memory area of the host program, then no vital informations
; will be ever overwritten just after being restored. The SP, in fact, will write just BELOW
; the area that, after memory recovery, will hold again the registers block and RETN address
; that was there at snapshot time.
; The only critical point is the subroutine that restores the memory content: it will be called
; by the main snapshot-restore routine, but during it's operations, the subroutine will
; overwrite it's own return-to-main-program address. For this reason, this subroutine POPS
; the stack into a register as the first thing. Then it performs the whole memory restore.
; Then, just before performing the RET, it PUSHES that register back to the stack.
; This way, the routine will restore the two bytes needed for return to main routine.
; Note that during it's operations, this routine CAN make subroutine calls, provided that
; no memory write (restore of bytes of snapshot memory) are performed in the meanwhile.
; On the other hand, the routine that calls the memory-restore subroutine should NEVER RELY
; on stack for temporary data, since it will be lost.
;
; For safety purposes, since the SP may point to a banked memory area, the RAM_BANK is restored
; along with SP content at the value stored in the snapshot data, before starting the memory
; restore. No other CPLD registers are altered until the entire process is completed.
;
; When restoring 128K snapshots, the stack area may disappear on RAM_BANK switch.
; The rs-232 routine, which is simpler, makes no use of the stack during memory read/write.
; The SD/MMC routine needs a lot of calls in order to perform flash memory read and write:
; the SP is checked before starting and, if it points to a banked area (above $BFFF), a special
; procedure takes place. The SP is loaded with $5800 (begin of attribute area) then the real
; SP content is pushed on that temporary stack, actually writing at $57FF and $57FE (last video
; memory bytes). The snapshot procedure then goes on as always. This way, we don't overwrite
; bytes in ALL BANKS as they are switched in during snapshot operations and neither we loose
; stack data due to bank switch itself. The video region is the ONLY PLACE where the stack can
; save informations without corrupting any memory area. The SP value saved in the snapshot data
; will reflect the actual SP state (pointing to video ram) so no special care should be taken
; at restore time, except for the very last step. After all memory and CPLD registers have
; been restored, then the SP is checked against value $57FE, which is the value it will hold
; in case we moved it to video area. If the comparison is TRUE, then the SP content is loaded
; back from the SP area itself, restoring the original value. The registers can now be recovered
; and the RETN will take place. The only payload is a little bit of garbage on the bottom
; line of the screen, that will change gracefully :-) as the cpu saves data on the stack
; during the entire process. When this trick takes place, the BORDER turns immediately RED
; to indicate that the SP was reloaded. It then proceed with the regular color cycling based
; on the ram bank being read/written.
;
;
; OTHER NOTES:
; When taking a 16K snapshot, the SP register is checked. If found to point to a region
; above the saved area, the snapshot will have no chances to be restored successfully.
; The rs-232 version will abort operation and prompt the user to either reload the
; software after a CLEAR 32599 command (which will move the SP below the first 16K area)
; or perform a 48K snapshot.
; The SD/MMC version will automatically switch to 48K snapshot and turn the border RED,
; to inform the user that the snapshot will be 48K in size.
;
;
;


no_newclock
	ld e,MMC_SNAP16			; codice snapshot da 16KBytes
	cp 11				; [Q] tasto SNAPSHOT 16K
	jr z,takesnap

	ld e,MMC_SNAP48			; codice snapshot da 48KBytes
	cp 12				; [W] tasto SNAPSHOT 48K
	jr z,takesnap

	ld e,MMC_SNAP128		; codice snapshot da 128KBytes
	cp 13				; [E] tasto SNAPSHOT 128K
	jr z,takesnap

	jr w_keyclock			; attesa di un tasto valido (eventualmente mettere TIMEOUT)

filename
	db	"New SnapShot  ...               "

takesnap
	ld hl,sh_flags		; if the SD CARD is not initialized for our proprietary snapshot scheme,
	bit sd_initsnap,(hl)	; turns the border WHITE then exits immediately
	jr nz,card_ok
	ld a,7
	out (BORDER),a
	jr exit_nmi

card_ok
	exx			; la strana sequenza di salvataggio ricalca quella usata nella routine UART,
	push bc			; al fine di mantenere una certa compatibilita` negli snapshot. Questi due
	push iy			; registri vengono usati attivamente nelle routines a seguire.

				; nello snapshot UART, tutti i registri che seguono non vengono salvati sullo
				; stack ma bensi` trasmessi e scritti nel file .snap, per minimizzare l'uso dello
				; stack stesso. Nel nostro caso, invece, li salviamo tutti sullo stack.
				; la sequenza e` stata rispettata, con la sola differenza che prima di AF'
				; viene trasmesso lo SP.

	ex af,af'		; ALTRI REGISTRI, salvati per averli sullo stack e poterli poi ripristinare
	push af			; in fase di recupero dello snapshot stesso.
	ex af,af'
	push de		
	push hl			; salvataggio secondo la sequenza con cui i registri vengono salvati nel file .snap
	exx			; (solo per avere una qualche forma di equivalenza qualora si decida di scrivere
				; un programma di conversione: sotto ogni altro punto di vista, e` inutile).
        push ix


; Ora verifica se ci sono dei colori nei registri ULA+ e, nel caso, li copia
; in ram video nelle ultime due righine in basso (sono 64 bytes).
; Inoltre setta il flag use_ulaplus, che verra' trasferiso nei dati dello
; snapshot.

        call UlaPlus_Save

;
; Procede con lo snapshot. Piu' avanti il flag ulap_in_use verra' salvato
; fra i dati nell'header dello snapshot.
;
takesnap_2
	ld bc,S128KREG		; registro $7FFD (contiene D5 = LOCK)
	in l,(c)

	ld bc,SPECTR_MODE	; registro $34DF (contiene D6 = LOCK registri non-standard)
	in h,(c)

	ld bc,UNLOCKREG		; unlocks both D5 on $7FFD and D6 on $54DF (standard and non-standard LOCKS)
	ld a,$AA
	out (c),a

	ld a,l
	ld (snap_7ffd),a	; salva $7FFD in una variabile in shadow_ram
	ld a,h
	ld (snap_34df),a	; salva $34DF in una variabile in shadow_ram


	exx
	ld bc,CLKREG		; new: gets the actual Z80 clock rate
	in a,(c)
        ld l,a                  ; saves to L'
        and $F8

        if WRSNAP_21MHz = 1
	or 3			; then selects 21MHz
        else
        or 2                    ; then selects 14MHz
	endif
	out (c),a
;       out (c),a
;	ld hl,rtc_data
;	call rtc_getdata	; updates the DATE/TIME informations from RTC
;       ld a,(prev_clock)


	exx			; E register (snapshot size) has been preserved

	ld hl,0
	add hl,sp		; SP in HL
	ld (real_sp),hl		; saves to a shadow_rom variable
	ld sp,SHADOW_STACK

	ld bc,PLUS3REG		; registro $1FFD
	in a,(c)
	ld (snap_1ffd),a	; lo salva in una variabile in shadow_ram
	bit 0,a			; D0 SET = modalita` di indirizzamento speciale: tutti i banchi dovranno essere salvati
	jr z,no_specialmode
	ld e,8			; se e` attiva, forza uno snapshot da 128K
	xor a
	out (c),a		; disattiva lo special mode: dovra` essere ripristinato prima di uscire
	ld a,4
	out (BORDER),a		; segnala l'intervento mediante border verde

no_specialmode
	ld a,e			; codice dello snapshot (1 = 16K, 2 = 48K, 8 = 128K)
	cp MMC_SNAP16
	jr nz,no16ksnap		; non e` 16K
	bit 7,h			; lo stack punta ad un'area oltre i primi 16K?
	jr z,proceed_snap	; se e` basso, prosegue
	inc e			; altrimenti FORZA UNO SNAPSHOT 48K: se no, lo stack rimarrebbe fuori! (no clear 32599?)
	ld a,e
	out (BORDER),a		; segnala l'intervento mediante BORDER ROSSO
no16ksnap
proceed_snap

	ld c,e			; saves the snap type
	call mmc_get_firstword	; legge i primi 2 bytes della MMC --> DE, contenenti il numero della prima ENTRY LIBERA

	push de

; Calcola l'indirizzo di scrittura in SD/MMC per questo snapshot; DE = numero dell'entry


	ld l,FATSIZE		; n. blocchi da 64KB che costituiscono la FAT (e` l'offset per la word MSB)
	ld a,(card_offset)
	ld h,a
	sla e
	rl d			; moltiplica per 2 (ogni entry corrisponde ad un cluster da 128KB)
	add hl,de		; HL = word MSB scrittura in SD/MMC
	ld de,0			; word LSB = 0 (insieme a D0 = 0 della word MSB = 128K)

	ld a,c			; dimensioni dello snapshot (1 = 16K, 2 = 48K, 8 = 128K)
	ld iyl,c		; backup size of snapshot

	ld b,16384/BLOCKSIZE	; dimensioni per uno snapshot da 16K
	cp MMC_SNAP48		; MMC_SNAP48 = 2
	jr nz,mmc_snapnow	; 2 = 48K
	ld b,49152/BLOCKSIZE	; e` da 48K


; Esegue lo snapshot

mmc_snapnow
	cp MMC_SNAP128
	jr z,snap128sd		; salta se e` uno snapshot da 128K

	ld ix,$4000		; ram buffer address
	call mmc_write_data	; takes the snapshot (16/48K)
	pop de			; recovers the cluster number from stack
	cp 0
	jr z,read_entry_mmc	; jumps if no error

	jp mmcwrite_error

snap128sd
	ld bc,S128KREG		; 128K snapshot. The SP is not banked.
	in a,(c)
	push af			; original ram_bank
	and $F8			; RAM_BANK = 0
	ld c,a			; C = RAM_BANK counter, that keeps all other bits
l_128ksd
	ld a,c			; actual bank
	exx
	ld bc,S128KREG		; BC' for CPLD addressing (mmc_read_data overwrites it)
	out (c),a		; new ram_bank
	exx
	ld b,16384/BLOCKSIZE
	ld ix,$C000
	call mmc_write_data	; saves this 16K bank
	jr nz,l_128k_error
	inc c
	ld a,c
	and 7
	out (BORDER),a		; shows progress
	jr nz,l_128ksd

	pop af			; original ram bank
	ld bc,S128KREG
	out (c),a
	pop de
	jr read_entry_mmc
l_128k_error
	pop af
	pop de
	ld a,c
	ld bc,S128KREG
	out (c),a
	jp mmcwrite_error       ; termina
	


; Calcola l'indirizzo in cui si trova il blocco da 64 bytes che descrive questa entry (nella FAT) e ne legge
; il blocco (512 bytes)

read_entry_mmc
	push de			; salva di nuovo il numero dello snapshot: serve per riscrivere l'entry
	call calc_hldec		; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C
	
	ld ix,SHADOW_RAM	; scratchpad area
	call mmc_read_data	; performs the single block read
	pop de			; clears the stack before possible jump
	cp 0
        jp nz,mmcwrite_error

	push de			; saves back to the stack (block number that contains the entry)


; Ora calcola l'offset di QUESTA entry all'interno dei 512 bytes del blocco letto (3 bit LSB = 8 entry da 64 bytes)

	ld d,c			; D = 3 bit LSB numero dell'entry
	ld e,0			; E = 0, quindi = entry * 256
	srl d			; divide /4: adesso DE e` un offset di 64 bytes per ogni valore unitario
	rr e			; che avevano i 3 bit LSB, ovvero l'offset della entry da 64 bytes
	srl d			; all'interno del blocco da 512 letto precedentemente
	rr e

	ld hl,SHADOW_RAM+2	; NOTA: SALTA I PRIMI 2 BYTES, utili solo nella prima entry
	add hl,de		; HL = puntatore al byte con offset [2] dell'entry da modificare

	ld e,iyl
	ld (hl),e		; E = size of snapshot (1 = 16K, 2 = 48K, 8 = 128K) at offset +2
	inc hl
	ex hl,de
	ld hl,filename		; copia 32 bytes di filename ad offset +3 (sostituire con il vero nome?)
	ld bc,32
	ldir

	ld hl,rtc_data		; copia 7 bytes di data ad offset +35
	ld bc,7
	ldir

	ex hl,de

	ld ix,(real_sp)
	ld de,ix
	ld (hl),e		; saves SP (2 bytes) at offset +42
	inc hl
	ld (hl),d
	inc hl

	ld a,i
	push af			; LD A,I transfers IFF2 state (INT enable) in P/V FLAG (D2 in flag register)
	ld d,a			; saves I to D too
	ld (hl),a		; saves I register at offset +44
	inc hl

	ld a,(snap_7ffd)	; previously saved
	ld (hl),a		; saves $7FFD register at offset +45
	inc hl

	call get_immode		; loads IM_MODE with the actual IM MODE selected in the processor

	pop bc			; C = flag register
	ld a,c
	and 4			; keeps D2 (IFF STATE)
	ld c,a
	ld a,(im_mode)		; IM MODE previously detected
	add c
	ld e,a			; E = result to be saved as F register

	ld a,iyl		; IYL still holds the snapshot size
	cp MMC_SNAP128		; 8 = 128K snapshot
	jr z,s128mode
	set 3,e			; selects 48K mode for spectrum_type bit
s128mode
	ld a,(snap_34df)
	bit 1,a			; D1 = INT doubler bit
	jr z,no_doublei
	set 4,e			; che deve essere salvato su D4 del registro 'F'
no_doublei
	and $60			; bit 5 and 6 should be copied to F register
	or e			; (D5 = context_switch disable; D6 = all non-standard registers LOCK)
	ld e,a
	exx
	ld a,l			; get the $24DF register content without our clock shift
	exx
	res 6,a			; disables the SHADOW_ROM WR BIT, although it's ignored at restore time
        ld b,a                  ; saves in B

        res 7,e                 ; ULA+ not in use by default
        ld a,(sh_flags)
        bit ulap_in_use,a
        jr z,no_ulap_used
        set 7,e                 ; flag 'ulap_in_use' is saved in D7 at +46 offset
no_ulap_used
        ld (hl),e               ; saves IM ENABLE, IM MODE and SPECTRUM_TYPE at offset +46
	inc hl

        ld (hl),b               ; saves $24DF register at offset +47
	inc hl

	ld iyl,e		; IYL now holds the F register (for INT ENABLE STATUS)

	xor a
        ld b,10                 ; 10 bytes joystick programming are set to ZERO
l_clearbytes
	ld (hl),a
	inc hl
	djnz l_clearbytes

	ld a,(snap_1ffd)	; original $1FFD register content
	ld (hl),a		; saves it at offset +58
	inc hl
	ld bc,PLUS3REG
	out (c),a		; restores the register content (previously set to zero in order to access all ram banks at $C000)

	ld bc,S128KREG
	ld a,(snap_7ffd)	; previously saved (restores LOCK status)
	out (c),a

	ld bc,FASTPAGE		; $54DF register
	in a,(c)
	ld (hl),a		; saves it at offset +59
	inc hl

        ld (hl),0               ; ZXMMC+ fastpage at offset +60: zero is an
        inc hl                  ; acceptable default

        ld bc,DDR_MSB           ; $64DF register
	in a,(c)
        ld (hl),a               ; saves it at offset +61
	inc hl

        xor a
        ld b,2                  ; 2 spare bytes are set to ZERO
l_sparebytes
	ld (hl),a
	inc hl
	djnz l_sparebytes


; Il blocco dati e` stato modificato: e` ora di riscriverlo nella SD/MMC

	pop de			; recupera l'indirizzo usato per la lettura
	call calc_hldec		; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C
	ld b,1			; number of blocks to write
	ld ix,SHADOW_RAM	; scratchpad area
	call mmc_write_data
	cp 0
	jr nz,mmcwrite_error


; A questo punto occorre aggiornare il numero della prima entry libera, che e` memorizzato nei due bytes
; iniziali della prima entry (ovvero i primi due bytes della SD/MMC). Notare che questo campo e` inutilizzato
; in tutte le altre entry.

; Lo incrementiamo +1

	ld de,0
	ld a,(card_offset)
	ld h,a
	ld l,0
	ld ix,SHADOW_RAM	; scratchpad area
	call mmc_read_data	; performs the single block read
	cp 0
	jr nz,mmcwrite_error
	ld hl,(SHADOW_RAM)
	inc hl			; incremento a 16 bit
	call check_hlspace	; se andiamo oltre il limite della SD/MMC, non incrementa
	jr nc,dontincr
	ld (SHADOW_RAM),hl
	ld de,0
	ld a,(card_offset)
	ld h,a
	ld l,0
	ld b,1			; 1 block
	ld ix,SHADOW_RAM	; scratchpad area
	call mmc_write_data	; performs the block write
	cp 0
	jr nz,mmcwrite_error

dontincr
	ld a,$D5		; test: wr ok
	ld ($5800),a
	jr end_write


mmcwrite_error
	ld a,18			; RED on RED
	ld ($5801),a		; error

end_write
	ld hl,(real_sp)
	ld sp,hl
;	call disable_wrshadow

	exx
	ld bc,CLKREG
	out (c),l		; restores Z80 CLOCK
	exx

	ld bc,SPECTR_MODE
	ld a,(snap_34df)	; restores D6 content (non-standard register's lock)
	out (c),a

	ld a,iyl
	bit 2,a
	jr z,no_intenable
	and 3			; checks the IM-MODE
	cp 2
	jr nz,no_im2		; IM-MODE is not IM2: safe to enable interrupts

	im 1			; IM2: if the host's interrupt handler deals with BASIC ROM,
	ei			; the program will crash as we are still running in BOOTROM.
	halt			; Shoots a couple of interrupts with IM1 mode selected, then
	halt			; switches back to IM2 with enough time for exiting the rom
	im 2			; before next interrupt will occur
	jr no_intenable
no_im2
	ei
no_intenable
	pop ix
	exx
	pop hl
	pop de
	ex af,af'
	pop af
	ex af,af'
	pop iy

	pop bc
	exx
	jp exit_nmi


;********************************************************************************************
; GESTIONE COMANDI DA SERIALE, fase di autenticazione con l'HOST per entrare nel main LOOP
;********************************************************************************************

qualcosa
	and 6				; isola i due bit di errore (D1 = overrun, D2 = framing)
	jr z,riceok			; salta se sono entrambi bassi
	in a,(RXREG)			; vuota il registro, cancellando l'errore
	jr nocode_exit			; esce
riceok
	in a,(RXREG)			; riceve il byte
	cp $55				; primo byte deve essere $55
	jr z,stop_nmi			; salta nella ricezione del resto della sequenza, disattivando NMI
	ld a,'!'			; altrimenti trasmette "!" e rientra.
	call rs_tx

nocode_exit
	xor a
	ld (nmi_count),a
        ld a,(sh_flags)
        bit bootrom,a
        jr z,no_bootrom
	pop af
	retn				; abbiamo interrotto la BOOTROM: rientra SENZA DISATTIVARE IL CONTEXT!

no_bootrom
	pop af
	jp $15				; rientro con RETN e disattivazione CONTEXT_SWITCH

;
;--- da qui in avanti vengono disabilitati NMI da parte della CPLD
;
stop_nmi
	xor a
	out (NMIREG),a	; disabilita ulteriori NMI in caso di ricezione da seriale

	ld a,'*'
	call rs_tx	; trasmette conferma ricezione primo codice della sequenza

	push hl
	push bc		; a questo punto sullo stack abbiamo BC, HL, AF

	ld b,4		; sequenza di altri 4 bytes identificativo entrata nella routine
	ld c,$65	; codice del secondo byte
lcheckl
	call rs_rx	; 17T + 81T riceve il byte
	jp c,exit_node	; 10T
	cp c		; 4T
        jp nz,exit_node ; 10T uscita senza pop DE e con riattivazione NMI da parte dell'UART
	sla c		; 8T terzo codice = $CA; poi $94, $28
	djnz lcheckl	; 13T attende il prossimo byte

	push de
	exx
	push bc
	exx		; (BC' viene usato nella gestione dei comandi R e T)
	push iy		; stack: IY, BC', DE, BC, HL, AF.

;	ld a,$F9
;	ld (ATTRIB1),a	; cubetto lampeggiante bianco/blu (test)

        ld a,(sh_flags)
        bit bootrom,a
        ld a,5          ; border CYAN se interrompiamo la bootrom
        jr nz,outborder
	ld a,(BORDCR)	; colore attuale del border *8 (variabile del basic sinclair)
	rra
	rra
	rra
	dec a		; lo decrementa di 1
	and 7
outborder
	out (BORDER),a	; gli altri bit possono essere ZERO (SPKR, TAPE)

ok_waitcmd
	ld a,'*'	; codice OK di conferma ricezione sequenza di entrata
	call rs_tx


;
;---------------------------------------------------------------------------------------------
; MAIN LOOP attesa comandi
;---------------------------------------------------------------------------------------------
waitcmd
	call rs_rx	; attende un altro byte
        jp c,rx_error   ; salta se si e` verificato un errore


;
;---------------------------------------------------------------------------------------------
; Gestione comando "R" (ricezione di un blocco)
;
; Distrugge AF, BC, BC', DE, HL
;
; Il banco richiesto in RX_DBLOCK viene caricato nel registro solo dopo il rientro dalla
; RX_DBLOCK stessa: lo stack potrebbe sparire. Per lo stesso motivo non usiamo la subroutine
; RS_RX ma una sua copia, in locale.
;---------------------------------------------------------------------------------------------

	cp 'R'		; codice "R" = riceviamo un blocco
	jr nz, nor
;        ld a,$FA
;        ld (ATTRIB1),a  ; cubetto lampeggiante bianco/rosso (test)

	call rx_dblock	; HL = puntatore; BC = len; attuale RAM_BANK --> D; RAM_BANK voluto --> E
	jr nc,rxd_ok	; salta se tutto OK
	ld a,d		; ripristina il ram_bank, che e` stato sicuramente letto da RX_DBLOCK
	ld bc,S128KREG
	out (c),a
        jp rx_error

rxd_ok
	ld a,'*'		; codice OK
	call rs_tx
	ld a,e			; RAM_BANK desiderato
	exx
	ld bc,S128KREG		; solo ora commuta il banco, usando BC alternativo
	out (c),a
	exx
	ld a,d
	exx
	ld b,a			; B' = banco su cui tornare alla fine
	ld c,0			; C' = flag di errore
	exx
	ld d,0			; DE usato per il calcolo del checksum
	ld iy,0			; contatore di checksum
l_ric
	in a,(STATUS)
	and 7
	cp 2
	jr c,okrx
	exx
	ld c,1			; flag di errore
	exx
okrx
        bit 0,a			; 8T ATTESA PRESENZA DI UN BYTE
        jr z,l_ric		; 7T
	in a,(RXREG)		; 11T
	ld (hl),a		; salva il byte
	ld e,a
	add iy,de		; somma al checksum
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_ric

	exx
	ld a,b			; ripristina il ram_bank, che era stato salvato in B'
	exx
	ld bc,S128KREG
	out (c),a

	call rs_rx
	cp iyl			; confronta LSB
	jr z,rtr_ok1
	exx
	ld c,1			; flag di errore
	exx
rtr_ok1
	call rs_rx
	cp iyh			; confronta la parte MSB
	jr nz,errex		; errore

	exx
	ld a,c			; recupera il flag di errore da C'
	exx
	cp 0
	jr z,ok_waitcmd		; rientra con trasmissione di codice OK
errex
	ld a,'#'		; codice di errore
	call rs_tx
	jr waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando "T" (trasmissione di un blocco)
;
; Distrugge AF, BC, DE, HL
;
; Il banco richiesto in RX_DBLOCK viene caricato nel registro solo dopo il rientro dalla
; RX_DBLOCK stessa: lo stack potrebbe sparire. Per lo stesso motivo non usiamo la subroutine
; RS_TX ma una sua copia, in locale.
;---------------------------------------------------------------------------------------------
nor
	cp 'T'		; codice "T" = trasmettiamo un blocco
	jr nz,not
;        ld a,$FC
;        ld (ATTRIB1),a          ; cubetto lampeggiante bianco/verde (test)

	call rx_dblock		; HL = puntatore; BC = len; attuale RAM_BANK --> D; RAM_BANK voluto --> E
	jr nc,rxd_ok2		; salta se tutto OK
	ld a,d			; ripristina il ram_bank, che e` stato sicuramente letto da RX_DBLOCK
	ld bc,S128KREG
	out (c),a
        jp rx_error
rxd_ok2
	ld a,'*'		; codice OK
	call rs_tx
	ld a,e			; RAM_BANK desiderato
	exx
	ld bc,S128KREG		; solo ora commuta il banco, usando BC alternativo
	out (c),a
	exx
	ld a,d
	exx
	ld b,a			; B' = banco su cui tornare alla fine
	exx
	ld d,0			; DE usato per il calcolo del checksum
	ld iy,0			; contatore di checksum
l_tra
	in a,(STATUS)
	bit 3,a			; attesa TRANSMITTER IDLE
	jr nz,l_tra
	ld a,(hl)		; trasmette il byte
	out (TXREG),a
	ld e,a
	add iy,de		; aggiorna il checksum
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_tra

	exx
	ld a,b			; ripristina il ram_bank
	exx
	ld bc,S128KREG
	out (c),a

	ld a,iyl		; trasmette LSB checksum
	call rs_tx
	ld a,iyh		; trasmette MSB checksum
	call rs_tx

        jp ok_waitcmd           ; trasmette OK e rientra in loop attesa comandi


;
;---------------------------------------------------------------------------------------------
; ESC command handler (NMI-handler exit).
;
; If we just took an rs-232 snapshot, then we know that the IM-MODE detection routine has
; overwritten (reset) the IFF2 interrupt enable flag, so maskable interrupt won't be enabled
; on NMI EXIT. This would lock the host program when the NMI handler will be exited, by
; pressing the LogOUT button on the zx-com PC software.
;
; To solve the problem, PC software checks if a take-snapshot with interrupt enabled took
; place. If this is the case, when the user pressed the LogOUT button, a special ESC code is
; sent:
;
; ESC+1 (28) = re-enable interrupt;
; ESC+2 (29) = set IM2 MODE and re-enable interrupt.
;
; Of course, the command 29 is sent only if IM2 mode was detected by snapshot routines.
;
; There should be no need to deal with IM MODE. Hovewer, since it's not possible to just set
; the IFF2 flag (that would enable maskable interrupt on RETN execution), our only chance is
; to execute a real EI (that sets both IFF1 & IFF2).
; Unfortunately, if IM MODE was 2 AND the host's program interrupt handler relies on the
; zx-spectrum BASIC ROM, then it will crash because at least one interrupt will be shooted
; while we are still in the BOOTROM firmware.
;
; To solve this problem, if the host sends ESC+2, the IM-MODE is initially set to IM1 and EI
; is executed, followed by a couple of HALT instructions (that will halt the processor until
; a maskable interrupt is executed). Then, we have about 20ms (10 if interrupt works @100Hz)
; to set IM2 then exit the bootrom firmware. The host's maskable interrupt handler will then
; find the correct rom.
;---------------------------------------------------------------------------------------------
not
	cp 27
	jr z,esc	; standard exit mode
	cp 28		; ESC+1 = exit EI execution
	jr z,esc_ei1
	cp 29
	jr nz,noesc

	ld a,'*'	; we should set IM2 and enable interrupts
	call rs_tx	; sends OK code now, to save time later
	im 1		; IM1
	ei		; enable interrupt
	halt		; sync with INT hardware signal
	halt
	im 2		; IM2: dangerous if the handler deals with BASIC ROM
	jr exit_sub	; exit immediately, before a new INT will be generated by ULA

esc_ei1
	ei
esc
	ld a,'*'
	call rs_tx	; codice OK


exit_sub
	pop iy
	exx
	pop bc		; recupera BC'
	exx
	pop de		; recupera DE
exit_node
	pop bc
	pop hl		; recupera il resto dei registri

        ld a,(sh_flags)
        bit bootrom,a
        ld a,0          ; border nero se siamo in bootrom
        jr nz,outbordere
	ld a,(BORDCR)	; colore attuale del border *8 (variabile del basic sinclair)
	rra
	rra
	rra
	and 7
outbordere
	out (BORDER),a	; gli altri bit possono essere ZERO (SPKR, TAPE)

	ld a,1
	out (NMIREG),a	; riabilita l'NMI per la prossima ricezione di un byte

	jp nocode_exit

;	xor a
;	ld (nmi_count),a
;
;	pop af
;	jp $15		; rientro con RETN


;
;---------------------------------------------------------------------------------------------
; Gestione comando "S" (invio all'host contenuto dei registri 7FFD, 24DF, 34DF,
; 1FFD, 54DF e 64DF)
;
; Distrugge AF, BC
;---------------------------------------------------------------------------------------------
noesc
	cp 'S'
	jr nz,nos
	ld bc,S128KREG		; $7FFD
	in a,(c)
	call rs_tx
	ld bc,CLKREG		; $24DF
	in a,(c)
	call rs_tx
	ld bc,SPECTR_MODE	; $34DF
	in a,(c)
	call rs_tx
	ld bc,PLUS3REG		; $1FFD
	in a,(c)
	call rs_tx
	ld bc,FASTPAGE		; $54DF
	in a,(c)
	call rs_tx
        ld bc,DDR_MSB           ; $64DF
	in a,(c)
	call rs_tx
        jp ok_waitcmd           ; trasmette OK e rientra in loop


;
;---------------------------------------------------------------------------------------------
; Gestione comando "G" (ricezione dall'host dati da scrivere nel registro 7FFD).
;
; GRAZIE AL CONTEXT SWITCH, possiamo scrivere anche nel bit D4 (LSB ROM_SELECT)
;
; Restituisce un codice di errore '$' se il registro risulta bloccato in scrittura.
;
; Distrugge AF, BC, E
;---------------------------------------------------------------------------------------------
nos
	cp 'G'
	jr nz,nog
	call rs_rx
        jp c,rx_error           ; in caso di errore esce dalla sequenza
	ld e,a
	ld bc,S128KREG		; Registro $7FFD
	in a,(c)
	bit 5,a
	jr z,proced		; salta se il registro non e` bloccato in scrittura
	ld a,'$'		; codice di errore: il registro e` bloccato in scrittura
	call rs_tx
        jp waitcmd
proced
	out (c),e		; scrive il dato nel registro

        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando "C" (ricezione dall'host dati da scrivere nel registro 24DF).
;
; Distrugge AF, BC, E
;---------------------------------------------------------------------------------------------
nog
	cp 'C'
	jr nz,noc
	call rs_rx		; attende il byte del parametro
        jp c,rx_error           ; in caso di errore esce dalla sequenza
	ld e,a
	ld bc,SPECTR_MODE
	in a,(c)
	bit 6,a
	jr z,proceed_24df
	ld a,'$'		; codice di errore: il registro e` bloccato in scrittura
	call rs_tx
        jp waitcmd
proceed_24df
	ld bc,CLKREG
	out (c),e
        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando "D" (impostazione registro $34DF)
;
; Distrugge AF, BC, E
;---------------------------------------------------------------------------------------------
noc
	cp 'D'
	jr nz,nod
	call rs_rx		; attende il byte del parametro
        jp c,rx_error           ; in caso di errore esce dalla sequenza
	ld e,a
	ld bc,SPECTR_MODE
	in a,(c)
	bit 6,a
	jr z,proceed_34df
	ld a,'$'		; codice di errore: il registro e` bloccato in scrittura
	call rs_tx
        jp waitcmd
proceed_34df
	out (c),e
        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando "H" (impostazione registro $1FFD)
;
; Distrugge AF, BC, E
;---------------------------------------------------------------------------------------------
nod
	cp 'H'
	jr nz,noh
	call rs_rx
        jp c,rx_error           ; in caso di errore esce dalla sequenza
	ld e,a
	ld bc,S128KREG		; Registro $7FFD
	in a,(c)
	bit 5,a
	jr z,proced1f		; salta se il registro non e` bloccato in scrittura
	ld a,'$'		; codice di errore: il registro e` bloccato in scrittura
	call rs_tx
        jp waitcmd
proced1f
	ld bc,PLUS3REG
	out (c),e		; scrive il dato nel registro
        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando "K" (impostazione registri $54DF e $64DF)
;
; Distrugge AF, BC, DE
;---------------------------------------------------------------------------------------------
noh
        cp 'K'
        jr nz,nok
	call rs_rx
        jp c,rx_error           ; in caso di errore esce dalla sequenza
	ld e,a
	call rs_rx
        jp c,rx_error           ; in caso di errore esce dalla sequenza
        ld d,a
	ld bc,SPECTR_MODE
	in a,(c)
	bit 6,a
        jr z,proceed_54df
	ld a,'$'		; codice di errore: il registro e` bloccato in scrittura
	call rs_tx
        jp waitcmd
proceed_54df
        ld bc,FASTPAGE
	out (c),e		; scrive il dato nel registro
        ld bc,DDR_MSB
        out (c),d               ; scrive il dato nel registro
        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando J, salto ad un indirizzo ricevuto come parametro. Il contesto va` perso
;---------------------------------------------------------------------------------------------
nok
	cp 'J'
	jr nz,noj

	call rs_rx
        jp c,rx_error
	ld l,a			; riceve 2 bytes: indirizzo del salto
	call rs_rx
        jp c,rx_error
	ld h,a
	push hl			; indirizzo del salto sullo stack
	ld a,'*'
	call rs_tx		; trasmette OK

	ld a,1
	out (NMIREG),a		; riabilita l'NMI per la prossima ricezione di un byte

	xor a
	ld (nmi_count),a

        ld a,(sh_flags)
        res bootrom,a
        ld (sh_flags),a

	jp $15			; rientro con RETN (grazie al salto a $15 la CPLD disattiva
				; la nostra ROM $C000 subito dopo il fetch del byte $16)


;
;---------------------------------------------------------------------------------------------
; Gestione comando 'Q' (trasmissione contenuto di tutti i registri che non abbiamo salvato
; sullo stack) ad uso SNAPSHOT.
;
; Ordine di trasmissione dei registri:
;
; SP	registro n. 1	2 bytes
; AF	sullo stack, non trasmesso
; BC	"
; DE	"
; HL	"
; AF'	registro n. 2	2 bytes
; BC'	sullo stack, non trasmesso
; DE'	registro n. 3	2
; HL'	registro n. 4	2
; IX	registro n. 5	2
; IY	sullo stack, non trasmesso
;
; I	registro n. 7 (1 byte)
; $7FFD CPLD (1 byte)
;
; F	altro byte per salvataggio stato IFF (1 byte)
; $24DF CPLD (1 byte)
;
; $1FFD CPLD (1 byte)
; $54DF CPLD (1 byte)
;
; $7F   CPLD ZXmmc+ (dummy: trasmette 0)
; $64DF FPGA (1 byte)
;
; TOTALE = 18 bytes
;
; NUOVA VERSIONE che maschera il registro F a ZERO tranne D2 (stato IFF flag) e carica nei
; bit D1-D0 lo stato IM MODE, realmente rilevato. Anche D7 viene preservato
; con lo stato di attivazione dell'ULA+.
; Tale rilevamento provoca l'azzeramento di eventuale IFF2, per cui non ci sara` riabilitazione
; degli INT all'uscita dall'handler NMI. A questo problema ovvia zx-com, che nel momento in cui
; l'utente decide di uscire dalla sessione (col bottone "LogOUT) provvede eventualmente a
; riabilitare gli interrupt.
;
; REGISTRO F:
; bit d1-d0 = IM MODE. 00 = IM 0, 01 = IM 1, 10 = IM 2.
; D2: 0 = DI, 1 = EI.
; D3: Spectrum mode 128/48K. Restituiamo SEMPRE ZERO, ci pensa il C++ a settarlo.
; D4: Double int status (proviene da D1 SPECTR_MODE)
; D5: Context_switch disable (proviene da D5 SPECTR_MODE)
; D6: All non-standard registers LOCKED (proviene da D6 SPECTRUM_MODE)
; D7: '1' = colori ULA+ abilitati
;
; Corrispondenza con i relativi bit nel registro SPECTR_MODE:
;
; F Register	D7	D6	D5	D4	D3	D2	D1	D0
; SPECTR_MODE	--	D6	D5	D1    not D0	--	--	--
;---------------------------------------------------------------------------------------------
noj
	cp 'Q'
	jp nz,noq

        call UlaPlus_Save       ; se ULA+ in uso, copia i colori in ram video
                                ; e setta il flag ulap_in_use

	ld iy,0			; checksum
	add iy,sp
	ld a,iyl
	call rs_tx		; trasmette SP
	ld a,iyh
	call rs_tx

	ex af,af'
	push af
	ex af,af'
	pop bc
	add iy,bc
	ld a,c
	call rs_tx		; trasmette AF'
	ld a,b
	call rs_tx

	exx

	add iy,de
	ld a,e
	call rs_tx		; trasmette DE'
	ld a,d
	call rs_tx

	push hl			; salva per checksum
	ld a,l
	call rs_tx		; trasmette HL'
	ld a,h
	call rs_tx

	exx

	pop bc
	add iy,bc		; somma HL' al checksum

        push ix
        pop bc
        add iy,bc
	ld a,ixl		; trasmette IX
	call rs_tx
	ld a,ixh
	call rs_tx

	ld a,i
	push af			; LD A,I ha trasferito lo stato di IFF2 nel flag P/V (D2 del flag register)
	ld e,a			; LSB somma checksum
	call rs_tx		; trasmette I

	ld bc,S128KREG		; $7FFD
	in a,(c)
	ld d,a			; MSB checksum
	add iy,de		; somma gli ultimi 2 bytes trasmessi
	call rs_tx

	ld d,e
	call get_immode		; loads IM_MODE with 1 for IM0 and IM1, and 2 for IM2. WARNING: this will overwrite
				; the IFF2 flag, so INT won't be re-enabled on exit from NMI handler.

	pop bc			; C = flag register
	ld a,c			; trasmettiamo il registro dei flag
	and 4			; preserva il solo D2 (IFF STATE)
	ld c,a
	ld a,(im_mode)
	add c
	ld e,a

	ld bc,SPECTR_MODE
	in a,(c)
	bit 1,a			; INT doubler bit
	jr z,no_indou
	set 4,e
no_indou
	and $60			; bit 5 and 6 should be just copied
	or e
        ld hl,sh_flags
        bit ulap_in_use,(hl)    ; UlaPlus_save ha rilevato ULA+ attiva?
        jr z,noulapuse
        set 7,a                 ; si: setta D7 del registro F
noulapuse
	ld e,a
        call rs_tx              ; trasmette il registro F

	ld bc,CLKREG
	in a,(c)
	ld d,a
	call rs_tx		; trasmette il registro $24DF

	add iy,de		; aggiorna il checksum

	ld bc,PLUS3REG
	in a,(c)
	ld e,a			; LSB checksum
	call rs_tx		; trasmette il registro $1FFD

	ld bc,FASTPAGE
	in a,(c)
	ld d,a
	call rs_tx		; trasmette il registro $54DF
	add iy,de		; aggiorna il checksum

	xor a
        call rs_tx              ; $7F zxmmc+ (dummy)
        ld e,a
        ld bc,DDR_MSB
	in a,(c)
	ld d,a
        call rs_tx              ; $64DF register

	add iy,de		; aggiorna il checksum

        ld a,iyl
	call rs_tx		; trasmette LSB checksum
	ld a,iyh
	call rs_tx		; trasmette MSB checksum

        jp ok_waitcmd


;
;---------------------------------------------------------------------------------------------
; Gestione comando 'W' (ricezione contenuto di tutti i registri che non abbiamo salvato
; sullo stack) + $7FFD, $1FFD e $54DF CPLD ad uso SNAPSHOT.
;
; Non ripristina F (non FLAG, ma quello usato per conoscere lo stato degli interrupt
; mascherabili al momento dello snap) ed il registro $24DF (gestiti dal comando 'P').
;
; Chiamare PRIMA DI RIPRISTINARE LA MEMORIA.
;
; L'utilita` di settare lo SP ed il RAM_BANK prima di ricevere i blocchi di memoria consiste
; nel fatto che le routines di ricezione sporcano almeno 2 livelli di stack, ed e` quindi
; opportuno che cio` avvenga dove lo stack e` effettivamente presente (cioe` dove si trovava
; al momento dello SNAP).
;
; E` tuttavia necessaria una operazione speciale per D5 del registro $7FFD (LOCK BIT):
; Non si deve bloccare adesso il registro, perche` il successivo comando 'P' avra` bisogno
; di scrivere in $1FFD. Per questo motivo, la presente routine salva il valore integrale
; nella variabile shadow "snap_7ffd" ed imposta il registro con D5 forzato basso.
; Il comando 'P', dopo che avra` opportunamente impostato $1FFD, provvedera` a traferire
; sul $7FFD il valore salvato nella variabile.
;
; E` importante relegare l'impostazione di $1FFD (e anche $54DF, fast_paging) al comando 'P'
; per poter effettuare il ripristino dei banchi di ram senza indirizzamenti speciali in corso.
; Anche in caso di indirizzamento speciale mediante 1FFD, l'intera ram dovra` essere scritta
; mediante il consueto ram_paging su $7FFD (su $C000 - $FFFF) per poi, alla fine, attivare
; la modalita` speciale su $1FFD.
;
;
; Ordine di trasmissione dei registri:
;
; SP	registro n. 1	2 bytes
; AF	sullo stack, non trasmesso
; BC	"
; DE	"
; HL	"
; AF'	registro n. 2	2 bytes
; BC'	sullo stack, non trasmesso
; DE'	registro n. 3	2
; HL'	registro n. 4	2
; IX	registro n. 5	2
; IY	sullo stack, non trasmesso
;
; I	registro n. 7 (1 solo byte)
; Registro CPLD $7FFD (1 byte)
;
; Totale 12 bytes
;---------------------------------------------------------------------------------------------
noq
	cp 'W'
        jp nz,now

	ld bc,UNLOCKREG
	ld a,$AA
	out (c),a

	call rs_rx	; riceve SP usando HL come area temporanea
	ld l,a
	call rs_rx
	ld h,a
	ld sp,hl	; in questo momento dovrebbe essersi ricostituita la situazione allo SNAP
        push hl
        pop iy          ; inizializza checksum

	call rs_rx
	ld c,a		; riceve AF' usando BC come area temporanea
	call rs_rx
	ld b,a
	push bc
	ex af,af'
	pop af
	ex af,af'
	add iy,bc	; somma al checksum

	exx

	call rs_rx	; riceve DE'
	ld e,a
	call rs_rx
	ld d,a
	add iy,de

	call rs_rx	; riceve HL'
	ld l,a
	call rs_rx
	ld h,a
	push hl		; salva per checksum

	exx

	pop de
	add iy,de	; somma al checksum

	call rs_rx	; riceve IX
	ld ixl,a
	call rs_rx
	ld ixh,a
        push ix
        pop de
        add iy,de

	call rs_rx	; riceve I
	ld i,a
	ld e,a		; LSB CHECKSUM

	call rs_rx	; riceve il registro $7FFD
	ld d,a		; salva per il checksum
	add iy,de	; aggiorna il checksum
	ld (snap_7ffd),a; salva il valore INTEGRALE del registro
	res 5,a		; dopodiche` disattiva il LOCK BIT
	ld bc,S128KREG
	out(c),a	; imposta il registro evitando il LOCK: verra` inserito dal comando 'P'

	call rs_rx
	ld c,a		; parte LSB checksum trasmesso dal PC
	call rs_rx
	cp iyh		; confronta la parte MSB
	jr nz,rty_err
	ld a,c
	cp iyl
	jr nz,rty_err

        jp ok_waitcmd	; termina con successo

rty_err
	ld a,'#'
	call rs_tx	; codice di errore
        jp waitcmd      ; NOTA: I REGISTRI SONO COMPROMESSI!


;
;---------------------------------------------------------------------------------------------
; Gestione comando 'P': ricezione di F (abilitazione interrupt), $24DF, $1FFD, $54DF e $7F CPLD.
;
; Termina con un SALTO NEL PROGRAMMA SNAPPATO.
;
; RIPRISTINA LO STATO DEL ROM_BANK ma NON LO SHADOW_ROM RD, che resta al contenuto attuale.
; Durante la nostra esecuzione l'indirizzamento e` forzato a "00011" a prescindere dal contenuto
; dei registri di ROM_BANK e possiamo quindi scriverci impunemente (il valore scritto diventa
; operativo al momento del JP $15, quando la CPLD revochera` il CONTEXT = "00011").
; Cio` non e` invece vero per il bit di controllo READ SHADOW_ROM, che esegue immediatamente
; lo SWITCH.
;
; La configurazione delle ROM non deve quindi piu` cambiare. Il ripristino del ROM_BANK e`
; necessario per gli SnapShot 128K, in cui questo puo` puntare alla rom 0 o 1.
;
; Nel caso fosse attiva una modalita` di indirizzamento speciale mediante il registro 1FFD o
; 54DF, ZX-Com si sara` occupato di disattivarli temporaneamente prima di impartire i comandi
; di lettura dei blocchi di ram (che, trattandosi necessariamente di uno snap 128K, sara`
; avvenuto nell'ultimo blocco da 16K con paging mediante $7FFD).
;
; REGISTRI RICEVUTI DA QUESTO COMANDO:
;
; F	byte per salvataggio stato INTERRUPT e spectrum_mode = 1 byte
; $24DF CPLD = 1 byte
; $1FFD	CPLD = 1 byte
; $54DF	CPLD = 1 byte
; $7F   CPLD = 1 byte (ignorato)
; $64DF FPGA = 1 byte
;
; Esegue un EI o DI a seconda dello stato che aveva IFF2 al momento della creazione dello snap.
; Rientra con RET invece di RETN in quanto non c'e' piu` bisogno di caricare IFF1 --> IFF2,
; sebbene questo non dovrebbe creare problemi dato che i due flag sono settati all'unisono
; da DI o EI.
;
; Resta cmq un margine di errore: un eventuale interrupt NMI in coincidenza di un EI
; salvera` in IFF2 lo stato LOW di IFF1, con conseguente mancata attivazione al termine
; rella routine (con RETN). Inoltre non sappiamo come salvare l'IM (0 - 2). Al momento
; non viene salvato.
;
;
; CONTENUTO DEL BYTE F:
;
; D1 - D0 = IM MODE (0, 1, 2)
;
; D2 = interrupt enable bit
;
; D3 = stato da caricare INVERTITO nello SPECTRUM_TYPE bit (D0 $34DF) (0 = 128K, 1 = 48K)
; (infatti quel bit deve essere 0 per selezionare il 48K)
;
; D4 = stato da caricare nel DOUBLE_INT bit (D1 $34DF) (0 = normale, 1 = double INT speed)
;
; D5 = '1' = context_switch disabled (no bootrom swap-in will occur on $66 opcode fetch)
;
; D6 = non-standard registers LOCK bit
;
; D7 = stato di attivazione dell'ULA+
;      Nel caso, sara' necessario ripristinare la tavolozza, che e' stata
;      salvata in ram video.
;---------------------------------------------------------------------------------------------
now
	cp 'P'
        jp nz,no_p

	call rs_rx	; riceve registro F per stabilire come settare IFF
        ld (tmp_reg),a  ; salva in una zona in shadow ram

	call rs_rx	; riceve il registro $24DF
	ld d,a
	res 7,d		; azzera il bit di selezione READ ENABLE SHADOW_ROM.
	ld bc,CLKREG	; ora scriviamo i nuovi dati nel registro CLKREG, salvati in D
	in a,(c)
	and $80		; preserva lo stato attuale dei bit di selezione rd shadow_rom
	or d
	out (c),a	; imposta il registro preservando solo eventuale RD shadow_rom

	call rs_rx
	ld bc,PLUS3REG
	out (c),a	; registro $1FFD (scrivibile grazie al fatto che il LOCK bit e` LOW)

	ld a,(snap_7ffd); e` ora di impostare (se necessario) il LOCK BIT. Il valore integrale
	ld bc,S128KREG	; del registro $7FFD e` stato salvato in questa variabile dal comando 'W',
	out (c),a	; che lo ha impostato tutto tranne D5 (LOCK).

	call rs_rx
	ld bc,FASTPAGE
	out (c),a	; registro $54DF

	call rs_rx	; $7F (ignorato)

        call rs_rx      ; dummy byte
        ld bc,DDR_MSB
        out (c),a       ; registro $64DF

	ld a,'*'
	call rs_tx

        ld a,(tmp_reg)  ; Registro F; D4, D3 da caricare in D1, D0 di SPECTR_MODE (D3 INVERTITO!)
        ld e,a
        bit 7,a                         ; flag ULA+ in uso
        call nz,UlaPlus_Restore         ; se necessario, ripristina colori

        ld a,e
        ld h,e
        and $60         ; D5, D6 da copiare pari-pari
        bit 3,h
	jr nz,bit3high
	set 0,a		; copia invertito
bit3high
        bit 4,h
	jr z,bit4low
	set 1,a		; copia
bit4low

	ld bc,SPECTR_MODE
	out (c),a	; ATTENZIONE: D6 SET blocca ogni altro accesso WR ai registri non standard!


	ld a,(BORDCR)	; colore attuale del border *8 (variabile del basic sinclair)
	rra
	rra
	rra
	and 7		; rimette a posto il colore originale
	out (BORDER),a	; gli altri bit possono essere ZERO (SPKR, TAPE)

        ld a,(sh_flags) ; non siamo piu' in BOOTROM (flag usato da NMI handler)
        res bootrom,a
        ld (sh_flags),a

        pop iy
	exx
	pop bc		; recupera BC'
	exx
	pop de		; recupera DE
	pop bc

	ld a,1
	out (NMIREG),a	; riabilita l'NMI per la prossima ricezione di un byte


	di		; not really necessary, but who knows...
	im 1		; default IM1
	bit 2,h		; interrupt enable real status
	jr z,no_enei
	ei		; this will execute our fake handler (EI, RETI)
	halt		; sync with the hardware maskable interrupt generator
	halt
no_enei			; now we have 20 (or 10) ms before next interrupt will occur
	ld a,h		; recupera lo stato di F dopo il LD A,I all'atto della creazione dello snapshot
	and 3		; bit d1-d0 = IM MODE. 00 = IM 0, 01 = IM 1, 10 = IM 2.
	cp 0
	jr z,imzero
	cp 1
	jr z,completedei
	im 2
	jr completedei
imzero
	im 0
completedei
	xor a
	ld (nmi_count),a

	pop hl
	pop af
	jp $15		; RETN


;
;---------------------------------------------------------------------------------------------
; Subroutine che TRASMETTE all'host un blocco da 512 bytes letti dalla SD/MMC.
; L'host deve trasmettere l'indirizzo del blocco da leggere. La SD/MMC e` indirizzata da un
; puntatore a 32bit, ma dato che leggiamo allineati ai 512 bytes l'host ne trasmette solo 24:
; si tratta del NUMERO DEL BLOCCO, che viene quindi SCALATO verso SX (moltiplicato *2) e poi
; caricato nei 3 bytes MSB dell'indirizzo (il byte LSB e` azzerato).
;
; Utilizza la SHADOW_ROM come buffer temporaneo
;---------------------------------------------------------------------------------------------
no_p
	cp 'Z'
        jr nz,noz

	ld a,'*'		; 'pronto'
	call rs_tx

	call rs_rx		; byte MSB indirizzo richiesto
	ld h,a

	call rs_rx
	ld l,a

	call rs_rx
	ld d,a

	sla d			; moltiplica *2
	rl l
	rl h

w_tx1	in a,(STATUS)
	bit 3,a			; attesa TRANSMITTER IDLE: la porta SPI e` condivisa con la UART
	jr nz,w_tx1		; a livello di contatori nella CPLD

	ld e,0			; byte LSB = 0. Adesso abbiamo l'indirizzo completo a 32bit
	ld ix,SHADOW_RAM
	call mmc_read_data
	cp 0
	jr z,r_sectok
	ld a,'#'		; error code
	call rs_tx
	jp waitcmd
r_sectok
	ld a,'*'
	call rs_tx

	ld d,0			; DE usato per il calcolo del checksum
	ld iy,0			; contatore di checksum
	ld hl,SHADOW_RAM
	ld bc,BLOCKSIZE
l_tra_mmc
	ld a,(hl)
	ld e,a
	call rs_tx
	add iy,de		; aggiorna il checksum
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_tra_mmc

	ld a,iyl		; trasmette LSB checksum
	call rs_tx
	ld a,iyh		; trasmette MSB checksum
	call rs_tx

        jp waitcmd


;
;---------------------------------------------------------------------------------------------
; Subroutine che RICEVE dall'host un blocco da 512 bytes, da scrivere nella SD/MMC.
; L'host trasmette i 512 bytes di dati da scrivere, seguiti dal numero del blocco in SD/MMC
; piu` il checksum.
;
; Utilizza la SHADOW_ROM come buffer temporaneo
;---------------------------------------------------------------------------------------------
noz
	cp 'X'
        jp nz,nox

	ld a,'*'		; codice OK
	call rs_tx

	ld iy,0			; contatore di checksum
	ld d,0			; DE usato per il calcolo del checksum
	exx
	ld c,0			; flag di errore
	exx
	ld hl,SHADOW_RAM
	ld bc,BLOCKSIZE
l_ric_mmc
	call rs_rx
	jr nc,no_err_m
	exx
	ld c,1
	exx
no_err_m
	ld (hl),a		; salva il byte
	ld e,a
	add iy,de		; somma al checksum
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_ric_mmc

	call rs_rx		; byte MSB indirizzo richiesto
	ld h,a
	ld e,a
	add iy,de		; aggiorna il checksum

	call rs_rx
	ld l,a
	ld e,a
	add iy,de		; aggiorna il checksum

	call rs_rx
	ld e,a
	add iy,de		; aggiorna il checksum

	ld d,a			; questo byte deve andare in D (adesso che abbiamo finito con i checksum)

	call rs_rx
	cp iyl			; confronta LSB
	jr z,rtr_ok1_mmc
	exx
	ld c,1			; flag di errore
	exx
rtr_ok1_mmc
	call rs_rx		; riceve MSB CHECKSUM
	cp iyh			; confronta la parte MSB
	jr nz,errex_mmc		; errore

	exx
	ld a,c			; recupera il flag di errore da C'
	exx
	cp 0
	jr nz,errex_mmc		; rientra con trasmissione di codice OK

	sla d			; moltiplica *2
	rl l
	rl h
	ld e,0			; byte LSB = 0. Adesso abbiamo l'indirizzo completo a 32bit

	ld ix,SHADOW_RAM
	ld b,1			; blocchi da scrivere
	call mmc_write_data
	cp 0
	jr nz,errex_mmc
        jp ok_waitcmd

errex_mmc
	ld a,'#'		; codice di errore
	call rs_tx
        jp waitcmd


;
;---------------------------------------------------------------------------------------------
; Subroutine che trasmette all'host 7 bytes relativi al contenuto dell'RTC.
; Per evitare che il dato sia inaccurato a causa di mancato incremento delle variabili
; locali, il PIC viene nuovamente interrogato per un update preventivo.
;---------------------------------------------------------------------------------------------
nox
	cp 'A'
        jp nz,noa

;	ld hl,rtc_data		; buffer for local date/time storage
;	call rtc_enquire	; read a copy of data from the PIC microcontroller

	ld hl,rtc_data
	ld b,7			; bytes to send
l_senddata
	ld a,(hl)
	call rs_tx
	inc hl
	djnz l_senddata

	ld a,'*'		; codice OK
	call rs_tx
        jp waitcmd


;
;---------------------------------------------------------------------------------------------
; Subroutine che riceve dall'host 7 bytes da scrivere all'interno del chip RTC.
;---------------------------------------------------------------------------------------------
noa
	cp 'B'
       jp nz,nob

	ld a,'*'		; codice OK
	call rs_tx

	ld hl,rtc_data
	ld b,7			; bytes to receive
l_getdata
	call rs_rx
	jp c,ext_error
	ld (hl),a		; data is in binary format
	inc hl
	djnz l_getdata

;	ld hl,rtc_data
;	call rtc_setup

	ld a,'*'
	call rs_tx		; RTC_SETUP returns an error code: '*' no error; '!' PIC/RTC error; 1 comm error
ext_error
        jp waitcmd



;
;---------------------------------------------------------------------------------------------
; This routine receives from host computer a 16/32/64 KB rom block, saves it in DDR memory
; in the second half of the current DDR_MSB megabyte (the "ROM" area), using fastpage mode
; at $C000.
;---------------------------------------------------------------------------------------------
nob
	cp 'E'
        jp nz,noe

        call SetFastpC000       ; stack and fastpaging registers are saved
        ld sp,SHADOW_STACK      ; stack in a safe place

        ld a,$E0                ; riceve sul banco ZERO di tipo "ROM", ovvero il banco 32 in assoluto (+512KB offset)
	ld bc,FASTPAGE
	out (c),a

	call rx_dblock		; HL = puntatore; BC = len; attuale RAM_BANK --> D; RAM_BANK voluto --> E
	jp c,e_f_common_exit	; error
;
; In realta` il parametri 'HL' ed 'E' vengono ignorati
;

	ld a,'*'		; codice OK
	call rs_tx

	exx
	ld c,0			; error flag
	exx
	call common_getrom	; ricezione del blocco sul DDR_BANK corrente, all'indirizzo $C000, autoincremento
	jr c,e_f_common_exit	; ricezione con errori: rientra

	xor a			; azzera il contatore di interrupt NMI: questa variabile sopravvive (a meno che
	ld (nmi_count),a	; la rom non sia da 32KB) ed e` vitale che non indichi erroneamente un nested NMI

	ld bc,S128KREG		; zero in all addressing banks, 3.5MHz
	out (c),a
        ld bc,PLUS3REG
	out (c),a
	ld bc,CLKREG
	out (c),a
	ld bc,FASTPAGE		; disables fastpage
	out (c),a

        inc a
	out (NMIREG),a		; riabilita l'NMI per la prossima ricezione di un byte

        ld a,(sh_flags)         ; non siamo piu' in BOOTROM (flag usato da NMI
        res bootrom,a           ; handler, fondamentale)
        ld (sh_flags),a

        ld hl,jp_reloc          ; si riloca in modo da poter disattivare
        ld de,16384             ; il context mode
	ld bc,jpreloc_end-jp_reloc
	ldir
	jp 16384

jp_reloc
        ld a,8                  ; disables $C000 fastpage mode, removes context switch
        ld bc,SPECTR_MODE
	out (c),a
        jp 0                    ; salta nella nuova rom, DDR_MSB rimane alterato
jpreloc_end



;
;---------------------------------------------------------------------------------------------
; Flash routines: programmazione della FLASH SPI a bordo della scheda Xilinx
;
; 64K of DDR are used as a buffer.
; Address is TMP_DDRMSB (6 msb bits addressing A25:A20)
; and TMP_FASTPAGE (one of 64 16K banks inside the megabyte specified by TMP_DDRMSB
;---------------------------------------------------------------------------------------------
noe
	cp 'F'
        jp nz,nof

                                ;!!! OCCORRE modificare una volta per tutte il meccanismo di rilevamento di
				;interruzione della bootrom stessa da parte del NMI, che si basa su registri vecchi

        call SetFastpC000       ; stack and fastpaging registers are saved
        ld sp,SHADOW_STACK      ; then moves it to a safe place

another_bank
	call rx_dblock		; BC = len; A = ROM_BANK da flashare; D = stato attuale del registro $7FFD
	jr nc,okf_rom		; salta se OK

e_f_common_exit			; ENTRY POINT USED BY 'E' CMD TOO
	ld a,'#'		; codice di errore
	call rs_tx

enof
        call StopFastpC000

	ld hl,(real_sp)		; restores stack pointer
	ld sp,hl
	jp waitcmd

;
; Il parametro 'HL' viene ignorato
;
okf_rom

	ld (flash_bank),a	; salva in una variabile non shadow (serviva che fosse sempre accessibile)
	ld (flash_towr),bc	; idem

	ld a,(sh_flags)		; in case this NMI interrupted the bootrom firmware, we signal that the screen
	set redraw_screen,a	; should be redrawn
	ld (sh_flags),a

	call cls
	xor a
	ld (x_pos),a
	ld (y_pos),a
	ld hl,backup64
	call disp_text

	ld iy,$5810		; attributi in ram video per rom_64Kcopy

	ld e,0			; 0 = operazione di copia e verifica
	call rom_64kcopy	; copia l'intero banco da 64Kbytes in RAM

	xor a
	ld (x_pos),a
	inc a
	ld (y_pos),a
	ld hl,receiving
	call disp_text

	exx
	ld c,0			; azzera il flag di errore per common_getrom
	exx

	ld a,(flash_bank)
	and 3			; seleziona il banco specifico da 16K per la ricezione dei dati
	or TMP_FASTPAGE
	ld bc,FASTPAGE
	out (c),a

	ld a,'*'		; codice OK
	call rs_tx

	ld bc,(flash_towr)
	call common_getrom	; gestisce la ricezione del blocco sul RAM_BANK selezionato
	jr c,enof		; ricezione con errori: rientra

	call flash_screen1	; displays user screen for eprom flashing

w_keyflash
	call key_scan
	jr nz,w_keyflash	; piu` di 2 tasti: ignora
	ld a,e
	cp 8			; N key
	jr nz,no_no
	ld hl,flash_abort
	jr espi_abort
no_no
	cp 2			; Y key
	jr nz,w_keyflash

	call cls
	xor a
	ld (x_pos),a
	ld (y_pos),a
	out (BORDER),a
comp_again
	call set_iy
	ld hl,comparing
	call disp_text
	ld e,1			; 1 = operazione di verifica (confronto)
	call rom_64kcopy
	cp 0
	jr z,espi_program	; OK: esce

	xor a
	ld (x_pos),a
	ld a,(y_pos)
	inc a
	ld (y_pos),a
	ld hl,erasing
	call disp_text

	call spi_erase		;cancella il settore da 64K che contiene il banco FLASH_BANK
	call set_iy
	ld (iy),32

	xor a
	ld (x_pos),a
	ld a,(y_pos)
	inc a
	ld (y_pos),a
	ld hl,programming
	call disp_text

	call set_iy
	call spi_64Kprogram
	xor a
	ld (x_pos),a
	ld a,(y_pos)
	inc a
	ld (y_pos),a
	jr comp_again

espi_program
	ld hl,flash_programok
espi_abort
	xor a
	ld (x_pos),a
	ld a,(y_pos)
	add a,3
	ld (y_pos),a
	call disp_text
	jp enof

set_iy
	ld a,(y_pos)		;sets IY to the screen attribute of current line +10 offset
	rrca
	rrca
	rrca			;*32
	ld c,a
	and $E0
	add 11
	ld iyl,a		;LSB attribute in screen line
	ld a,c
	and 3
	add $58
	ld iyh,a		;MSB attribute
	ret

flash_screen1
	call cls
	xor a
	ld (y_pos),a
        ld a,39
	ld (x_pos),a
	ld hl,flashmsg1
        call disp_text
        ld hl,flashmsg1b
        call disp_text
	ld a,(flash_bank)
	ld c,a
	call disp_hex

        ld a,2
	ld (y_pos),a
        ld a,39
	ld (x_pos),a
        ld hl,flashfpgaarea
        ld a,c
        cp $3C
        jr c,flashmsgok
        ld hl,flashbootarea
        cp $40
        jr c,flashmsgok
        ld hl,flashromarea
flashmsgok
        call disp_text
        ld a,10
	ld (y_pos),a
        ld a,39
	ld (x_pos),a
	ld hl,flashmsg2
	call disp_text
	ret


flashmsg1
        db      2,"About to FLASH ROM_BANK",0
flashmsg1b
        db      7," $",0
flashfpgaarea
        db      6,"[FPGA Configuration Area (!)]",0
flashbootarea
        db      6,"[Bootrom Firmware Area]",0
flashromarea
        db      6,"[User ROM Area]",0
flashmsg2
	db	7,"Proceed? (Y/N) ",0
backup64
	db	7,"64K Backup.......",0
receiving
	db	7,"Receiving Data...",0
programming
	db	7,"Programming...",0
comparing
	db	7,"Comparing.....",0
erasing
	db	7,"Erasing.......",0
flash_programok
	db	7,"OK. New roms will take effect on reboot.",0
flash_abort
        db      7,"Aborted. LogOut for main menu.",0

;
; Subroutine che copia 64Kbytes di ROM in 4 blocchi DDR specificati da TMP_DDRMSB e TMP_FASTPAGE.
; Richiede che il chiamante abiliti la modalita` FASTPAGE in $C000.
; Se E = 1 esegue solo il confronto, e rientra con A = 0 se OK oppure A = 1 se errore.
; IY deve puntare ad una zona di attributi dove visualizzare quadratini verdi o rossi
;
rom_64kcopy
        ld ix,tmp_reg

	ld a,(flash_bank)	; numero del BANCO ROM
	srl a			; n. banco e` da 16K: lo trasforma in 64K
	srl a
	ld (ix+1),a		; A23:A16 indirizzo SPI FLASH di lettura
	xor a
	ld (ix+2),a		; 16 bit LSB = 0000
	ld (ix+3),a
	ld a,TMP_DDRMSB
	ld (ix+4),a		; IX+3 = ddr_msb da usare
	ld a,TMP_FASTPAGE
	ld (ix+5),a		; IX+4 = fastpage da usare

	ld b,4			; banchi da copiare
	ld c,e			; flag: 0 = copia/verifica; 1 = solo verifica
loop_64Kcopy
	push bc
retry_64Kcopy
	ld a,c
	cp 1			;1 = verify only
	jr z,qcopy
	ld hl,$C000		;destination write pointer
	ld bc,16384		;bytes to copy
	ld e,0			;E = 0 = READ; 1 = VERIFY
	call spi_16Kread	;legge 16K in DDR
qcopy
	ld hl,$C000		;destination write pointer
	ld bc,16384		;bytes to copy
	ld e,1			;VERIFY
	call spi_16Kread	;confronta 16K in DDR
	ld a,e
	cp 0
	jr z,bank_rok
	ld (iy),16		;red box
	inc iy
	pop bc
	ld a,c
	cp 1			;era solo confronto?
	ret z			;rientra con A = 1 in caso di errore nel confronto
	push bc
	jr retry_64Kcopy	;altrimenti ritenta all'infinito (mettere contatore?)

bank_rok
	ld (iy),32		;green box
	inc iy
	ld a,(ix+2)
	add $40			;advance 16K in SPI address
	ld (ix+2),a
	inc (ix+5)		;next FASTPAGE
	pop bc
	djnz loop_64Kcopy	;next bank
	xor a			;se era confronto, A = 0 = OK
	ret


;
;---------------------------------------------------------------------------------------------
; 64K block program into SPI FLASH from TMP_FASTPAGE address, into the 64K SPI area which
; contains the 16K block specified in flash_bank.
;
; Caller should set proper DDR_MSB and enable fastpaging at $C000.
;---------------------------------------------------------------------------------------------
spi_64Kprogram
	ld a,TMP_FASTPAGE
	ld bc,FASTPAGE
	out (c),a		;first of four 16K banks is selected

	ld a,(flash_bank)	;numero del BANCO ROM destinazione
	srl a			;lo allinea al banco da 64K SPI che lo contiene
	srl a
	ld b,a			;MSB (A23:A16) SPI address

	ld hl,0			;byte pointer (64K area)
next_page
	call wr_enable
	call page_program	;programs 256 bytes
	call wait_idle
	ld a,h
	and $3F			;D5:D0 = parte che indirizza 16K (esclude D7:D6)
	jr nz,no_pagecross	;salta se non stiamo passando ad un nuovo banco
	ld (iy),48
	inc iy
	ld e,b
	ld bc,FASTPAGE
	in a,(c)
	inc a
	out (c),a
	ld b,e
no_pagecross
	ld a,h
	cp 0			;64K completed?
	jr nz,next_page
	ret


;
;---------------------------------------------------------------------------------------------
; SPI flash sector erase (64K) - cancella il settore che contiene il banco da programmare
;---------------------------------------------------------------------------------------------
spi_erase
	call wr_enable
	xor a
	out (OUT_PORT),a	;seleziona CS3 = spi onboard

	ld a,216		;sector erase
	out (SPI_PORT),a	;command byte

	ld a,(flash_bank)	;numero del BANCO ROM
	srl a			;n. banco e` da 16K: lo trasforma in 64K
	srl a
	out (SPI_PORT),a
	xor a
	nop
	out (SPI_PORT),a
	nop
	nop
	out (SPI_PORT),a
	ld a,3
	out (OUT_PORT),a

	call wait_idle
	ret

;
;---------------------------------------------------------------------------------------------
; This routine enables WR mode for the M25P16 spi flash chip
;---------------------------------------------------------------------------------------------
wr_enable
	xor a
	out (OUT_PORT),a
	ld a,6			;wr enable
	out (SPI_PORT),a
	ld a,3
	nop
	out (OUT_PORT),a
	ret


;
;---------------------------------------------------------------------------------------------
; This routine waits for idle M25P16 spi flash chip
;---------------------------------------------------------------------------------------------
wait_idle
	xor a
	out (OUT_PORT),a
	ld a,5
	out (SPI_PORT),a
	nop
	nop
	in a,(SPI_PORT)
l_waiti
	nop
	nop
	in a,(SPI_PORT)
	bit 0,a
	jr nz,l_waiti
	ld a,3
	out (OUT_PORT),a
	ret


;
;---------------------------------------------------------------------------------------------
; This routine programs 256 bytes into the M25P16 spi flash chip.
; B  = one of 256 possible 64K sector containing the 256 bytes to be programmed (16MBytes,
;      but only 2MB available on this chip)
;
; HL = 0-FFFF address into selected sector and C000-FFFF memory read pointer for data buffer.
;      for proper operation, L should be zero.
;
;
; Returns with H = H + 1 (256 more bytes programmed)
;---------------------------------------------------------------------------------------------
page_program
	xor a
	out (OUT_PORT),a

	ld a,2			;page program command
	out (SPI_PORT),a

	ld a,b
	nop
	out (SPI_PORT),a	;MSB address
	ld a,h
	nop
	out (SPI_PORT),a
	xor a
	nop
	out (SPI_PORT),a

	ld a,h			;DE = HL | C000 (16K memory pointer)
	or $C0
	ld d,a
	ld e,l
l_prog
	ld a,(de)
	out (SPI_PORT),a
	inc de
	ld a,e
	cp 0
	jr nz,l_prog
	ld a,3
	out (OUT_PORT),a
	inc h			;next page
	ret
switch_end



;
; Routine comune per la ricezione di un blocco ROM. Puo` ricevere fino a 64Kbytes.
; Chiamare con:
; BC = nbytes (0 = 65536)
; C' = 0 (flag di errore)
;
; Modificata per FPGA: scrive in $C000 utilizzando indirizzamento FASTPAGE. Il chiamante deve occuparsi di selezionare
; il banco di partenza in fastpage stesso, oltre a settare il bit 'fastpage_C000' che provoca lo spostamento del
; fastpaging in C000 invece di 0.
; Provvede ad incrementare il puntatore al banco in caso di rollover, durante la ricezione.
;
; In caso di errore, rientra con CARRY SET dopo aver trasmesso '#' all'host
; altrimenti trasmette '*' e rientra con carry reset.
;
; Distrugge AF, BC, HL, DE, IY
;
;
common_getrom
	ld d,0			; DE usato per il calcolo del checksum
	ld iy,0			; contatore di checksum
	ld hl,$C000		; indirizzo di scrittura
l_ric_rom
	in a,(STATUS)		;!!! perche` non chiama rs_rx?
	and 7
	cp 2
	jr c,okrxrom
	exx
	ld c,1			; flag di errore
	exx
okrxrom
        bit 0,a			; 8T ATTESA PRESENZA DI UN BYTE
        jr z,l_ric_rom		; 7T
	in a,(RXREG)		; 11T
	ld (hl),a		; salva il byte
	ld e,a
	add iy,de		; somma al checksum
	inc hl
	ld a,l
	or h
	jr nz,no_bankrollover
	push bc
	ld bc,FASTPAGE		; registro di RAM BANKING
	in a,(c)
	inc a
	or $C0			; ensures no rollover on D7:D6 (write enable, read enable)
	out (c),a
	pop bc
	ld h,$C0		; beginning of new ram bank
no_bankrollover
	dec bc
	ld a,b
	or c
	jr nz,l_ric_rom

	call rs_rx
	cp iyl			; confronta LSB
	jr z,rtr_ok1_rom
	exx
	ld c,1			; flag di errore
	exx
rtr_ok1_rom
	call rs_rx
	cp iyh			; confronta la parte MSB
	jr nz,cm_rom_error	; errore

	exx
	ld a,c			; recupera il flag di errore da C'
	exx
	cp 0
	jr z,ok_rom1

cm_rom_error
	ld a,'#'		; codice di errore
	call rs_tx
	scf
	ret

ok_rom1				; possiamo procedere
	ld a,'*'		; codice OK
	call rs_tx
	and a
	ret


;
;---------------------------------------------------------------------------------------------
; This subroutine sends (or receives) a 16K bytes block, accessing memory through FASTPAGE
; register. Up to 1 MegaByte (512KB ROM and 512KB RAM) can then be accessed.
;
; FASTPAGE, CLKREG and CONTEXT SWITCH are altered but restored on exit.
; However, REGISTER ARE UNLOCKED and lock is NOT restored.
;---------------------------------------------------------------------------------------------
nof
	cp 'I'
        jr nz,noi

	call rs_rx	; receives the byte to be written to the FASTPAGE register
	ld e,a		; E = desired bank
	and $E0		; at least one of these three bits should be HIGH
	cp 0
	jr nz,valid_bank
	ld a,'!'	; error code
	call rs_tx
	jp waitcmd

valid_bank
	bit 7,a
	jr z,no_ramwrite
	ld a,'#'	; receiving a ram block is currently not supported
	call rs_tx
	jp waitcmd

no_ramwrite
	ld a,'*'
	call rs_tx	; OK

        call SetFastpC000       ; stack and fastpaging registers are saved
	ld sp,SHADOW_STACK	; then moves it to a safe place (we don't know what we interrupted and will switch C000 bank)

	ld bc,FASTPAGE
	out (c),e		; new content (bank to be sent)

m_entry
	ld d,0			; DE is used for checksum computation
	ld iy,0			; IY is the checksum register
	ld hl,$C000		; address pointer

l_sendpage
	ld a,(hl)
	call rs_tx
	ld e,a
	add iy,de		; checksum
	inc hl
	ld a,h
	or l
	jr nz,l_sendpage

        call StopFastpC000

	ld a,iyl		; LSB checksum
	call rs_tx
	ld a,iyh		; MSB checksum
	call rs_tx

	ld hl,(real_sp)		; restores stack pointer
	ld sp,hl
	jp waitcmd		; end of job

;
;---------------------------------------------------------------------------------------------
; This subroutine sends a 16K SPI block, using DDR as a buffer.
;---------------------------------------------------------------------------------------------
noi
	cp 'M'
        jr nz,nom

        call SetFastpC000       ; stack and fastpaging registers are saved
        ld sp,SHADOW_STACK      ; moves it to a safe place

        ld bc,FASTPAGE
	ld a,TMP_FASTPAGE
	out (c),a		; new content

	call rs_rx		; A = desired bank (16K offset)

        ld ix,tmp_reg

        ld c,0
        srl a                   ; n. banco e` da 16K
        rr c
	srl a
        rr c
        ld (ix+1),a             ; A23:A16 indirizzo SPI FLASH di lettura
        ld (ix+2),c             ; A15:A14
	xor a
	ld (ix+3),a
	ld a,TMP_DDRMSB
	ld (ix+4),a		; IX+3 = ddr_msb da usare
	ld a,TMP_FASTPAGE
	ld (ix+5),a		; IX+4 = fastpage da usare

	ld a,'*'
	call rs_tx	; OK
w_txe	in a,(STATUS)
	bit 3,a			; attesa TRANSMITTER IDLE prima di usare la SPI
	jr nz,w_txe

	ld hl,$C000		;destination write pointer
	ld bc,16384		;bytes to copy
	ld e,0			;E = 0 = READ; 1 = VERIFY
        call spi_16Kread        ;legge 16K in DDR
	jp m_entry


nom
nomorecmd
	ld a,'?'	; fine comandi disponibili: trasmette un "?"
	call rs_tx
        jp waitcmd


rx_error
	in a,(STATUS)
	res 0,a
	set 7,a		; codice di errore
	call rs_tx	; trasmette il codice di errore
	in a,(RXREG)	; toglie il byte ricevuto (e lo scarta)
	ld a,'#'	; trasmette '#' (fuori sincrono)
	call rs_tx
        jp waitcmd      ; rientra in loop


;
;---------------------------------------------------------------------------------------------
; Subroutine che riceve i dati per il trasferimento. Per primo il RAM_BANK, poi address e len.
;
; RAM_BANK puo` essere 0 - 7 (seleziona il relativo RAM_BANK 0 - 7).
; Rientra con:
;
; HL = indirizzo del blocco
; BC = lunghezza del blocco
;  D = ram_bank attuale, da ripristinare alla fine del trasferimento
;  E = ram_bank richiesto, da impostare subito prima dell'inizio del trasferimento
;  A = primo byte ricevuto come parametro (ram_bank) non modificato
;
; NON commuta il banco per non rischiare di farsi sparire lo STACK da sotto i piedi, cosa che
; potrebbe creare qualche problema al rientro :-)
;
; Distrugge AF, BC, DE, HL
;---------------------------------------------------------------------------------------------
rx_dblock
	ld a,'p'
	call rs_tx	; clear to receive parameters

	ld bc,S128KREG
	in a,(c)
	ld d,a		; salva lo stato attuale del registro RAM_BANK

	call rs_rx	; riceve il primo byte di dati
	push af
	jr c,rx_der
	and 7
	ld e,a		; salva in E (RAM_BANK richiesto)
	ld a,d
	and $F8		; maschera i bit di competenza RAM_BANK
	or e		; setta quelli richiesti
	ld e,a		; E = byte pronto da scrivere nel registro S128KREG

	call rs_rx	; riceve un byte
	jr c,rx_der
	ld l,a
	call rs_rx
	jr c,rx_der
	ld h,a		; HL = puntatore al blocco
	call rs_rx
	jr c,rx_der
	ld c,a
	call rs_rx
	jr c,rx_der
	ld b,a		; BC = lunghezza blocco
rx_der
	pop af
	ret


;
; Ricezione di 1 byte con attesa arrivo. Distrugge AF. Rientra con CARRY SET in caso di errore.
; 1 ciclo con byte gia` pronto = 81T
;
rs_rx	in a,(STATUS)		; 11T
	and 7
	cp 2
	jr nc,errores
        bit 0,a			; 8T ATTESA PRESENZA DI UN BYTE
        jr z,rs_rx		; 7T
	in a,(RXREG)		; 11T
	and a			; 4T rientro OK: carry reset
	ret			; 10T
errores	scf			; set carry: errore
	ret			; notare che rientra SENZA leggere il byte, lasciando l'errore in STATUS


;
; Trasmissione di 1 byte con attesa TX IDLE. Distrugge F.
;
rs_tx	push af
w_tx	in a,(STATUS)
	bit 3,a			; attesa TRANSMITTER IDLE
	jr nz,w_tx
	pop af
	out (TXREG),a
	ret

;
;------------------------------------------------------------------------------
; This subroutine enables $C000 FASTPAGE after saving register's content
;------------------------------------------------------------------------------
SetFastpC000
        ld hl,2
        add hl,sp               ; SP in HL (without THIS call return address)
	ld (real_sp),hl		; saves to a shadow_rom variable

	ld bc,UNLOCKREG		; unlocks both $7FFD + $1FFD and all non-standard registers
	ld a,$AA
	out (c),a

	ld bc,SPECTR_MODE	; enables $C000 mode for fastpaging
	in a,(c)
        ld (spectrm_save),a     ; previous value is saved
	set 7,a
	out (c),a

	ld bc,FASTPAGE
        in a,(c)
        ld (fastp_save),a

        ld bc,DDR_MSB
        in a,(c)
        ld (ddrmsb_save),a
        ret


;
;------------------------------------------------------------------------------
; This subroutine disabled $C000 fastpage mode and restores registers
;------------------------------------------------------------------------------
StopFastpC000
	ld bc,SPECTR_MODE	; enables $C000 mode for fastpaging
        ld a,(spectrm_save)
	out (c),a

	ld bc,FASTPAGE
        ld a,(fastp_save)
	out (c),a

        ld bc,DDR_MSB
        ld a,(ddrmsb_save)
	out (c),a
        ret

;
;********************************************************************************************************
;********************************************************************************************************
;*													*
;* Routine di POWER-ON. Seleziona una ROM ed esegue un salto $0, oppure ripristina uno Snapshot da MMC.	*
;* Se lanciata da shadow_rom (banco 1), presenta una ulteriore opzione per saltare nella rom caricata	*
;* sul banco 0 della shadow_rom stessa, lasciandola quindi attiva in lettura.				*
;*													*
;* NOTA: Per poter funzionare in SHADOW_ROM, una ROM del basic deve trovarsi su di un BANCO PARI,	*
;*	 ovvero con LSB ROM_BANK = 0 (D4 $7FFD). Questo perche` la nostra ROM deve per forza essere	*
;*	 sul banco con LSB ROM_BANK = 1. La SHADOW_ROM dispone di soli 2 banchi, selezionati da quel	*
;*	 bit. Questo comporta anche l'impossibilita` di gestire in SHADOW_ROM la coppia di rom dello	*
;*	 Spectrum 128K e contemporaneamente la nostra ROM. Per le prove del firmware occorrera`		*
;*	 necessariamente appoggiarsi alla ROM del BASIC 48K, posta su di un ROM_BANK PARI.		*
;*													*
;********************************************************************************************************
;********************************************************************************************************
reset_vector
        xor a                   ; this flag makes the difference between
        ld (nmi_count),a        ; cold and warm reset

warm_reset:
	xor a
	out (NMIREG),a		; disables NMI from UART
	in a,(RXREG)		; clears UART RX register

	ld sp,STACK		; STACK in nonbanked area

	call set_cpldregs	; CPLD registers setup

        ld a,(nmi_count)        ; if nmi_count > 0, it was a warm reset
        ld e,a                  ; set_vars does not alter E register
	call set_vars		; set variables. in order to access the last KB in RD/WR mode,
				; this subroutine will ACTIVATE THE CONTEXT SWITCH MODE anyway.

        ld a,e                  ; if NMI_COUNT > 0, then we came here by double
        cp 0                    ; NMI (quick reset): no DDR init is needed
        jr nz,no_ddrinit
	call spi2ddr_copy	; all the 16K roms should be copied from SPI FLASH to DDR memory
no_ddrinit

	call joystick_reset	; cleans any programmed data in the joystick controller

        call cls                ; clears the screen

	call residos_detect	; sets the ResiDOS presence bit. to be called after battery low detection.
				; it also initialize RESIDOS_SPEED to a default value


	in a,(RXREG)		; Inizializzazione UART: vuota il registro di ricezione dopo aver visualizzato
	ld a,1			; varie cose, in modo da dare il tempo all'uart di concludere eventuale ricezione spuria
	out (NMIREG),a		; abilita il pin DATA_AVAILABLE, che generera` un NMI alla ricezione di un byte

        call UlaPlus_Colors     ; inizializza la tavolozza colori ULA+

	ld a,6			; D0 LOW = MMC0 selected
	ld (cs_outreg),a

	call sd_check		; controlla eventuale presenza di SD/MMC
	ld a,(sh_flags)
	bit sd_initsnap,a
	jr nz,card_okforus

	push af
	ld a,5			; D1 LOW = MMC1 selected
	ld (cs_outreg),a
	call sd_check		; controlla eventuale presenza di SD/MMC
	pop bc

	ld a,(sh_flags)		; la seconda card e' stata trovata: salta
	bit sd_present,a
	jr nz,card_okforus
	ld a,6			; altrimenti riseleziona la prima per default: non si sa mai che
	ld (cs_outreg),a	; sia l'unica presente
	ld a,b
	ld (sh_flags),a		; ripristina lo stato di sd_present riguardante la prima card

card_okforus
	call disp_menu		; displays the main menu

	ld a,(sh_flags)
	bit sd_present,a
	jr z,no_mmc		; salta se non l'ha trovata

	bit sd_initsnap,a
	jr z,no_mmc

	ld hl,0
	ld (sd_entry),hl	; numero di snapshot per visualizzazione

	call display_mmcdata

no_mmc
	ld a,5
	ld (start_item),a	; oggetto con cubetto lampeggiante lanciato per default alla pressione di ENTER

	ld hl,flags
	set save_attrib,(hl)	; alla prima chiamata, SHOWSEL_ITEM salvera` l'attributo prima di sovrascriverlo


; RTC chip time/date registers are updated on falling edge of the 1Hz signal.
; At poweron, the PIC gets the entire set of registers and keeps an internal copy,
; which is updated every second on the falling edge of the 1Hz signal.
;
; At midnight, the RTC chip is read again by PIC firmware in order to properly update the DATE.
;
; When the RTC chip is first read at power on, the flag used to detect the falling
; edge of the 1Hz signal is reset. This avoids an increment of PIC registers
; immediately after RTC is read. The PIC registers, hence, maybe 1 second LATE but
; not EARLY. This is fine, since next midnight's read won't be performed too early
; (causing a double read, one at RTC's 23:59:59 and the other at the real 0:0:0)


	ld hl,rtc_data		; buffer for local date/time storage
	call rtc_enquire	; no rtc chip: date just initialized

	ld a,$AA		; signals to the ZX-Com program that we LOGGED OUT
	call rs_tx

waitkey_countdown
	xor a
	ld (frames+3),a
	
;---------------------------------------------------------------------------------------------
; MAIN WAIT-FOR-KEYPRESS LOOP
;---------------------------------------------------------------------------------------------
wait_key
        ld hl,frames
        inc (hl)		; incrementa byte LSB prescaler
        jr nz,wait_key
        inc hl
        inc (hl)		; incrementa byte MSB prescaler
        jr nz,wait_key
        ld a,$E8                ; che viene pero` reinizializzato a $E8. Quando questo conteggio va` in overflow,
	ld (hl),a		; si ha un ciclo di lettura KEYBOARD
        inc hl
        inc (hl)		; incrementa byte LSB contatore
	jr nz,nomsbinc
	inc hl
	inc (hl)		; incrementa byte MSB contatore (10 secondi/incremento @3.5MHz)
nomsbinc
        ld a,(hl)		; cambia colore ad ogni ciclo
	call showsel_item	; carica attributo A alla voce indicata da START_ITEM

	call gest_rewrite	; eventuale gestione riscrittura di FIRSTFREE in SD/MMC

;	call inc_rtc		; gestione incremento DATE/TIME basato sul segnale a 1Hz dal chip RTC
				; se reload_rtc in sh_flags e` settato, il chip viene riletto.


	ld a,(sh_flags)		; if an NMI call corrupted the screen content, we redraw it now
	bit redraw_screen,a
	jr z,qredraw
	res redraw_screen,a
	ld (sh_flags),a
	call restore_screen
qredraw

	call key_scan		; lettura tastiera
	jr nz,wait_key		; piu` di 2 tasti: ignora
	ld a,e
	cp $ff			; nessun tasto premuto?
	jr nz,something
	in a,(KEMPSTON)		; un tasto del joystick e` attivo?
	and $1F
	jr z,no_joystick
	call kempston_code	; carica in A un ASCII corrispondente ad uno dei 5 possibili tasti
	jr kempston_too

no_joystick
	call g_nokeys		; gestione reset flags tasto premuto per eventuali autorepeat
	jr wait_key		; salvare il prossimo snapshot

something
	call key_number		; restituisce il numero del tasto premuto (1 - 40)
	cp 9			; tasti da 1 a 9 = selezione di uno snapshot
	jr nc,no_mmcread
	dec a
jp_snapshot			; entry point per pressione ENTER, A contiene n. snapshot 0-based
	ld hl,sh_flags
	bit sd_initsnap,(hl)	; is the card initialized for snapshots?
	jr z,wait_key		; no, ignores the command
	ld e,a
	ld d,0
	ld hl,(sd_entry)	; base n. entry visualizzata nel menu
	add hl,de		; HL = numero di snapshot da ripristinare
	call remove_residos
        jp read_snap

no_mmcread
	call key_decode		; restituisce il codice ASCII del tasto premuto

	ld l,0			; default ROM_BANK ZERO
        cp 'z'
        jr z,r_loadthis
        inc l                   ; banco 1
	cp 'x'
        jr z,r_loadthis
        inc l                   ; banco 2
        cp 'c'
        jr z,r_loadthis
	inc l			; steps over ROM3 (bootrom)
        inc l
        cp 'v'                  ; banco 4
	jr z,r_loadthis

        ld l,8                  ; banco 8
        cp 'b'
        jr z,r_loadthis

	ld l,$FE		; ResiDOS
	cp 'r'
	jr z,r_loadthis


kempston_too
	cp 13			; ENTER attiva l'opzione selezionata dal cubetto lampeggiante
	jr nz,noenter
start_the_item
	ld a,(start_item)
	cp 5
	jr nc,no_rom		; salta se non e` selezionata una ROM
	ld l,a
	cp 0
	jr z,r_loadthis		; la rom 0 corrisponde in effetti al valore 00 in start_item
	cp 1			; dato che sulla prima riga sono presenti 3 rom, occorre saltarle
	jr nz,no_plus3e		; (salta anche la 03 che e` la BOOTROM, per cui 01 = rom 4)
	ld l,4
	jr r_loadthis
no_plus3e
	ld l,$FE


r_loadthis
	jp loadthis
no_rom
	sub 5
	jr jp_snapshot


noenter
	ld hl,sh_flags
	bit sd_present,(hl)
	jp z,no_cardpresent	; se la SD/MMC non e` stata trovata, salta diverse opzioni
	bit sd_initsnap,(hl)
	jp z,no_oursdcard	; se la SD/MMC non e` inizializzata per snapshot, salta diverse opzioni

	cp 'e'			; EDIT filename
	jp nz,noedit
	call wait_keyrelease	; waits for the key to be released

	call get_hlentry	; loads the entry from SD/MMC and sets HL to point to the correct item
	jp nz,wait_key		; returns NONZERO if entry is invalid

	ld a,(start_item)
	add 5
	ld (y_pos),a
	inc hl
	push hl			; address of the ascii text in TEMP_RAM (sd/mmc scratch area)
	call make_blank
	ld hl,tmp_str
	ld (hl),$0F		; attribute for edit space
	ld b,32			; available space
	inc hl
l_edit
	push hl
	push bc
	ld a,'_'
	ld (hl),a
	ld hl,tmp_str
	ld a,51
	ld (x_pos),a
	call disp_text
l_key
	call key_scan
	jr nz,l_key
	call key_decode		; codice ASCII
	pop bc
	pop hl
	cp 13
	jr z,end_edit
	cp 8			; symbol shift = DELETE (cambiare con codice 8)
	jr nz,nodelete
	ld a,b
	cp 32
	jr z,l_edit		; niente da cancellare
	ld (hl),' '		; toglie il cursore
	dec hl
	inc b
	inc b
	jr next_key
nodelete
	cp ' '
	jr c,l_edit		; non accetta nessun altro codice inferiore a SPAZIO
	ld (hl),a
	inc hl
next_key
	call wait_keyrelease	; waits for the key to be released

	djnz l_edit
end_edit
	ld a,b
	cp 0
	jr z,notrailing
	ld (hl),' '		; se la stringa non e` piena, elimina il codice del cursore
notrailing
	pop de			; address of string in SD/MMC scratchpad area
	ld hl,tmp_str+1
	ld bc,32
	ldir
	call put_entry
	call display_entries
        jp wait_key


noedit
	cp 'w'			; EDIT Parameters of snapshot
	jp nz,noparameters
	call wait_keyrelease	; waits for the key to be released
	call get_hlentry	; loads the entry from SD/MMC and sets HL to point to the correct item
	jp nz,wait_key		; returns NONZERO if entry is invalid

	inc hl
	push hl
	ld a,7
	ld (attribute),a
	call cls
        ld a,48
	ld (x_pos),a
	xor a
	ld (y_pos),a
	ld hl,logo		; stringa asciiz LOGO di sistema
	call disp_text

	ld hl,istruz1
	xor a
	ld (x_pos),a
	ld a,19
	ld (y_pos),a
	call disp_text

        ld hl,istruz1b
	xor a
	ld (x_pos),a
        ld a,20
	ld (y_pos),a
        call disp_text

	ld hl,istruz2
	xor a
	ld (x_pos),a
	ld a,21
	ld (y_pos),a
	call disp_text

	ld hl,istruz3
	ld a,45
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
	call disp_text

	pop hl			; visualizza il nome dello snapshot
	push hl
	ld de,tmp_str+1
	ld bc,32
	ldir
	xor a
	ld (x_pos),a
        ld a,2
	ld (y_pos),a
	ld hl,tmp_str
	ld (hl),7
	call disp_text

	pop iy			; indirizzo blocco dati
	ld de,39		; offset al byte LSB dello SP
	add iy,de
refresh_param
	call disp_param		; visualizza i parametri di questo snapshot

l_paramkey
	call key_scan
	jr nz,l_paramkey
	call key_number		; A = numero del tasto premuto
	cp 0
	jr z,l_paramkey
	cp 1
	jr nz,nofreqch
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+5)
	inc a			; next frequency step on bits D2:D0
	and 7
	ld c,a
	ld a,(iy+5)
	and $F8
	or c
	ld (iy+5),a
	call wait_keyrelease
	jr refresh_param


nofreqch
	cp 2
	jr nz,noimch
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+4)		; F BYTE
	inc a			; IM MODE +1
	ld c,a
	and 3			; IM MODE 3 non esiste
	cp 3
	jr nz,no_i3
	res 0,c
	res 1,c
no_i3
	ld (iy+4),c
	call wait_keyrelease
	jr refresh_param

noimch
	cp 3
	jr nz,noint_toggle
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+4)
	xor 4			; bit abilitazione INT
	ld (iy+4),a
	call wait_keyrelease
        jp refresh_param

noint_toggle
        cp 4
        jr nz,noula_toggle
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+4)
        xor 128                 ; bit abilitazione ULA+
	ld (iy+4),a
	call wait_keyrelease
        jp refresh_param

noula_toggle
        cp 32                   ; 'Z'
	jr nz,nospectrum_toggle
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+4)
	xor 8			; bit SPECTRUM MODE
	ld (iy+4),a
	call wait_keyrelease
	jp refresh_param

nospectrum_toggle
	cp 33			; 'X'
	jr nz,no_doubleinttoggle
	ld hl,flags
	set update_mmc,(hl)
	ld a,(iy+4)
	xor 16			; bit DOUBLE_INT mode
	ld (iy+4),a
	call wait_keyrelease
	jp refresh_param

no_doubleinttoggle
	cp 30			; ENTER
	jp nz,l_paramkey	; fine tasti decodificati

	ld hl,flags
	bit update_mmc,(hl)
	jr z,no_upmmc1
	res update_mmc,(hl)
	call put_entry		; salva in SD/MMC
no_upmmc1
	call restore_screen
	jp wait_key

istruz1
	db	4,"[1] Z80 Clock; [2] IM MODE; [3] INT On/Off",0
istruz1b
        db      4,"[4] ULA+ On/Off",0
istruz2
	db	4,"[Z] Toggle Spectrum Type; [X] INT Doubler",0
istruz3
	db	2,"Press [ENTER] to Terminate",0


restore_screen
	ld a,7
	ld (attribute),a
	call cls
	call disp_menu
	call display_mmcdata
	ret

noparameters
	cp 'q'			; EDIT Joystick mapping
	jp nz,nojoystick
	call wait_keyrelease	; waits for the key to be released
	call get_hlentry	; loads the entry from SD/MMC and sets HL to point to the correct item
	jp nz,wait_key		; returns NONZERO if entry is invalid

	ld de,46		; HL points at snapshot size byte (+2); joystick buffer starts at offset +48
	add hl,de		; HL = 10 bytes joystick programming data
	ld (joystick_pnt),hl

	ld a,7
	ld (attribute),a
	call cls
	xor a
	ld (y_pos),a
	ld a,30
	ld (x_pos),a
	ld hl,joy1		; stringa asciiz nome pagina
	call disp_text

	xor a
	ld (x_pos),a
        ld a,17
	ld (y_pos),a
	ld hl,joy2
	call disp_text
	xor a
	ld (x_pos),a
        ld a,18
	ld (y_pos),a
	ld hl,joy3
	call disp_text
	ld a,24
	ld (x_pos),a
        ld a,20
	ld (y_pos),a
	ld hl,joy4
	call disp_text
	ld a,39
	ld (x_pos),a
        ld a,22
	ld (y_pos),a
	ld hl,joy5
	call disp_text
        ld a,24
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
        ld hl,joy6
	call disp_text

	xor a
	ld (joystick_item),a	; input being programmed
	call disp_joymenu	; displays the joystick menu

disp_assign
        call joy_mapped         ; displays current key assignment from
        call wait_keyrelease    ; ((JOYSTICK_PNT)) buffer

wait_jrelease
        ld de,$8000             ; initial autorepeat timeout
	ld hl,flags
	bit key_pressed,(hl)
	jr z,wait_r
        ld de,$2000             ; fast autorepeat timeout
wait_r
	set key_pressed,(hl)
wait_rep
	dec de
	ld a,d
	or e
	jr z,wait_something
	ld bc,10
wt_ime
	in a,(KEMPSTON)
        and $1F
	jr nz,wait_rep
        ld de,$8000             ; initial autorepeat timeout
	djnz wt_ime		; debounce
	dec c
	jr nz,wt_ime
end_rep
	ld hl,flags
	res key_pressed,(hl)

wait_something
	in a,(KEMPSTON)
	bit 2,a			; DOWN
	jr z,no_jdown
	ld a,(joystick_item)
	cp 9
        jr z,end_rep
	inc a
	ld (joystick_item),a
	call disp_joymenu
	jr wait_jrelease

no_jdown
	bit 3,a			; UP
	jr z,no_jup
	ld a,(joystick_item)
	cp 0
        jr z,end_rep
	dec a
	ld (joystick_item),a
	call disp_joymenu
	jr wait_jrelease

no_jup
	bit 4,a
	jr z,no_jfire		; the fire button erases the key
	ld c,0
	jr entry_erasekey

no_jfire
        bit 1,a                 ; left button quits
        jr nz,no_upmmc2
	bit 0,a
	jr z,no_right		; the right button exits the routine
	ld hl,flags
	bit update_mmc,(hl)
	jr z,no_upmmc2
	call put_entry		; salva in SD/MMC
no_upmmc2
	res update_mmc,(hl)
	call restore_screen
	jp wait_key

no_right
	call key_scan		; reads the keyboard
	jr nz,wait_something	; more than 2 keys: ignore
	ld a,e
	cp $FF
	jr z,wait_something	; no keys pressed
        cp 39                   ; CAPS: tenuto premuto, esce senza salvare
        jr nz,no_jspace
        ld bc,$8000             ; 1,5 secondi a 14MHz
l_wspace
        push bc
        call key_scan
        pop bc
        ld a,e
        cp $FF
        jr z,jreleased          ; rilasciato: occorre programmarlo
        dec bc
        ld a,b
        or c
        jr nz,l_wspace
        jr no_upmmc2            ; CAPS premuto a lungo: esce senza salvare
jreleased
        ld e,39                 ; ricarica il codice del CAPS
no_jspace
	call key_joycode	; transforms in ADDRESS/DATA line number for joystick programming
	ld c,a
entry_erasekey
	ld hl,flags
	set update_mmc,(hl)
	ld hl,(joystick_pnt)	; base address of 10-bytes joystick key assignment
	ld a,(joystick_item)	; item being edited
	ld e,a
	ld d,0
	add hl,de
	ld (hl),c		; saves the new assignment to the buffer
	jp disp_assign

joy1
	db	7,"Programmable JoyStick Setup Page",0
joy2
	db	1,"Use JoyStick UP/DOWN to Select an Item,",0
joy3
	db	1,"then press a Key to map it to active Item.",0
joy4
	db	1,"JoyStick FIRE to delete a Key Item",0
joy5
	db	2,"JoyStick RIGHT to Save & Exit",0
joy6
        db      2,"JoyStick LEFT or hold CAPS to Quit",0
joyn
	db	7,"IN 0",0
joydx
	db	7,"(Right)",0
	db	7,"(Left) ",0
	db	7,"(Down) ",0
	db	7,"(Up)   ",0
	db	7,"(Fire) ",0

;
; Joystick menu
;

disp_joymenu
	ld a,4
	ld (y_pos),a		; display position for first row

	ld hl,joyn
	ld de,tmp_str
	ld bc,6
	ldir
	ld b,10			; items to display
j_newline
	xor a
	ld (x_pos),a
	ld a,7
	ld (tmp_str),a

	ld a,10+'0'
	sub b
	ld (tmp_str+4),a
	sub '0'
	ld l,a
	ld a,(joystick_item)
	cp l
	jr nz,no_evi
	ld a,$57
	ld (tmp_str),a
no_evi
	ld hl,tmp_str
	push bc
	call disp_text
	ld a,(tmp_str+4)
	cp '5'
	jr nc,nospiega
	ld hl,joydx
	ld de,9
ce_stri
	cp '0'
	jr z,bleble
	add hl,de
	dec a
	jr ce_stri
bleble
	ld a,32
	ld (x_pos),a
	call disp_text

nospiega
	ld a,(y_pos)
	inc a
	ld (y_pos),a
	pop bc
	djnz j_newline
	ret

;
; Joystick mapped keys display
;
joy_mapped
	ld a,4
	ld (y_pos),a
	ld b,10
	ld hl,(joystick_pnt)	; base address of 10-bytes joystick key assignment
disp_keyval
	push bc
	push hl
	call make_15blank
	ld a,100
	ld (x_pos),a
	ld e,(hl)		; E = JOYSTICK ADDRESS/DATALINE for current input
	call joy_keycode	; A = KEYSCAN which corresponds to the key map data
	ld e,a			; simulates a real keyscan
	ld d,0			; no caps or symbol shift
	call key_decode		; gets the ascii value of the programmed key
	ld hl,tmp_str
	cp 0
	jr z,no_spacekey	; leaves blank if no valid key is assigned
	ld (tmp_str+1),a
	cp 13
	jr nz,no_enterkey
	ld hl,enter_keyname
no_enterkey
	cp 1
	jr nz,no_capskey
	ld hl,caps_keyname
no_capskey
	cp 2
	jr nz,no_symbolkey
	ld hl,symbol_keyname
no_symbolkey
	cp 32
	jr nz,no_spacekey
	ld hl,space_keyname
no_spacekey
	call disp_text
	pop hl
	pop bc
	inc hl
	ld a,(y_pos)
	inc a
	ld (y_pos),a
	djnz disp_keyval
	ret

make_15blank
	push bc
	push hl
	ld a,7
	ld hl,tmp_str
	ld (hl),a
	inc hl
	ld a,' '
	ld b,15
l_makeblank
	ld (hl),a
	inc hl
	djnz l_makeblank
	xor a
	ld (hl),a
	pop hl
	pop bc
	ret

enter_keyname
	db	7,"[ENTER]       ",0
caps_keyname
	db	7,"[CAPS Shift]  ",0
symbol_keyname
	db	7,"[SYMBOL Shift]",0
space_keyname
	db	7,"[SPACE]       ",0


nojoystick
        cp 'u'                  ; backup entire zxmmc+ RAM content into four snapshot slots (512K)
        jp nz,noramrombackup
	xor a
	ld (flash_bank),a	; source: RAM. utilizza flash_bank
        ld hl,backup_confirm
;	jr do_ramrombackup
;norambackup
;	cp 'i'			; backup entire zxmmc+ FLASHROM content
;	jp nz,noramrombackup
;	ld a,1
;	ld (flash_bank),a	; source: ROM
;	ld hl,backup_confirmrom

do_ramrombackup
	call yesno_confirm
	jp nc,wait_key

        xor a
        ld (flash_bank),a
        ld (x_pos),a
        ld a,2
        ld (y_pos),a
        ld hl,backup_inprogress
        call disp_text

        call make_512kbackup
	jr c,no_backup

        ld a,5
        ld (y_pos),a
        xor a
        ld (x_pos),a
        out (BORDER),a
        ld hl,back_done
        call disp_text
anykey
        call key_scan
        ld a,e
        cp $ff                          ; no keys
        jr z,anykey

no_backup
        call cls
        call disp_menu
        call display_mmcdata
        call wait_keyrelease            ; waits for the key to be released
        jp wait_key


backup_confirm
        db      7,"Confirm 512K Ram Backup into sd-card (Y/N)",0
;backup_confirmrom
;        db      7,"Confirm 512K FlashBackup into sdcard (Y/N)",0
backup_inprogress
        db      7,"Backup in progress. Please wait.",0
back_done
        db      6,"Backup Successful. Press any key.",0


noramrombackup                  ; fine comandi specifici per sd-card presente ed inizializzata
	jr common_keys



no_oursdcard			; COMMANDS AVAILABLE WHEN THE CARD IS NOT INITIALIZED FOR SNAPSHOTS
	cp 'q'			; Q = inizializza la card per l'uso con gli snapshot
	jr nz,common_keys
	call gest_cardinit	; initializes the card
	jp wait_key


no_cardpresent
common_keys
;
; Following KEYS are subject to AUTOREPEAT function
;
	ld c,a			; salva il codice del tasto
	ld a,(repeat_timer)
	cp 0
	jr z,nomz_pro		; autorepeat scaduto: procede al rilevamento della pressione dei tasti
	dec a
	ld (repeat_timer),a	; altrimenti decrementa semplicemente la variabile
	ld a,c
	jp cha_wrnum		; e salta la gestione di queste funzioni


nomz_pro
	ld a,2
	ld (repeat_timer),a	; velocita` di autoripetizione
	ld hl,flags		; gestione ritardo/autorepeat per i comandi a seguire
	bit key_pressed,(hl)
	jr nz,no_moredelay
        ld a,20                 ; ritardo iniziale prima dello scatto dell'autorepeat
	ld (repeat_timer),a
	set key_pressed,(hl)
no_moredelay
	ld a,c			; recupera il codice del tasto

	ld hl,sh_flags		; se non c'e' una card con snapshots, salta le opzioni avanti/indietro veloci
	bit sd_initsnap,(hl)
	jr z,nomz

	ld de,8
	cp '9'			; 9 = indietro di un blocco da 8 entries
	jr z,indblock
	ld de,128
	cp 'n'
	jr nz,nolessentry
indblock
	ld hl,(sd_entry)
	and a
	sbc hl,de
        jr nc,nroll2
	ld hl,0
nroll2
	ld (sd_entry),hl
	call display_entries
	jp wait_key


nolessentry
	ld de,8
	cp '0'			; 0 = avanti di un blocco da 8 entries
	jr z,avablock
	ld de,128
	cp 'm'
	jr nz,nomz
avablock
	ld hl,(sd_entry)
	add hl,de
	push hl
	ld de,MMCSIZE
	and a
	sbc hl,de
	pop hl
	jr c,noroll
	ld hl,MMCSIZE-8
noroll
	ld (sd_entry),hl
        call display_entries
        jp wait_key


nomz
	cp 'p'			; P = avanti di 1 item
	jr nz,nomp
	ld hl,flags
	set restore_attrib,(hl)
	call showsel_item	; ripristina attributi normali sulla voce che sta per essere lasciata
	set save_attrib,(hl)	; alla prossima chiamata, SHOWSEL_ITEM salvera` l'attributo prima di sovrascriverlo
	ld a,(start_item)
	inc a
	ld hl,flags
	bit weare_shadow,(hl)	; stiamo girando in shadow_rom?
	jr nz,sh_selectable	; si, ne permette la selezione
	cp 2
	jr nz,sh_selectable
	inc a
sh_selectable
	bit residos_present,(hl)
	jr nz,r_selectable
	cp 3
	jr nz,r_selectable
	inc a			; se assente, salta l'opzione relativa al residos

r_selectable	
	cp 4
	jr nz,no3_ss
	ld hl,sh_flags
	bit sd_initsnap,(hl)
	jr nz,sdfound1
	ld a,3
	jr re_checkback		; utilizza la routine 'indietro 1 item' per stabilire una opzione presentabile
sdfound1
	ld a,5			; salta la parte vuota (notare che non puo` selezionare la SHADOW)
no3_ss
	cp 13			; siamo oltre il menu`?
	jr nz,no_omenu
	ld a,5
	ld (start_item),a
	ld de,8
        jp avablock             ; prosegue nella routine di avanzamento di un blocco
no_omenu
	ld (start_item),a
	ld hl,flags2
	set force_points,(hl)
	call disptimedate	; displays DATE and TIME for the (start_item-5) entry
	jp r_wait_key	


nomp
	cp 'o'			; O = indietro di 1 item
	jr nz,noop
	ld hl,flags
	set restore_attrib,(hl)
	call showsel_item	; ripristina attributi normali sulla voce che sta per essere lasciata
	set save_attrib,(hl)	; alla prossima chiamata, SHOWSEL_ITEM salvera` l'attributo prima di sovrascriverlo
	ld a,(start_item)
	cp 0
	jr z,r_wait_key
	dec a
re_checkback
	ld hl,flags
	cp 3			; opzione shadow_rom
	jr nz,nor_back
	bit residos_present,(hl)
	jr nz,nor_back
	dec a
nor_back
	cp 2			; opzione shadow_rom
	jr nz,proceed_back
	bit weare_shadow,(hl)
	jr nz,proceed_back
	dec a
proceed_back
	cp 4			; siamo oltre il menu`?
	jr nz,no_omenu2
	ld hl,(sd_entry)	; ci sono altri blocchi di entry su cui retrocedere?
	ld a,l
	or h
	jr nz,some_moreitems	; si, salta
	ld a,3			; altrimenti rientra nella zona della selezione di una ROM (3 = residos)
	jr re_checkback
some_moreitems
	ld a,12
	ld (start_item),a
	ld de,8
	jp indblock		; prosegue nella routine di indietreggiamento di un blocco
no_omenu2
	ld (start_item),a
	ld hl,flags2
	set force_points,(hl)
	call disptimedate	; displays DATE and TIME for the (start_item-5) entry
	jr r_wait_key	


noop

;
; Gestione pulsanti [s] e [ ] (symbol shift e spazio) per la gestione di incremento/decremento del numero di entry
; in cui il prossimo snapshot avra` luogo. E` basato su di un contatore di ritardo per l'aggiornamento della SD/MMC
; che produce anche un meccanismo di autorepeat.
;
; Utilizza un conteggio di autorepeat dedicato, che gestisce anche la riscrittura del dato aggiornato.
; Non aggiungere decodifica di altri tasti dopo di questi, perche` il conteggio di autorepeat interferirebbe.
;
cha_wrnum
	ld c,a
	ld a,(rewrite_timer)
        cp $30
        jr nc,r_wait_key        ; attesa autorepeat

	xor a			; impedisce uscita dal menu` per timeout
	ld (frames+3),a

	ld a,c

	cp 2			; Symbol Shift decrementa il numero di ENTRY per il prossimo snapshot
	jr nz,nosymb		; 2 = codice del symbol shift premuto da solo

	ld hl,(firstfree)	; ma solo se e` > 0
	ld a,l
	or h
        jr z,r_wait_key
	dec hl
loadfirstfree
	ld (firstfree),hl
	call disp_firstfree

        ld a,$40                ; ritardo autorepeat e timer riscrittura in SD/MMC
	ld hl,flags
	bit sp_pressed,(hl)	; era premuto anche prima?
	jr z,first_pressed	; alla prima pressione carica $40, alle successive $30 (autorepeat veloce)
        ld a,$30
first_pressed
	ld (rewrite_timer),a
	set sp_pressed,(hl)
r_wait_key
        jp wait_key

nosymb
	cp ' '			; spazio incrementa il numero di ENTRY per il prossimo snapshot
        jr nz,r_wait_key
	ld hl,(firstfree)
	inc hl			; incrementa il puntatore
	call check_hlspace	; verifica che non abbia raggiunto il limite
	jp nc,wait_key
	jr loadfirstfree


;
; Gestione nessun tasto premuto
;
g_nokeys
	ld hl,flags

	res key_pressed,(hl)	; gestione autorepeat generico
	xor a
	ld (repeat_timer),a

	bit sp_pressed,(hl)	; gestione autorepeat cambio entry in cui salvare il prossimo snapshot,
	ret z			; che ne gestisce anche la riscrittura in SD/MMC

	res sp_pressed,(hl)
	ld a,$30
	ld (rewrite_timer),a	; pronta risposta in caso venisse premuto di nuovo
	ret


;
;------------------------------------------------------------------------------------------------
; This subroutine saves the entire 512K zxmmc+ ram or rom content into 4 empty snapshot slots
;------------------------------------------------------------------------------------------------
make_512kbackup
        ld hl,BACKUPSTR         ; stringa nomefile primo slot utilizzato
;        ld c,MMC_RAMBACKUP      ; tipo di dati: backup ram
;	ld a,(flash_bank)	; 0 = RAM; 1 = FLASHROM
;	cp 0
;	jr z,backram
;	ld hl,BACKUPSTRROM
;        ld c,MMC_FLASHBACKUP    ; tipo di dati: backup flash
;backram
        ld (tmp_reg),hl
;        ld a,c
        ld a,MMC_RAMBACKUP      ; tipo di dati: backup ram
        ld (tmp_reg+2),a

        ld hl,(firstfree)       ; primo slot libero
        ld (tmp_reg+3),hl       ; serve piu` tardi per il calcolo della posizione del blocco dati

        call write_slotdata     ; lo occupa, assegnando il nomefile ed il tipo

        ld hl,RAWSTR            ; ora occupa i successivi 3 slot (512K in totale)
	ld de,tmp_str
	ld bc,32
	ldir
	ld hl,tmp_str		; usa tmp_str in modo da poter variare il numero nel nome del blocco
        ld (tmp_reg),hl
        ld a,MMC_RAWDATA        ; tipo di dati: RAW (non ripristinabile)
        ld (tmp_reg+2),a
        ld hl,(firstfree)       ; primo slot libero (non e` ancora aggiornato)
        ld b,3
loop_3slot
	ld a,'5'
	sub b
	ld (tmp_str+12),a	; scrive il numero di questo blocco nel nome del file
        inc hl
        push bc
        push hl
        call write_slotdata
        pop hl
        pop bc
        ret c                   ; rientra in caso di errore
        djnz loop_3slot
        inc hl
        ld (firstfree),hl       ; aggiorna il n. di slot libero (+4)
        call save_firstfree     ; lo salva nella card

;        ld hl,ramcopy           ; parte che deve essere eseguita in ram spectrum
;        ld bc,ramcopyend-ramcopy
;        ld de,EXECUTE_RAM
;        ldir

        call SetFastpC000       ; stack and fastpaging registers are saved

; Calcola l'indirizzo di scrittura in SD/MMC per questo snapshot

        ld hl,(tmp_reg+3)       ; numero del primo dei 4 slot occupati
        ex hl,de
        ld l,FATSIZE            ; n. blocchi da 64KB che costituiscono la FAT (e` l'offset per la word MSB)
        ld a,(card_offset)
        ld h,a                  ; offset area snapshot nella card
        sla e
        rl d                    ; moltiplica per 2 (ogni entry corrisponde ad un cluster da 128KB)
        add hl,de               ; HL = word MSB scrittura in SD/MMC
        ld de,0                 ; word LSB = 0 (insieme a D0 = 0 della word MSB = 128K)

        ld b,32                 ; blocchi da 16K da copiare

	ld c,$60		; RAM backup: diventa 0-31 con D6 set dopo la sottrazione di B (ram selected)
	ld a,(flash_bank)
	cp 0			; 0 = backup ram
	jr z,l_rambackup
	ld c,$80		; ROM backup: diventa 0-31 con D6+D5 set
l_rambackup
        push bc

        ld a,c
        sub b                   ; lo trasforma in numero 0-31 con solo D6 SET (ram paged in) o D6+D5 SET (rom paged in)
	ld bc,FASTPAGE
	out (c),a		; banco leggibile su $C000 - $FFFF
        and 7
        out (BORDER),a

        ld b,16384/BLOCKSIZE    ; dimensioni per uno snapshot da 16K

        ld ix,$C000             ; ram buffer address
        call mmc_write_data     ; takes the snapshot
        and a
        jr z,wr_okbackup
        ld a,2
        out (BORDER),a
        pop bc
        call StopFastpC000
	scf
        ret

wr_okbackup
        pop bc
        djnz l_rambackup        ; other banks? (HL/DE is incremented by mmc_write_data)

        call StopFastpC000
	and a
        ret

BACKUPSTR
        db      "512K Ram Backup                 "
;BACKUPSTRROM
;        db      "512K Flash Backup               "
RAWSTR
        db      "Backup part x of 4              "


;
; Routine di scrittura di una entry nella FAT
;
write_slotdata
        ex hl,de
        push de
        call calc_hldec         ; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C

        ld ix,SHADOW_RAM        ; scratchpad area
        call mmc_read_data      ; performs the single block read
        and a
        jr z,b512_1
b512_error_de
        pop de
b512_error
        ld a,2
        out (BORDER),a
        scf                     ; carry set = error
        ret

b512_1
        ld d,c                  ; D = 3 bit LSB numero dell'entry
        ld e,0                  ; E = 0, quindi = entry * 256
        srl d                   ; divide /4: adesso DE e` un offset di 64 bytes per ogni valore unitario
        rr e                    ; che avevano i 3 bit LSB, ovvero l'offset della entry da 64 bytes
        srl d                   ; all'interno del blocco da 512 letto precedentemente
        rr e

        ld hl,SHADOW_RAM+2      ; NOTA: SALTA I PRIMI 2 BYTES, utili solo nella prima entry
        add hl,de               ; HL = puntatore al byte con offset [2] dell'entry da modificare

        ld a,(tmp_reg+2)
        ld (hl),a               ; data block type
        inc hl
        ex hl,de
        ld hl,(tmp_reg)         ; copia 32 bytes di filename ad offset +3
        ld bc,32
        ldir

        xor a
        ld b,29                 ; 29 spare bytes are set to ZERO
        ex hl,de
l_spare
        ld (hl),a
        inc hl
        djnz l_spare

        pop de                  ; recupera l'indirizzo usato per la lettura
        call calc_hldec         ; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C
        ld b,1                  ; number of blocks to write
        ld ix,SHADOW_RAM        ; scratchpad area
        call mmc_write_data
        and a
        jr nz,b512_error
        and a                   ; carry reset = no error
        ret


;
;--------------------------------------------------------------------------------------------------
; MMC card initialization for proprietary SNAPSHOT scheme
;--------------------------------------------------------------------------------------------------
gest_cardinit
	ld a,7
	ld (attribute),a
	call cls
	xor a
	ld (x_pos),a
	ld (y_pos),a
	ld hl,init_confirm_msg
	call disp_text

	ld bc,1			; B = offset, C = key number
	ld hl,tmp_str
	ld a,7
	ld (hl),a		; attributi
	inc hl
	inc hl
	ld a,' '
	ld (hl),a
	inc hl
	ld a,'+'
	ld (hl),a
l_dispoffsets
	ld iy,tmp_str+4		; posizione di caricamento per CONV_4CIFRE
	ld a,c
	add '0'
	ld (iy-3),a
	ld l,b			; valore offset attuale (unitario = 16MB)
	ld h,0
	sla l			; moltiplica per 16 per esprimere in megabytes
	rl h
	sla l
	rl h
	sla l
	rl h
	sla l
	rl h
	ld a,(flags2)
	set suppress_leadz,a
	ld (flags2),a
	call conv_4cifre
	xor a
	ld (x_pos),a
	ld a,c
	inc a
	ld (y_pos),a
	ld hl,tmp_str
	call disp_text
	inc c
	ld a,b
	add OFFSET_GRAN		; offset granularity (16 = +256MB)
	ld b,a
	cp MAXOFFSET+1
	jr nc,max_exceeded
	ld a,(card_offset)
	cp b
	jr nc,l_dispoffsets
max_exceeded
	push bc
	call wait_keyrelease	; waits for the key to be released

	xor a
	ld (x_pos),a
	ld a,(y_pos)
	add a,2
	ld (y_pos),a
	ld hl,noinit
	call disp_text

w_keyinitsd
	call key_scan
	jr nz,w_keyinitsd	; piu` di 2 tasti: ignora
	ld a,e
	cp $ff			; no keys
	jr z,w_keyinitsd
	call key_number		; restituisce il numero del tasto premuto (1 - 40)
	pop bc
	cp c
	jr nc,dont_init		; supera il massimo (tasto 0 = key n. 10)
	dec a
	cp 0
	jr z,l_cadone		; 0 = non c'e' bisogno di calcolare l'offset
	ld b,a
	xor a
l_calco
	add OFFSET_GRAN		; computes the corresponding value
	djnz l_calco
l_cadone
	ld (card_offset),a	; and saves it to the offset variable
	call sd_init_for_snap
	jp 0

dont_init
	call restore_screen
	jp wait_key

init_confirm_msg
	db	7,"Select an Initialization Offset (MBytes):",0
noinit
	db	7,"...or press '0' to Abort",0



;
;--------------------------------------------------------------------------------------------------
; This subroutine builds a printable string with DATE/TIME informations at address TMP_STR,
; reading data from (IX) (7 bytes in BINARY format)
;
; The ':' symbol is displayed only when the RTC's 1Hz line is LOW (update occurs on falling edge)
; or when flags2,force_points is SET.
;
; Destroys AF, BC, DE, HL
;--------------------------------------------------------------------------------------------------
build_timestring
	ld de,tmp_str
	ld a,6				; yellow on black
	ld (de),a
	inc de
	ld c,(ix+3)			; day of week
	sla c				; *4
	sla c
	ld b,0
	ld hl,day_names
	add hl,bc
	ld bc,4				; length of day's name
	ldir
no_dayname
	ld b,':'			; the ':' symbol
	ld hl,flags2
	bit force_points,(hl)		; this flag forces the ':' symbol
	jr nz,dupoint
	in a,(STATUS)
	bit 5,a				; otherwise, we test the 1Hz line
	jr z,dupoint
	ld b,' '			; blank
dupoint
	res force_points,(hl)

	ex hl,de			; write pointer to HL

	ld a,(ix+4)			; day of month
	cp 32
	jr nc,rtc_invalid
	cp 0
	jr z,rtc_invalid
	call ldhl_binary
	ld (hl),'/'
	inc hl
	ld a,(ix+5)			; month
	cp 13
	jr nc,rtc_invalid
	cp 0
	jr z,rtc_invalid
	call ldhl_binary
	ld (hl),'/'
	inc hl
	ld (hl),'2'
	inc hl
	ld (hl),'0'			; someone should change this on year 2100 (please)
	inc hl
	ld a,(ix+6)			; year
	cp $9A
	jr nc,rtc_invalid
	call ldhl_binary
	ld (hl),' '
	inc hl
	ld (hl),' '
	inc hl
	ld a,(ix+2)			; hours
	cp 24
	jr nc,rtc_invalid
	call ldhl_binary
	ld (hl),b
	inc hl
	ld a,(ix+1)			; minutes
	cp 60
	jr nc,rtc_invalid
	call ldhl_binary
	ld (hl),b
	inc hl
	ld a,(ix)			; seconds
	cp 60
	jr nc,rtc_invalid
	call ldhl_binary
	ld (hl),0			; null terminated string
	ret

rtc_invalid
	ld hl,tmp_str+1
	ld b,24				; string length
	ld a,' '			; returns a blank string suitable for deleting any previous data
l_clearstr
	ld (hl),a
	inc hl
	djnz l_clearstr
	xor a
	ld (hl),a			; null terminated string
	ret


ldhl_binary
	ld c,'0'
l_bincalc
	sub 10
	jr c,over_bin
	inc c
	jr l_bincalc
over_bin
	ld (hl),c
	add 10+'0'
	inc hl
	ld (hl),a
	inc hl
	ret


;ldhl_bcd				; loads 2 characters from 'A' > (HL), (HL+1)
;	ld c,a				; destroys A, C; updates HL
;	and $F0
;	rrca
;	rrca
;	rrca
;	rrca
;	add '0'
;	ld (hl),a
;	inc hl
;	ld a,c
;	and $0F
;	add '0'
;	ld (hl),a
;	inc hl
;	ret

day_names
	db	"    Mon Tue Wed Thu Fri Sat Sun "


;
;--------------------------------------------------------------------------------------------------
; This subroutines returns NO CARRY if HL points to a snapshot number which is above the maximum
; SD/MMC capacity
;
; Destroys AF, DE
;--------------------------------------------------------------------------------------------------
check_hlspace
	push hl
	ld de,FATSIZE		; dimensioni della FAT in termini di blocchi da 64K
	srl e			; lo trasforma in numero di blocchi da 128K
	ld hl,MMCSIZE		; dimensioni della MMC in termini di blocchi da 128K
	and a
	sbc hl,de		; HL = numero complessivo di snapshot memorizzabili
	ex hl,de
	pop hl
	push hl
	and a
	sbc hl,de		; siamo oltre il massimo?
	pop hl
	ret			; rientra con CARRY SET se non siamo oltre il massimo


;--------------------------------------------------------------------------------------------------
; This subroutine is called by the [E] (edit filename, [W] (edit parameters) and [Q] edit Joystick.
; It loads the SD/MMC BLOCK and sets HL to the desired item, at offset +2 (snapshot size).
;--------------------------------------------------------------------------------------------------
get_hlentry
	ld a,(start_item)
	cp 5
	jr nc,it_ok		; aborts if no snapshot entry is selected
	ret			; returns with NONZERO
it_ok
	call get_entry		; reads the sd/mmc block to TEMP_RAM
	ld a,(start_item)
	sub 5			; 0-based entry pointer
	ld d,a
	ld e,0
	srl d			; each item takes 64 bytes
	rr e
	srl d
	rr e
	ld hl,TEMP_RAM+2	; base address (+2 = snapshot type)
	add hl,de
	ld a,(hl)
	cp 1
	ret z			; only snapshots with code 1, 2 or 8 are valid: return ZERO
	cp 2
	ret z			; 2 = ok = return ZERO
	cp 8
	ret			; return NONZERO if <> 8


;---------------------------------------------------------------------------------------------
; Subroutine che associa i 5 tasti del joystick KEMPSTON ad altrettanti tasti ASCII
;---------------------------------------------------------------------------------------------
kempston_code
	ld hl,kempston_table
	ld b,5
lfind_k
	rrca
	jr c,found_k
	inc hl
	djnz lfind_k
found_k
	ld a,(hl)
	ret

kempston_table
	defb	'0','9','p','o',13



;---------------------------------------------------------------------------------------------
; Subroutine che visualizza i parametri di snapshot puntati da IY (che punta al byte LSB SP)
;
; Distrugge AF, BC, DE, HL, IX
;---------------------------------------------------------------------------------------------
disp_param
        ld a,4
	ld (y_pos),a

	ld a,(iy+5)			; CPLD $24DF register
	and 7				; clock select bits
	ld ixl,a

	ld hl,M35
	ld a,(iy+4)			; F register (IM MODE(D1,0), INT Enable(D2), spectrum type(D3), int doubler(D4))
	bit 4,a
	jr z,hl_isok
	ld de,M35_bis-M35		; string length offset (points to the "double INT frequency" version)
	add hl,de
hl_isok
	ld a,0
l_displayfreq
	ld de,tmp_str
	ld bc,16			; string length
	ldir
	push af
	push hl

	ld hl,tmp_str
	cp ixl				; is it the currently selected frequency?
	jr nz,noqq1
	ld (hl),$0F			; yes, highlights
noqq1
	xor a
	ld (x_pos),a
	call disp_text
	ld a,(y_pos)
	inc a
	ld (y_pos),a

	pop hl
	pop af
	inc a
	cp 8
	jr nz,l_displayfreq


	ld hl,im_modestr
	ld de,tmp_str+1
	ld bc,5
	ldir

	ld b,3
l_dis_p1
	ld a,(iy+4)			; byte F
	and 3				; D1 - D0 = IM MODE
	inc a

	ld ix,tmp_str
	ld c,7				; attributi per voce non attiva
	cp b				; e` quella attiva?
	jr nz,noattiva
	ld c,$0F
noattiva
	ld (ix),c
	ld a,b
	add '0'-1
	ld (ix+4),a
	ld a,12
	sub b
	ld (y_pos),a
	ld a,120
	ld (x_pos),a
	push bc
	ld hl,tmp_str
	call disp_text
	pop bc
	djnz l_dis_p1


        ld hl,ulap_used
        bit 7,(iy+4)
        jr nz,ula_k
        ld hl,ulap_notused
ula_k
	ld a,170
	ld (x_pos),a
        ld a,(y_pos)
        dec a
        ld (y_pos),a
	call disp_text


        ld hl,int_enabledt
	bit 2,(iy+4)
	jr nz,int_k
	ld hl,int_disabledt
int_k
	ld a,170
	ld (x_pos),a
        ld a,(y_pos)
        inc a
        ld (y_pos),a
        call disp_text

        ld hl,spec_typet
	ld a,14
	ld (y_pos),a
	ld a,70
	ld (x_pos),a
	call disp_text


	ld hl,t48k_t
	bit 3,(iy+4)
	jr nz,type_k
	ld hl,t128k_t
type_k
	call disp_text

	ld a,16				; displays the SP register content (from snapshot data)
	ld (y_pos),a
	ld a,70
	ld (x_pos),a
	ld hl,spval_t
	call disp_text
	ld c,(iy+1)			; MSB byte
	call disp_hex
	ld c,(iy)			; LSB byte
	call disp_hex

	ld a,142
	ld (x_pos),a
	ld hl,ival_t			; displays the I register content (from snapshot data)
	call disp_text
	ld c,(iy+2)
	call disp_hex

	ret


im_modestr
	db	"IM  ",0
ulap_used
        db      2,"ULA+ In Use  ",0
ulap_notused
        db      7,"ULA+ Not Used",0
int_enabledt
	db	2,"INT Enabled ",0
int_disabledt
	db	7,"INT Disabled",0
spec_typet
	db	7,"Spectrum Type: ",0
t48k_t
	db	7,"48K ",0
t128k_t
	db	7,"128K",0

M35
	db	7,"3.5MHz -  50Hz",0	; "000"
	db	7,"  7MHz -  50Hz",0	; "001"
	db	7," 14MHz -  50Hz",0	; "010"
	db	7," 21MHz -  50Hz",0	; "011"
	db	7,"  7MHz - 100Hz",0	; "100"
	db	7," 14MHz - 100Hz",0	; "101"
	db	7," 28MHz - 100Hz",0	; "110"
	db	7," 42MHz - 100Hz",0	; "111"
M35_bis
	db	7,"3.5MHz - 100Hz",0	; "000"	- doubled INT version
	db	7,"  7MHz - 100Hz",0	; "001"
	db	7," 14MHz - 100Hz",0	; "010"
	db	7," 21MHz - 100Hz",0	; "011"
	db	7,"  7MHz - 200Hz",0	; "100"
	db	7," 14MHz - 200Hz",0	; "101"
	db	7," 28MHz - 200Hz",0	; "110"
	db	7," 42MHz - 200Hz",0	; "111"

spval_t
	db	7,"SP: $",0
ival_t
	db	7,"I: $",0

;
;------------------------------------------------------------------------------------------------
; Subroutine that displays the C content in HEX format, 2 chars.
;------------------------------------------------------------------------------------------------
disp_hex
	ld hl,tmp_str
	ld (hl),7			; white color
	inc hl
	ld a,c
	call ldhl_hex
	ld (hl),0			; asciz
	ld hl,tmp_str
	call disp_text
	ret

ldhl_hex
	ld c,a
	srl a
	srl a
	srl a
	srl a
	ld b,2
nextcc
	ld e,'0'
	cp 10
	jr c,letter
	ld e,'A'-10
letter
	add e
	ld (hl),a
	inc hl
	ld a,c
	and $0F
	djnz nextcc
	ret

;
;---------------------------------------------------------------------------------------------
; ROM SELECTION AND JUMP TO BASIC
;---------------------------------------------------------------------------------------------
loadthis
        ld a,(sh_flags)
        res bootrom,a
        ld (sh_flags),a
	ld a,l				; salva in A
        cp $FE                          ; e` richiesto il ResiDOS?
	jr nz,no_residos

;
; -- attivazione eventuale ResiDOS presente in ram non volatile
;


	ld a,(flags)
	bit residos_present,a
	jp z,wait_key

	call wait_keyrelease		; aspettiamo il rilascio del tasto, altrimenti rischiamo di
					; selezionare qualche opzione nel menu` del ResiDOS

	ld a,(residos_bank)		; l'area variabili in shadow_rom e` ancora accessibile
	and $1F				; il page_number e` a 5 bit (A18-A14 = 512KBytes)
	or $40				; D6 = high = fast_page register attivo; D5 low = RAM enabled
	ld bc,FASTPAGE
	out (c),a
	xor a				; 3.5MHz, no shadow rd/wr enabled
	ld bc,CLKREG
	out (c),a

	ld hl,0				; contando sul fatto che il CONTEXT SWITCH e` attivo anche se stavamo
	push hl				; girando in shadow_rom (che e` prioritaria su di esso), possiamo
	jp $14				; tranquillamente utilizzare il JP $14 (si presuppone che la EPROM, che e`
					; attiva grazie al context_switch stesso, sia abbastanza simile alla
					; versione del firmware che girava in shadow_rom da contenere l'apposita
					; istruzione RET).

no_residos
	call remove_residos

	ld bc,CLKREG
        xor a                           ; 3.5MHz, ROM_MSB = "000"
	out (c),a

        ld bc,0                         ; loop: il BBC basic parte in 48K
lbbc                                    ; se trova la 'Z' premuta!
        dec bc
        ld a,b
        or c
        jr nz,lbbc
;
; -- lancio di una ROM oppure della SHADOW_ROM
;
        ld a,l                 ; ricarica il numero della rom richiesta
	call spectrum_select

        ld l,a                 ; A contiene il numero del banco desiderato
        ld bc,S128KREG         ; CAMBIO DI BANCO: possiamo impunemente scrivere nel registro S128KREG grazie
        in a,(c)               ; al modo "CONTEXT" attivo nella CPLD, che forza comunque "0011" sui pin di indirizzo.
        res 4,a                ; Stesso discorso non vale per il CLKREG, a causa della presenza del bit SHADOW_RD.
        bit 0,l                ; bit LSB indirizzo ROM (A14)
	jr z,aok
        set 4,a                ; LSB rom_bank in base al contenuto di D0 dI L
aok
	out (c),a

        ld bc,PLUS3REG         ; D2 del registro $1FFD contiene il secondo bit di indrizzamento
        xor a                  ; default basso (ed azzera anche D0, D1 in modo da selezionare la modalita` STANDARD (non +3))
        bit 1,l                ; bit A15 indirizzo rom
	jr z,aaok
        ld a,4                 ; setta A15 (D2 registro $1FFD)
aaok
	out (c),a

        res 0,l                ; i bit LSB devono essere eliminati (finirebbero sul clock_select)
	res 1,l
	ld bc,CLKREG
        sla l                  ; allinea il registro
        out (c),l

	ld hl,0
	push hl
        jp $14                 ; in questo caso usiamo un RET
end_reloc


;
; Subroutine di eliminazione della firma del ResiDOS dalla ram non volatile
;
remove_residos
	ld a,(flags)		; non carichiamo il residos: se e` presente in ram, occorre eliminarne
	bit residos_present,a	; la firma per evitare che venga successivamente scelto (quando ormai la ram
	jr z,no_needtoreset	; sara` stata compromessa).
	ld a,(residos_bank)
	and $18			; Se il banco di lancio del ResiDOS e` > 7, significa che abbiamo piu` di 128K
	jr nz,no_needtoreset	; ed il residos non utilizza nessuna parte dei primi 128K, per cui possiamo
	xor a			; mantenerlo operativo
	ld (residos_sign),a	; altrimenti azzera il primo byte della firma
no_needtoreset
	ret


;---------------------------------------------------------------------------------------------
; This subroutine clears the screen and displays a (HL) string.
; Returns carry SET if key Y is pressed. If any other key is pressed, the main menu is
; displayed and the routine returns carry reset.
;---------------------------------------------------------------------------------------------
yesno_confirm
	push hl
        ld a,7
        ld (attribute),a
        call cls
        xor a
        ld (x_pos),a
        ld (y_pos),a
        pop hl
        call disp_text
        call wait_keyrelease            ; waits for the key to be released

w_keybackup
        call key_scan
        jr nz,w_keybackup               ; piu` di 2 tasti: ignora
        ld a,e
        cp $ff                          ; no keys
        jr z,w_keybackup
        cp 2                            ; Y key
        jr nz,answerno
	scf
	ret				; return carry SET (Y was pressed)
answerno
        call cls
        call disp_menu
        call display_mmcdata
        and a
	ret


;---------------------------------------------------------------------------------------------
; Subroutine che visualizza 8 entries e la posizione del prossimo snapshot
;---------------------------------------------------------------------------------------------
display_mmcdata
	ld a,(sh_flags)
	bit sd_present,a	; rientra subito se la SD/MMC non e` stata trovata
	ret z
	bit sd_initsnap,a	; rientra subito se la SD/MMC non e` stata trovata
	ret z

	call display_entries	; visualizza 8 entries a partire da quella puntata da SD_ENTRY

	ld hl,next_snapstr
	ld a,170
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
	call disp_text
;				prosegue in DISP_FIRSTFREE
;
;---------------------------------------------------------------------------------------------
; Subroutine che visualizza il contenuto di FIRSTFREE in basso a destra
;---------------------------------------------------------------------------------------------
disp_firstfree
	ld hl,(firstfree)
	ld iy,tmp_str+1		; posizione di caricamento per CONV_4CIFRE
	ld a,7
	ld (iy-1),a		; attributi
	call conv_4cifre
	ld (iy+5),0
	ld a,220
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
	ld hl,tmp_str
	call disp_text
	ret


;---------------------------------------------------------------------------------------------
; Subroutine che gestisce la temporizzazione per la riscrittura in SD/MMC di FIRSTFREE
;---------------------------------------------------------------------------------------------
gest_rewrite
	ld a,(rewrite_timer)
	cp 0
	ret z				; nessuna temporizzazione in corso: rientra subito

	dec a
	ld (rewrite_timer),a
	ld ($5AF9),a			; cubetto multicolore che evidenzia il conteggio in corso
	ld ($5AFA),a			; cubetto multicolore che evidenzia il conteggio in corso
	cp 0
	ret nz				; se non e` arrivato a zero, rientra
	ld a,2
	ld ($5AF9),a			; cubetto multicolore che evidenzia il conteggio in corso
	ld ($5AFA),a			; cubetto multicolore che evidenzia il conteggio in corso

save_firstfree
	ld a,(card_offset)
	ld h,a
	ld l,0
	ld de,0
	ld ix,TEMP_RAM
	call mmc_read_data		; performs the single block read
	cp 0
	ret nz

	ld hl,(firstfree)
	ld (TEMP_RAM),hl

	ld a,(card_offset)
	ld h,a
	ld l,0
	ld de,0
	ld b,1				; 1 block
	ld ix,TEMP_RAM
	call mmc_write_data
	ret



;-------------------------------------------------------------------------------------------------
; ULA3 and I/O CPLDs initialization.
;
; Forces "00011" on ROM_BANK (although probably already active due to CONTEXT switch) and selects
; 7MHz / 50Hz frequency for the processor.
;
; Destroys AF, BC, DE, HL
;-------------------------------------------------------------------------------------------------
set_cpldregs
	call cs_high		; disables SD/MMC chip select

	ld bc,UNLOCKREG		; unlocks both $7FFD + $1FFD and all non-standard registers
	ld a,$AA
	out (c),a

	ld bc,FASTPAGE		; disables fastpaging register
	xor a
	out (c),a
	ld bc,SPECTR_MODE	; 48K, no doubleint, rom wrprotect disabled (FPGA), context switch enabled
	out (c),a
	ld bc,DDR_MSB		; ensures first megabyte of DDR is addressed
	out (c),a

;        ld bc,S128KREG          ; for safety reasons loads $C000 as the ROM_BANK
;        in a,(c)
;        set 4,a                 ; LSB bit of ROM_BANK (A14)
;        out (c),a

;        ld bc,PLUS3REG
;        ld a,4
;        out (c),a               ; bit A15

	ld bc,CLKREG		; se siamo qui significa che il banco indirizzato e` gia` il "00011" (A18:A14),
	in a,(c)		; tuttavia carichiamo "00011" nei registri ROM_BANK in modo che, in caso
	and $C0			; di NMI mentre ci troviamo all'interno del menu` di scelta ROM,
				; la successiva uscita dall'NMI (con salto a $15) non tolga questo indirizzo.
	or 2			; clock 14MHz
	out (c),a

	ret


;---------------------------------------------------------------------------------------------
; Subroutine che imposta le variabili ed il colore del BORDER.
; Dovendo accedere in scrittura all'ultimo KB di SHADOW_ROM (che e` la nostra area variabili)
; si accerta preventivamente che il CONTEXT sia attivo, condizione necessaria per avere
; quel KB in modalita` RD/WR.
;
; Per questo motivo, la presente routine provoca l'attivazione del CONTEXT MODE.
;
; Nota: E non deve essere alterato
;---------------------------------------------------------------------------------------------
set_vars
	xor a
	ld (x_pos),a
	ld (y_pos),a
	ld (frames),a
	ld (frames+2),a
	ld (frames+3),a
	ld (rewrite_timer),a
	ld (repeat_timer),a
	ld (flags),a
	ld (flags2),a
	out (BORDER),a		; border nero
	ld (BORDCR),a		; variabile BASIC che conserva il colore del border

	ld a,$C0		; MSB contatore viene inizializzato in modo da sveltire la comparsa
	ld (frames+1),a		; del cubetto lampeggiante (e scansione tastiera)

	ld a,7			; PAPER 0, INK 7
	ld (attribute),a


;
; SHADOW_ROM variables setup
; These variables resides in the last KB of the 16K rom space, that is RD/WR due to the context switch.
;

	ld bc,SPECTR_MODE
	in a,(c)
	set 2,a			; NUOVO: forza lo stato di CONTEXT SWITCH
	out (c),a

	xor a
	ld (nmi_count),a	; NMI counter (nested NMI calls detect)
        set bootrom,a           ; this indicates that BOOTROM is active
	ld (sh_flags),a

	ld hl,0
	ld (snap_torestore),hl


	ret


;---------------------------------------------------------------------------------------------
; Subroutine che rileva la presenza del ResiDOS in ram non volatile
; *** modifica per FPGA ***
; Il banco utilizzato per lo scambio di dati con il residos era il $7
; in ram tamponata, ovvero quello usato per le variabili shadow_rom.
; In questa implementazione, quel banco corrisponde al RAM_BANK 7 dello
; spectrum 128K (!!!) in quanto solo la prima meta` di quel banco e'
; occupato da blockram video a doppia porta.
;
; Di conseguenza, segue un piccolo artificio (ma il residos dovra' essere
; modificato)
;---------------------------------------------------------------------------------------------
residos_detect
;
;Old version detect (used to write signature at address $3D00 of RAM_BANK 7)
;This is the bootrom bank (context_switch activated) of the first CPLD BADALOC
;and $3D00 is the beginning of variables area.
;
;        ld bc,S128KREG
;        in l,(c)
;        ld a,7                          ; rende accessibile il RAM_BANK 7
;        out (c),a
;
;        ld a,RESISPEED_DEFAULT
;        ld (residos_speed+$C000),a                   ; ResiDOS max allowable speed
;
;        exx
;        ld hl,$C000+residos_sign
;        ld de,residos_sign
;        ld bc,end_residos-residos_sign
;        ldir
;        exx
;        out (c),l                       ; rimette a posto il RAM_BANK


;
;New version of ResiDOS now writes at address 0000 of same bank. In FPGA
;implementation, this lower half of bank 7 is never accessed because shaded-out
;by 8KB of video blockram. Only FASTPAGE access can reveal it.
;On the original CPLD Badaloc, it will overwrite a few bytes at beginning
;of bootrom code.
;

	ld bc,UNLOCKREG		; unlocks both $7FFD + $1FFD and all non-standard registers
	ld a,$AA
	out (c),a

	ld bc,SPECTR_MODE	; enables $C000 mode for fastpaging
        in a,(c)
        ld e,a
        set 7,a
        out (c),a

	ld bc,FASTPAGE
        in d,(c)
        ld a,$C7                ;RAM bank 7 now accessible at $C000
        out (c),a               ;(all from DDR: no video blockram on C000-BFFF)


	ld a,RESISPEED_DEFAULT
        ld ((residos_speed-start_residos)+$C000),a  ; ResiDOS max allowable speed

        exx
        ld hl,$C000
        ld de,residos_sign
        ld bc,end_residos-residos_sign
        ldir
        exx

        out (c),d                       ; FASTPAGE is restored
        ld bc,SPECTR_MODE
        out (c),e                       ; SPECTR_MODE is restored

	ld hl,flags
        res residos_present,(hl)
;	ld a,(battery_wrn)
;	cp 0
;	ret nz				; no ResiDOS if battery low was detected
        ld a,(residos_sign)             ; this is a shadow_rom shared variable written by ResiDOS
	cp 'R'				; first signature byte should be 'R'
	ret nz
        ld a,(residos_sign+1)
	cp 'D'				; second signature byte should be 'D'
	ret nz
	set residos_present,(hl)
	ret

;---------------------------------------------------------------------------------------------
; Subroutine che inizializza la tavolozza dei colori ULA+
;---------------------------------------------------------------------------------------------
Ula_DefaultColors
        db 0, 1, 8, 9, $40, $41, $48, $49 ;ink tabella 1 (flash, bright 0,0)
        db 0, 1, 8, 9, $40, $41, $48, $49 ;paper tabella 1

        db 0, 2, $10, $12, $80, $82, $90, $92 ;ink tabella 2 (flash, bright 0,1)
        db 0, 2, $10, $12, $80, $82, $90, $92 ;paper tabella 2

        db 0, 1, 8, 9, $40, $41, $48, $49 ;ink tabella 1 (flash, bright 0,0)
        db 0, 1, 8, 9, $40, $41, $48, $49 ;paper tabella 1

        db 0, 2, $10, $12, $80, $82, $90, $92 ;ink tabella 2 (flash, bright 0,1)
        db 0, 2, $10, $12, $80, $82, $90, $92 ;paper tabella 2

UlaPlus_Colors
        ld e,0
        ld hl,Ula_DefaultColors
ulaloop
        ld bc,ULAP_CTRL
        out (c),e
        ld bc,ULAP_DATA
        ld a,(hl)
        out (c),a
        inc hl
        inc e
        bit 6,e
        jr z,ulaloop

        ld bc,ULAP_CTRL
        ld a,64
        out (c),a
        ld bc,ULAP_DATA
        xor a                   ; ULA+ disabled
        out (c),a
        ld bc,ULAP_CTRL
        ld a,255                ; invalid value
        out (c),a
        ret


;---------------------------------------------------------------------------------------------
; Subroutine che confronta il contenuto dei registri ULA+ con la tavolozza
; di default. Se non corrisponde, li salva sulle ultime due righe dello schermo.
; In caso di salvataggio, setta il flag 'ulap_in_use'.
;
; Distrugge AF, BC, D, HL
;---------------------------------------------------------------------------------------------
UlaPlus_Save
        ld d,0
        ld hl,Ula_DefaultColors
l_checkula
        ld bc,ULAP_CTRL         ; registro controllo ULA+
        out (c),d               ; seleziona il colore puntato da D (0-63)
        ld bc,ULAP_DATA         ; registro dati ULA+
        in a,(c)                ; legge il valore del colore
        cp (hl)                 ; confronta con la tabella di power-on
        jr nz,Ula_in_use        ; salta se almeno un colore e' stato alterato
        inc hl
        inc d
        bit 6,d
        jr z,l_checkula
        ld hl,sh_flags          ; segnala ula+ non in uso
        res ulap_in_use,(hl)
        jr QUlap_detect         ; fine operazioni ULA+
;
; Copia i colori ULA+ in zona ram video
;
Ula_in_use
        ld hl,sh_flags          ; segnala ula+ utilizzata
        set ulap_in_use,(hl)

        ld d,0
        ld hl,$56E0             ; penultima righina in basso
l_saveula
        ld bc,ULAP_CTRL         ; registro controllo ULA+
        out (c),d               ; seleziona il colore puntato da D (0-63)
        ld bc,ULAP_DATA         ; registro dati ULA+
        in a,(c)                ; legge il valore del colore
        ld (hl),a               ; lo salva
        inc hl
        inc d
        ld a,d
        cp 32
        jr nz,no_2riga
        ld hl,$57E0             ; ultima riga
no_2riga
        bit 6,d
        jr z,l_saveula
QUlap_detect
        ret


;---------------------------------------------------------------------------------------------
; Subroutine che ripristina la tavolozza ULA+ attingendo dalle ultime due
; righe dello schermo (64 bytes).
;
; Distrugge AF, BC, D, HL
;---------------------------------------------------------------------------------------------
UlaPlus_Restore
        ld d,0
        ld hl,$56E0             ; penultima righina in basso
l_restoreula
        ld bc,ULAP_CTRL         ; registro controllo ULA+
        out (c),d               ; seleziona il colore puntato da D (0-63)
        ld bc,ULAP_DATA         ; registro dati ULA+
        ld a,(hl)               ; legge il colore dalla ram
        out (c),a               ; e lo ripristina
        inc hl
        inc d
        ld a,d
        cp 32
        jr nz,no_riga22
        ld hl,$57E0             ; ultima riga
no_riga22
        bit 6,d
        jr z,l_restoreula

        ld bc,ULAP_CTRL
        ld a,64                 ; control register
        out (c),a
        ld bc,ULAP_DATA
        ld a,1                  ; enable ula+
        out (c),a
        ld bc,ULAP_CTRL
        ld a,255                ; invalid setting
        out (c),a
        ret

;---------------------------------------------------------------------------------------------
; Subroutine che visualizza il menu principale (scelta rom)
;---------------------------------------------------------------------------------------------
disp_menu
	xor a
	ld (x_pos),a
	ld (y_pos),a
	ld hl,logo		; stringa asciiz LOGO di sistema
	call disp_text

	ld hl,rom0		; elenco ROM disponibili
	xor a
	ld (x_pos),a
	ld a,5
	ld (y_pos),a
	call disp_text

	ld hl,rom3e
	xor a
	ld (x_pos),a
	ld a,6
	ld (y_pos),a
	call disp_text

        ld hl,bbcbasic
	xor a
	ld (x_pos),a
        ld a,7
	ld (y_pos),a
	call disp_text

	ld a,(flags)		; e` stato rilevato il ResiDOS in ram non volatile?
	bit residos_present,a
	jr z,no_respres

	ld hl,resi_ram
	xor a
	ld (x_pos),a
	ld a,8
	ld (y_pos),a
	call disp_text

	ld hl,tmp_str			; now displays the ResiDOS version
	ld (hl),7			; white color
	inc hl
	ld a,(residos_version)
	ld c,a
	srl a
	srl a
	srl a
	srl a
	cp 0
	jr z,suppress_zero
	add '0'
	ld (hl),a
	inc hl
suppress_zero
	ld a,c
	and $0F
	add '0'
	ld (hl),a
	inc hl
	ld (hl),'.'
	inc hl
	ld (hl),0			; asciz
	ld hl,tmp_str
	call disp_text
	ld a,(residos_version+1)	; remaining byte can be processed by standard disp_hex call
	ld c,a
	call disp_hex

	
no_respres
	ld hl,sh_flags
	bit sd_initsnap,(hl)	; is the present SD/MMC initialized for our proprietary snapshot scheme?
	jr z,no_sdprop
	call proprietary_menu
	ret
no_sdprop
	call extended_menu	; displays alternate menu
	ret


proprietary_menu
	ld a,19			; posizione di visualizzazione
	ld (y_pos),a
	xor a
	ld (x_pos),a
	ld hl,yellowarrow
	call disp_text

	ld hl,please		; [O,P] [9,0] [N,M] menu for snapshot navigation
	xor a
	ld (x_pos),a
	ld a,2
	ld (y_pos),a
	call disp_text

	ld hl,snap1		; snapshot buttons
	xor a
	ld (x_pos),a
	ld a,3
	ld (y_pos),a
	call disp_text
	ld hl,snap2
	call disp_text
	ld hl,snap3
	call disp_text
	ld hl,snap4
	call disp_text

        ld hl,rambackup
        xor a
        ld (x_pos),a
        ld a,4
        ld (y_pos),a
        call disp_text

	ld hl,keys1		; other snapshot properties buttons
	xor a
	ld (x_pos),a
	ld a,20
	ld (y_pos),a
	call disp_text

	ld hl,keys2
	xor a
	ld (x_pos),a
	ld a,21
	ld (y_pos),a
	call disp_text
	ret

extended_menu
	ld hl,sh_flags
	bit sd_present,(hl)	; is an sd-card present?
	ld hl,extend_1		; card init message
	jr nz,msg_correct
	ld hl,nocardmsg
msg_correct
	xor a
	ld (x_pos),a
	ld a,2
	ld (y_pos),a
	call disp_text
	ret



logo            db      7,"Nexys2 FPGA ZX-Badal~c V1.01",0
please		db	1,"[O,P] = Cursor;  [9,0] +- 8;  [N,M] +- 128",0
rom0            db      7,"[Z] 128K ROM 0; [X] 128K ROM 1; [C] 16/48K",0
rom3e           db      7,"[V] 128K +3E ROM by Garry Lancaster",0
bbcbasic        db      7,"[B] BBC Basic",0
resi_ram        db      7,"[R] ResiDOS (in Ram) V",0
nocardmsg	db	1,"No SD-CARD Detected.",0
mmc		db	7,"SD CID = ",0
next_snapstr	db	7,"Next --> ",0
keys1		db	1,"[Q] Joystick; [W] Parameters; [E] Filename",0
keys2		db	1,"[s,' '] Change 'Next Snap' slot number",0
snap1           db      1,"Snap: NMI --> ",0
snap2		db	4,"[Q]=16K; ",0
snap3		db	5,"[W]=48K; ",0
snap4		db	2,"[E]=128K",0
;rambackup       db      6,"[U]Ram Backup [I] Flash Backup",0
rambackup       db      6,"[U] 512K Ram Backup",0

extend_1	db	1,"Press [Q] to initialize CARD for SnapShots",0


;
;---------------------------------------------------------------------------------------------
; Copia le rom da SPI a DDR utilizzando il FASTPAGING su $C000
;---------------------------------------------------------------------------------------------
spi2ddr_copy

	call cls

	xor a
	ld (x_pos),a
	ld (y_pos),a
	ld hl,spiddr_msg
	call disp_text

	ld bc,SPECTR_MODE	;enables $C000 mode for fastpaging
	in a,(c)
	set 7,a
	out (c),a

	ld ix,copytable		;Address TABLE pointer

spiddr_nextbank
	ld a,(ix)
	cp 255			;255 = end of table
	jr nz,dobank

	xor a			;disables fastpaging
	ld bc,FASTPAGE
	out (c),a

	ld bc,SPECTR_MODE	;disables $C000 mode for fastpaging
	in a,(c)
	res 7,a
	out (c),a

	ld e,20
	ld bc,0			;small delay loop
l_ddr
	dec bc
	ld a,b
	or c
	jr nz,l_ddr
	dec e
	jr nz,l_ddr
	ret			;DONE

dobank
	call display_romaddress	;displays a row with SPI and DDR addresses
	ld a,(ix)
	rrca
	rrca
	rrca			;*32
	ld c,a
	and $E0
	add 16
	ld iyl,a		;LSB attribute in screen line
	ld a,c
	and 3
	add $58
	ld iyh,a		;MSB attribute

retryddr_loop
	ld hl,$C000		;destination write pointer
	ld e,1			;E = 0 = READ; 1 = VERIFY

	call spi_16Kread
	ld a,e			;0 = no error
	cp 0
	jr z,spiddr_complete

	ld (iy),16		;red box
	inc iy

	ld hl,$C000		;destination write pointer
	ld bc,16384		;bytes to copy
	ld e,0			;E = 0 = READ; 1 = VERIFY

	call spi_16Kread
	jr retryddr_loop	;mettere contatore di tentativi

spiddr_complete
	ld (iy),32		;green box

	ld bc,6
	add ix,bc		;next item in the table
	jr spiddr_nextbank

spiddr_msg
        db 7,"SPI ROM     RAM Address",0

;
;The Copy Table:
;
;1 byte screen line number to display information for this entry
;3 bytes SPI address of source 16K ROM
;1 byte ddr_msb register for destination address
;1 byte FASTPAGE register content during copy (destination address)
;
; FASTPAGE format:
; D7 	= Write enable
; D6 	= Read enable (enable paging)
; D5 	= MSB bank address: for compatibility, '0' = ram bank; '1' = rom bank (now both in ddr ram)
; D4:D0 = Bank number (0-31)
;
; In original Badaloc and ZXmmc+, D5 was the RAM/ROM select. Here they both reside into a 1MB DDR area.
; When normally addressed, the DDR bank will provide 512K of "rom" and 512K of ram. For compatibility reason,
; the FPGA will hard-wire RAM accesses to first half of the megabyte (D5 = '0' = RAM) and ROM accesses to second half.
;
;
;Example:	0	  first screen line
;
;	 	$10 00 00 SPI address at +1MB offset
;		
;		$00	  ddr_msb: unit value = +1MB offset into DDR address
;
;		$E2	  WR enable, RD enable, Write to "ROM" half (+512K) because D5 = '1', bank n.2 (D1 = '1')
;
copytable
	db 2
        db $10,0,0      ;copy SPI at offset +1MB into ROM BANK 0 (128K ROM0)
	db 0,$E0

	db 3
	db $10,$40,0	;copy SPI at offset +1MB+16K into ROM BANK 1
	db 0,$E1

	db 4
        db $10,$80,0    ; 16/48K rom on bank 2
	db 0,$E2

	db 5
        db $10,$C0,0    ; unused: was bootrom bank (3) but still copied
	db 0,$E3

	db 6
        db $11,0,0      ; 4 banks for +3E rom
	db 0,$E4

	db 7
	db $11,$40,0
	db 0,$E5

	db 8
	db $11,$80,0
	db 0,$E6

	db 9
	db $11,$C0,0
	db 0,$E7

        db 10
        db $12,0,0      ; 2 banks for BBC roms on 8,9
        db 0,$E8

        db 11
        db $12,$40,0
        db 0,$E9

	db 255		;end of table



;
; This routine reads 16K of SPI FLASH data into (HL).
; If E = 0 read is performed; if E = 1 verify is performed: on return, E = 0 (OK), E = 1 (error)
;
; Source Address in SPI chip and DDR MSB address should be specified in a table pointed by IX:
;
; (IX+1)  MSB SPI address byte (A23:A16)
; (IX+2) LMSB SPI address byte (A15:A8)
; (IX+3)  LSB SPI address byte (A7:A0)
;
; (IX+4) DDR_MSB register (unit value = 1MB offset)
; (IX+5) 16K bank select: FASTPAGE (see caller documentation for fastpage format)
;
spi_16Kread
	xor a
	out (OUT_PORT),a	;seleziona CS3 = spi onboard

	ld a,3			;READ command
	out (SPI_PORT),a	;command byte

	ld a,(ix+1)		;MSB Byte address
	out (SPI_PORT),a
	ld a,(ix+2)
	out (SPI_PORT),a
	ld a,(ix+3)
	out (SPI_PORT),a

	ld a,(ix+4)		;DDR MSB address (unit value = 1MB)
	ld bc,DDR_MSB
	out (c),a

	ld a,(ix+5)		;D7 = fastpage WR enable; D6 = fastpage active; D5 low = ram is paged, A4:A0 = bank
	ld bc,FASTPAGE
	out (c),a

	in a,(SPI_PORT)		;dummy read
	ld a,e			;read or verify command
	ld de,16384		;data to be transferred
	cp 1
	jr z,ddrverify

l_spiread
	in a,(SPI_PORT)
	ld (hl),a
	inc hl
	dec de
        ld a,d
;        and 7
;        out (BORDER),a
;        ld a,d
	or e
	jr nz,l_spiread
qspiddr
	ld e,1			; 1 = error (only meaningful when verifying)
espiddr
	ld a,3
	out (OUT_PORT),a	; disabilita SPI flash
	ret

ddrverify
	in a,(SPI_PORT)
	cp (hl)
	jr nz,qspiddr		; NZ = ERROR
	inc hl
	dec de
	ld a,d
	or e
	jr nz,ddrverify
	ld e,0			; verify OK
	jr espiddr


;
;-- Displays SPI and DDR addresses (from (IX) data)
;
display_romaddress
	ld (y_pos),a
	xor a
	ld (x_pos),a

	ld hl,tmp_str
	ld (hl),7
	inc hl
	ld (hl),'$'
	inc hl
	ld a,(ix+1)
	call ldhl_hex
	ld a,(ix+2)
	call ldhl_hex
	ld a,(ix+3)
	call ldhl_hex

	ld (hl),' '
	inc hl
	ld (hl),'-'
	inc hl
	ld (hl),'-'
	inc hl
	ld (hl),'>'
	inc hl
	ld (hl),' '
	inc hl

	ld a,(ix+4)
	call ldhl_hex
;        ld (hl),' '
;        inc hl

	ld a,(ix+5)
	and $3C
	rrca
	rrca
	call ldhl_hex
	dec hl
	ld a,(hl)
	dec hl
	ld (hl),a
	inc hl

	ld a,(ix+5)
	and 3
	rrca
	rrca
	call ldhl_hex
	xor a
	call ldhl_hex
	ld (hl),0
	ld hl,tmp_str
	call disp_text
	ret


;---------------------------------------------------------------------------------------------
; Subroutine che verifica la presenza di una SD/MMC inserita nello slot.
; Se la trova ne visualizza il CID e verifica che i primi due bytes del secondo blocco
; contengano $AA55. Se e` cosi`, legge il numero del prossimo snapshot dai primi due bytes
; della card, altrimenti inizializza entrambi i campi. Setta il flag SD_PRESENT.
;---------------------------------------------------------------------------------------------
sd_check
	ld a,(sh_flags)
	res sd_present,a
	res sd_initsnap,a
	ld (sh_flags),a

	ld hl,TEMP_RAM		; tenta di leggere il CID di una eventuale SD/MMC salvandolo a TEMP_RAM
	call mmc_getcid_poweron
	cp 0
	ret nz			; errore

	ld hl,mmc		; visualizza il CID della SD rilevata
	xor a
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
	call disp_text
	ld hl,TEMP_RAM+8
	xor a			; la stringa e` ASCIZ
	ld (hl),a
	ld hl,TEMP_RAM		; primo carattere stampabile (per 7 caratteri) = +1; questo viene
	ld a,7			; sovrascritto con il codice dell'attributo desiderato
	ld (hl),a
	call disp_text
	ld hl,sh_flags
	set sd_present,(hl)

	xor a
	ld (card_offset),a	; starts from beginning of card (offset 0)
l_newoffset
	ld l,0
	ld h,a			; card_offset is loaded on 4o byte (address bits 31-24) of card absolute address
	ld de,0
	ld b,2
	ld ix,TEMP_RAM
	call mmc_read_multidata	; reads the first 2 blocks
	cp 0
	jr z,rseek_ok
	ld a,(card_offset)	; read error: marks the last successful read address
	dec a			; and returns
	ld (card_offset),a
	ret

rseek_ok
	ld hl,(TEMP_RAM+512)	; checks the first 2 bytes of the second block
	ld a,l			; that should contain $AA55 for cards that have been
	cp $55			; initialized for our proprietary snapshot filesystem
	jr nz,no_mmcok
	ld a,h
	cp $AA
	jr z,check_ok

no_mmcok			; now we seek for header on other card offsets
	ld a,(card_offset)
	add OFFSET_GRAN		; offset granularity (16 = +256MB)
	ld (card_offset),a
	cp MAXOFFSET+1
	ret nc			; none of card offsets contains a valid header
	jr l_newoffset
	ret

sd_init_for_snap		; SD INIT ENTRY POINT
	ld a,(card_offset)
	ld h,a
	ld l,0
	push hl
	ld de,0
	ld b,2
	ld ix,TEMP_RAM
	call mmc_read_multidata	; reads the first 2 blocks
	ld hl,$AA55
	ld (TEMP_RAM+512),hl	; initializes the identifier
	ld hl,0
	ld (TEMP_RAM),hl	; initializes the "first snapshot entry" counter

	pop hl
	ld de,0
	ld b,2
	ld ix,TEMP_RAM
	call mmc_write_data	; writes the data back to the SD/MMC

check_ok
	ld hl,(TEMP_RAM)
	ld (firstfree),hl
	ld bc,10000
	and a
	sbc hl,bc
	jr c,no_overn
	ld hl,1
	ld (firstfree),hl

no_overn
	ld hl,sh_flags		; this flag indicates that the present SD/MMC has been found to be
	set sd_initsnap,(hl)	; initialized for our proprietary snapshot scheme

	ld iy,tmp_str+2
	ld (iy-2),7		; color
	ld (iy-1),'+'
	ld a,(card_offset)
	ld l,a			; valore offset attuale (unitario = 16MB)
	ld h,0
	sla l			; moltiplica per 16 per esprimere in megabytes
	rl h
	sla l
	rl h
	sla l
	rl h
	sla l
	rl h
	ld a,(flags2)
	set suppress_leadz,a
	ld (flags2),a
	call conv_4cifre
	ld (iy+1),'M'
	ld (iy+2),'B'
	ld (iy+3),0
	ld a,120
	ld (x_pos),a
	ld hl,tmp_str
	call disp_text

	ret


;
;---------------------------------------------------------------------------------------------
; Subroutine che carica l'attributo A alla posizione indicata da START_ITEM
;---------------------------------------------------------------------------------------------
showsel_item
	push hl
	sra a
	ld c,a			; attributo

	ld a,(start_item)
	ld d,a
	ld e,0
	srl d
	rr e
	srl d
	rr e
	srl d
	rr e
	ld hl,$58A0		; posizione prima voce di menu
	add hl,de

	ld a,(flags)
	bit restore_attrib,a	; dobbiamo ripristinare l'attributo precedentemente sovrascritto?
	jr z,norestore
	res restore_attrib,a
	ld (flags),a
	ld a,(saved_attrib)
	ld c,a
	jr nosave
norestore
	bit save_attrib,a	; dobbiamo salvare l'attributo prima di sovrascriverlo?
	jr z,nosave
	res save_attrib,a
	ld (flags),a
	ld a,(hl)
	ld (saved_attrib),a
nosave
        ld (hl),c	    	; cubetto multicolore scelta di default?
        inc hl
        ld (hl),c
	pop hl
	ret


;
;------------------------------------------------------------------------------------------------
; The DISPLAY_SD_ENTRIES subroutine
;------------------------------------------------------------------------------------------------
display_entries
	call get_entry
	ld b,8			; number of entries to display
	ld ix,TEMP_RAM+3	; offset inizio filename

	ld a,10			; posizione di visualizzazione
	ld (y_pos),a

l_dispe
	ld a,(ix-1)		; tipo di snapshot: 1 = 16K, 2 = 48K, 8 = 128K
	ld c,4			; verde per snapshot 16K
	cp MMC_SNAP16		; 1
	jr z,l_dispp
	ld c,5			; cyan per snapshot 48K
	cp MMC_SNAP48		; 2
	jr z,l_dispp
	ld c,2			; rosso per snapshot 128K
	cp MMC_SNAP128		; 8
	jr z,l_dispp
	ld c,7			; bianco per raw_data
	cp MMC_RAWDATA		; 3
	jr z,l_dispp
        ld c,6                  ; giallo per 512K RAM backup
        cp MMC_RAMBACKUP
        jr z,l_dispp
        cp MMC_FLASHBACKUP
        jr z,l_dispp
	ld c,1			; invalid: blu
l_dispp
	push bc
	call disp_number	; displays KEY and SD/MMC entry number

	ld a,(attribute)
	cp 1
	jr nz,displayit
	call make_blank		; creates a blank string at TMP_STR+1
	jr godisplay

displayit
	ld a,7			; white
	ld (tmp_str),a
	push ix
	pop hl
	ld de,tmp_str+1
	ld b,32			; copia il nome del file
copyname
	ld a,(hl)
	cp 0			; sopprime eventuali 0 sostituendoli con spazi
	jr nz,noblanka
	ld a,' '
noblanka
	ld (de),a
	inc de
	inc hl
	djnz copyname
	ex hl,de

godisplay
	ld (hl),0

	ld a,(x_pos)
	add a,3
	ld (x_pos),a
	ld hl,tmp_str
	call disp_text		; visualizza la stringa

next_ent
	ld de,64
	add ix,de
	ld a,(y_pos)
	inc a
	ld (y_pos),a

	pop bc
	djnz l_dispe

	ld hl,flags2
	set force_points,(hl)
	call disptimedate	; displays DATE and TIME for the (start_item-5) entry
	ret


get_entry
	ld hl,(sd_entry)	; gets the SD_ENTRY data in TEMP_RAM
	ex hl,de
	call calc_hldec		; HL, DE = 32bit address of block to be read
	ld ix,TEMP_RAM
	call mmc_read_data	; reads the single block
	ret

put_entry
	ld hl,(sd_entry)	; puts the SD_ENTRY data --> SD/MMC card
	ex hl,de
	call calc_hldec		; HL, DE = 32bit address of block to be read
	ld ix,TEMP_RAM
	ld b,1
	call mmc_write_data	; writes the block
	ret

make_blank
	ld hl,tmp_str+1		; creates a BLANK STRING at tmp_str+1
	ld b,32
	ld a,' '
l_blankname
	ld (hl),a
	inc hl
	djnz l_blankname
	ld (hl),0		; asciz
	ret

;
;------------------------------------------------------------------------------------------------
; Subroutine that displays entry key-select and entry number, with C attribute
;
; (SD_ENTRY) = base number (first entry number)
; 8-B = this entry offset number
; IX = Address of filename in entry space
; C = attribute code
;
; Destroys AF, BC, DE, HL, IY
;------------------------------------------------------------------------------------------------
disp_number
	ld hl,(sd_entry)	; entries base number
	ld a,8
	sub b
	ld e,a
	ld d,0
	add hl,de

	ld iy,tmp_str+5
	ld (iy-4),'['
	ld (iy-2),']'
	ld (iy-1),' '
	add '1'
	ld (iy-3),a
	ld (iy-5),c

	call conv_4cifre	; trasforma HL in ascii all'indirizzo IY

	ld hl,tmp_str
	xor a
	ld (x_pos),a
	call disp_text
	ret


;
;---------------------------------------------------------------------------------------------
; Subroutine that displays TIME and DATE saved in the currently highlighted snapshot item
;---------------------------------------------------------------------------------------------
disptimedate
	ld ix,TEMP_RAM+35	; indirizzo inizio dati RTC del primo item
	ld de,64		; offset per blocco
	ld a,(start_item)
	sub 5
	ret c
	jr z,ix_disok
	ld b,a
l_calix
	add ix,de
	djnz l_calix
ix_disok
	call build_timestring

	ld a,19			; posizione di visualizzazione
	ld (y_pos),a
	ld a,51
	ld (x_pos),a
	ld hl,tmp_str
	call disp_text
	ret

yellowarrow			; questa e` visualizzata da 'disp_menu'
	db	6,"Took -->",0

;
;------------------------------------------------------------------------------------------------
; Subroutine that converts an integer (HL) to a string (max 4 characters), writing at (IY).
; Leading zero are suppressed if proper flag is set.
;
; Returns: String written at (IY), IY pointing to last valid character in string
; Destroys AF, DE, HL, IY
;------------------------------------------------------------------------------------------------
conv_4cifre
	push bc
	ld a,(flags2)
	ld c,a

	ld de,1000
	call conv_cifra
	ld de,100
	call conv_cifra
	ld de,10
	call conv_cifra
	ld a,l
	add '0'
	ld (iy),a
	ld (iy+1),0			; asciz

	res suppress_leadz,c
	ld a,c
	ld (flags2),a
	pop bc
	ret

conv_cifra
	ld a,'0'
l_1000
	and a
	sbc hl,de
	jr c,nomille
	inc a
	res suppress_leadz,c
	jr l_1000
nomille
	ld (iy),a
	bit suppress_leadz,c
	jr nz,sup_1000
	inc iy
sup_1000
	add hl,de
	ret


;*****************************************************************************************************************
; RIPRISTINO SNAPSHOT DA SD/MMC 16K/48K/128K
;
; Lo snapshot viene preso mediante pulsante NMI + tasto perdeterminato, in modo da non MODIFICARE NULLA
; in ram o sullo schermo (come per il cambio di clock). Il ripristino, invece, avviene dal menu` principale
; al power-on (come per la scelta di una ROM). SE STIAMO GIRANDO IN SHADOW_ROM per collaudare il software,
; non avremo il CONTEXT MODE in quanto saremo arrivati NON da un vero POWER-ON ma da un upload che lo avra`
; disattivato. Quindi, con buone probabilita` di riuscita, testiamo lo stato del bit SHADOW RD ENABLE e
; se lo troviamo attivo eseguiamo la rilocazione in RAM (video: non c'e' altro posto sicuro, dopo il ripristino
; dello snapshot) mentre in caso contrario cio` non avviene (collaudare se funziona una volta che sara`
; fatta l'EPROM). La rilocazione comprende la parte di codice che altera il ROM_BANK.
;
; Non esegue il ripristino se il numero di snapshot specificato e` oltre la capacita` della SD/MMC.
;*****************************************************************************************************************
; READ EXAMPLE
read_snap
	call check_hlspace	; controlla se lo snapshot specificato in HL e` oltre la capacita` della SD/MMC
	jr c,snap_ok		; no, possiamo ripristinare

	ld a,2
	out (BORDER),a		; segnala l'errore
	jp wait_key		; errore: rientra in loop

snap_ok
	ex hl,de		; number of snapshot in DE

	ld bc,UNLOCKREG			; UNLOCK di tutti i registri
	ld a,$AA
	out (c),a

	ld bc,CLKREG			; gets the actual Z80 clock rate
	in a,(c)
	ld (prev_clock),a		; saves it
	and $F8

	or RDSNAP_SPEED			; then selects 14MHz or 21MHz / 50Hz screen. Speed should be SWITCHED
					; TO 14MHz before communication to PIC is attempted (joystick
					; programming)
	out (c),a


	ld hl,0
	add hl,sp			; SP in HL
	ld (real_sp),hl			; saves to a shadow_rom variable
	ld sp,SHADOW_STACK

	ex hl,de
	ld (snap_torestore),hl		; salva il numero dello snapshot da ripristinare
	ex hl,de

	call calc_hldec			; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C

	ld ix,SHADOW_RAM		; scratchpad area
	call mmc_read_data		; performs the single block read
	cp 0
        jr z,restore_ok1

restore_error
	ld a,2
	out (BORDER),a			; segnala l'errore

restore_abort
	ld hl,(real_sp)			; restores the SP
	ld sp,hl

	ld a,(prev_clock)
	ld bc,CLKREG			; restores the clock
	out (c),a

        jp waitkey_countdown		; errore: rientra in loop azzerando il contatore di lancio su timeout

restore_ok1

	ld b,c				; BC = offset 64 bytes entry da leggere
	ld c,0
	srl b
	rr c
	srl b
	rr c

	ld hl,SHADOW_RAM		; scratchpad area
	add hl,bc			; aggiunge l'offset relativo all'entry da leggere (ce ne sono 8 in 512 bytes)
	ld (entry_addr),hl

	ld ix,(entry_addr)
	ld a,(ix+2)			; n. blocchi da ripristinare: 1 = 16K, 2 = 48K, 8 = 128K
	cp MMC_SNAP16			; 1
	jr z,snap_valid_nb
	cp MMC_SNAP48			; 2
	jr z,snap_valid			; 48K: ripristinare subito il ram_bank
	cp MMC_SNAP128			; 8
	jr z,snap_valid_nb
        cp MMC_RAMBACKUP
        jr z,snap_valid_nb
	ld a,2
	out (BORDER),a
	jr restore_error


snap_valid
	ld a,(ix+45)			; stato del registro $7FFD (S128KREG)
	and 7
	ld l,a				; L = stato del RAM_BANK al momento dello SNAP
	ld bc,S128KREG
	in a,(c)
	and $F8				; mantiene tutti i bit tranne il RAM BANK
	or l				; impostiamo subito il RAM_BANK: in caso di snap 48K sara` bene scrivere
	out (c),a			; nel banco che alla fine diventera` quello attivo

	ld bc,PLUS3REG			; si accerta che non sia attiva una modalita` di indirizzamento speciale:
	xor a				; impedirebbe l'azione del RAM_BANK nella scelta dei banchi 0-7 su $C000-$FFFF,
	out (c),a			; che e` l'indirizzo di ripristino per i 128K
					; notare che D2 e` A15 della ROM, ma noi stiamo girando in CONTEXT MODE

snap_valid_nb
	push ix
	pop hl
	ld de,48			; offset of joystick programming block
	add hl,de			; HL = read pointer
        call joy_program                ; programmazione del joystick

	ld hl,(snap_torestore)
	ex hl,de
	ld l,FATSIZE			; n. blocchi da 64KB che costituiscono la FAT (e` l'offset per la word MSB)
	ld a,(card_offset)
	ld h,a
	sla e
	rl d				; moltiplica per 2 (ogni entry corrisponde ad un cluster da 128KB)
	add hl,de			; HL = word MSB scrittura in SD/MMC

	ld de,0				; word LSB = 0 (insieme a D0 = 0 della word MSB = 128K)


        ld a,(ix+2)                     ; n. blocchi da ripristinare: 1 = 16K, 2 = 48K, 8 = 128K
        cp MMC_RAMBACKUP
        jp nz,no_restorebackup


;
; Ripristino di un backup RAM 512K zxmmc+
;
	push hl
	push de
        ld hl,restore_confirm
	call yesno_confirm
	pop de
	pop hl
        jp nc,restore_abort

	xor a
	ld (flash_bank),a		; flag "restore ram"

	ld bc,FASTPAGE
	in a,(c)
        ld (fastp_save),a       	; FASTPAGE content is saved (in spectrum ram)

	ld bc,SPECTR_MODE
	in a,(c)
        ld (spectrm_save),a
	set 7,a				; $C000 mode for FASTPAGE
	out (c),a

        ld a,32                 	; blocchi da 16K da leggere
l_ramrestore
        push af

	ld b,a
        ld a,$E0
        sub b                   	; lo trasforma in numero 0-31 con D6 (ram paged in) e D7 (ram WR) attivi
	ld bc,FASTPAGE
	out (c),a			; seleziona il banco in cui scrivere
        and 7
        out (BORDER),a

        ld b,16384/BLOCKSIZE    	; dimensioni per uno snapshot da 16K
        ld ix,$C000	           	; ram buffer address
        call mmc_read_multidata 	; reads the card
        and a
        jp nz,restore_error
        ld a,$40
        add d                   	; adding $40 to D means 16KBytes displacement
        jr nc,no_flohl
        inc hl                  	; increase MSB pointer
no_flohl
        ld d,a

        pop af
        dec a
        cp 0
        jr nz,l_ramrestore

end_ramromrestore
        ld hl,(real_sp)         	; restores the SP
        ld sp,hl

	ld bc,FASTPAGE
        ld a,(fastp_save)
	out (c),a

	ld bc,SPECTR_MODE
        ld a,(spectrm_save)
	out (c),a

        call cls
        xor a
        ld (x_pos),a
        ld (y_pos),a
        out (BORDER),a
        ld hl,restore_done1
        call disp_text
        xor a
        ld (x_pos),a
        ld a,2
        ld (y_pos),a
        ld hl,restore_done2
        call disp_text
ranykey
        call key_scan
        ld a,e
        cp $ff                          ; no keys
        jr z,ranykey
;        call wait_keyrelease            ; waits for the key to be released
;        call cls
;        call disp_menu
;        call display_mmcdata
        jp 0

restore_done1
        db      7,"512K Ram content has been restored.",0
restore_done2
        db      6,"Press any key to restart the machine",0

restore_confirm
	db	7,"Confirm 512K Ram Restore? (Y/N)",0


no_restorebackup

;        cp MMC_FLASHBACKUP
;        jp nz,no_flashrestorebackup



	ld b,16384/BLOCKSIZE
	cp MMC_SNAP48
	jr nz,ripr_this
	ld b,49152/BLOCKSIZE

ripr_this
	cp MMC_SNAP128			; A = 8 = snapshot 128K
	jr z,restore_128sd

	ld ix,$4000			; ram buffer (restore point)
	call mmc_read_multidata		; 16K/48K restore. we can't rely on stack for temporary data
	cp 0				; error?
        jp nz,restore_error
	jr proceed_restore

restore_128sd
	ld bc,S128KREG			; 128K restore. we can't rely on stack for temporary data
	in a,(c)
	and $F8				; RAM_BANK = 0
	ld c,a				; C = RAM_BANK counter, that keeps all other bits
	ld ix,$C000			; restore buffer write address
r_128ksd
	ld a,c				; actual bank
	exx
	ld bc,S128KREG			; BC' for CPLD addressing
	out (c),a			; new ram_bank
	exx

	ld b,16384/BLOCKSIZE
	call mmc_read_multidata		; reads this 16K bank without affecting IX, HL and DE
	cp 0
	jp nz,restore_error
	ld a,$40
	add d				; adding $40 to D means 16KBytes displacement
	jr nc,no_flowhl
	inc hl				; increase MSB pointer
no_flowhl
	ld d,a

	inc c				; next ram bank
	ld a,c
	and 7
	out (BORDER),a			; shows progress
	jr nz,r_128ksd			; roll over = compleded


proceed_restore
	ld a,(BORDCR)			; colore attuale del border *8 (variabile del basic sinclair)
	rra
	rra
	rra
	and 7				; rimette a posto il colore originale
	out (BORDER),a			; gli altri bit possono essere ZERO (SPKR, TAPE)

	ld ix,(entry_addr)


        ld a,(ix+46)
        bit 7,a                         ; flag ULA+ in uso
        call nz,UlaPlus_Restore         ; se necessario, ripristina colori

	ld a,(ix+44)
	ld i,a				; ripristina lo stato del registro I

	ld l,(ix+42)			; RIPRISTINA LO STATO DELLO STACK POINTER
	ld h,(ix+43)
	ld sp,hl

        ld a,(sh_flags)                 ; indica che non siamo piu' in
        res bootrom,a                   ; bootrom (per l'NMI handler)
        ld (sh_flags),a

	; FASE RIPRISTINO REGISTRI CPLD, IM_MODE E INT ENABLE da E', D', H'

	ld d,(ix+47)			; stato del registro $24DF
	res 7,d				; azzera il bit di selezione READ ENABLE SHADOW_ROM
	ld bc,CLKREG
	in a,(c)
	and $80				; preserva lo stato attuale dei bit di selezione rd shadow_rom
	or d
	out (c),a

	ld a,(ix+58)			; stato del registro $1FFD
	ld bc,PLUS3REG
	out (c),a

	ld bc,S128KREG			; ripristina il registro $7FFD
	ld a,(ix+45)			; ora il $1FFD e $7FFD potrebbero essere lockati mediante D5 SET
	out (c),a

	ld a,(ix+59)			; stato del registro $54DF
	ld bc,FASTPAGE
	out (c),a

        ld a,(ix+61)                    ; stato del registro $64DF
        ld bc,DDR_MSB
	out (c),a

	ld a,(ix+46)			; registro F: D5, D6 da copiare; D4 --> D1; D3 --> NOT D0
	ld h,a
	and $60
	bit 3,h
	jr nz,bit3highf
	set 0,a				; copia invertito
bit3highf
	bit 4,h
	jr z,bit4lowf
	set 1,a				; copia
bit4lowf
	ld bc,SPECTR_MODE
	out (c),a			; ATTENZIONE: D6 SET blocca ogni altro accesso WR ai registri non standard!

        ld bc,PLUS3REG                  ;if SW2 is ON, adds a pause
        in a,(c)                        ;before actually launching the program.
        bit 6,a                         ;this allows any screen to be seen for
        jr z,no_pause                   ;a moment
        ld a,3
        ld bc,0
l_screen_watch
        dec c
        jr nz,l_screen_watch
        dec b
        jr nz,l_screen_watch
        dec a
        jr nz,l_screen_watch
no_pause

	di				; not really necessary, but who knows...
	im 1				; default IM1
	bit 2,h				; interrupt enable real status
	jr z,no_eneimmc
	ei				; this will execute our fake handler (EI, RETI)
	halt				; sync with the hardware maskable interrupt generator
	halt
no_eneimmc				; now we have 20 (or 10) ms before next interrupt will occur
	ld a,h				; for exiting the handler and restore the BASIC ROM.
	and 3				; if the host's int handler need the basic rom, it would crash
	cp 0
	jr z,imzerommc
	cp 1
	jr z,completedeimmc
	im 2
	jr completedeimmc
imzerommc
	im 0
completedeimmc

	exx
	pop ix
	pop hl
	pop de
	ex af,af'
	pop af
	ex af,af'
	pop iy
	pop bc
	exx

	pop de
	pop bc
	pop hl
	pop af
	jp $15				; RETN



; Subroutine che calcola il blocco da 512 bytes che contiene l'entry (file) nella FAT.
; Richiede: DE = n. del file (snapshot)
; Restituisce: HL, DE = indirizzo 32 bit settore nells SD/MMC
; C = offset 0-7 che identifica l'header desiderato (ciascuno e` 64 bytes) all'interno
; di questo blocco da 512 bytes

calc_hldec
	ld a,e
	and 7
	ld c,a				; salva i 3 bit LSB in C
	ld a,e
	and $F8				; e li elimina da DE
	ld e,a

	ld a,(card_offset)
	ld h,a
	ld l,0
	ld b,6				; moltiplica *64 (6 shift) = dimensioni di ciascuna entry
mul6
	sla e
	rl d
	rl l				; shift a 24 bit: dovendo indirizzare 512KB, coinvolge anche i 3 bit LSB di L
	djnz mul6
	ret

;********************************************************************************************
; FINE RIPRISTINO SNAPSHOT DA SD/MMC 16K/48K/128K
;
; Gestione scelta di una ROM DI BOOT
;********************************************************************************************


;
;------------------------------------------------------------------------------------------------
; Subroutine che imposta D0 del registro $34DF (spectrum_mode) come segue:
;
; SET per lo spectrum 128L
; RESET per lo spectrum 48K
;
; Il tipo di spectrum da selezionare deve essere specificato dal chiamante nel registro A,
; che contiene il numero della rom attivata dal menu: 0 per 128K, != 0 per tutte le altre.
;
; In caso di ulteriori rom che richiedano la modalita` 128K, sara` sufficiente modificare
; solo questa subroutine.
;
; DISTRUGGE BC, L.
;------------------------------------------------------------------------------------------------
spectrum_select
	ld bc,SPECTR_MODE
	in l,(c)	; stato altri bit preservato (D1 = double_int mode)
	res 0,l

	cp 0		; ALTO solo se viene scelta la ROM 0, che e` quella dello spectrum 128K
	jr nz,noromzero
	inc l
noromzero
	out (c),l
	ret

;
;------------------------------------------------------------------------------------------------
; CLS Subroutine. Inizializza con il colore indicato in ATTRIBUTE
;------------------------------------------------------------------------------------------------
cls
	ld hl,$4000
	ld bc,6144
l_clear
	ld (hl),0		; pulisce il video
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_clear

	ld bc,768		; spazio attributi
	ld a,(attribute)
	ld e,a
l_attr
	ld (hl),e
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,l_attr

;
;------------------------------------------------------------------------------------------------
; THE 'KEYBOARD SCANNING' SUBROUTINE, copiata pari-pari da quella della ROM 48K
;
; The E register is returned with a value in the range of +00 to +27, the value being different
; for each of the forty keys of the keyboard, or the value +FF, for no-key.
; The D register is returned with a value that indicates which single shift key is being pressed.
; If both shift keys are being pressed then the D and E registers are returned with the values
; for the CAPS SHIFT and SYMBOL SHIFT keys respectively. 
; If no keys is being pressed then the DE register pair is returned holding +FFFF.
; The zero flag is returned reset if more than two keys are being pressed, or neither key of a
; pair of keys is a shift key.
;
; KEY MAP FOR 'E' REGISTER:
;
;      COL 1  2  3  4 5 6  7  8  9 10
;
;ROW 1	  36 28 20 12 4 3 11 19 27 35	1234567890
;ROW 2	  37 29 21 13 5 2 10 18 26 34	QWERTYUIOP
;ROW 3	  38 30 22 14 6 1  9 17 25 33	ASDFGHJKL[ENTER]
;ROW 4	  39 31 23 15 7 0  8 16 24 32	[caps]ZXCVBNM[symbol][SPACE]
;
; DISTRUGGE AF, BC, DE, HL
;------------------------------------------------------------------------------------------------
key_scan
	ld l,$2F
	ld de,$FFFF
	ld bc,$FEFE
key_line
	in a,(c)
	cpl
	and $1F
	jr z,key_done
	ld h,a
	ld a,l
key_3keys
	inc d
	ret nz
key_bits
	sub 8
	srl h
	jr nc,key_bits
	ld d,e
	ld e,a
	jr nz,key_3keys
key_done
	dec l
	rlc b
	jr c,key_line
	ld a,d
	inc a
	ret z
	cp $28
	ret z
	cp $19
	ret z
	ld a,e
	ld e,d
	ld d,a
	cp $18
	ret


;
;------------------------------------------------------------------------------------------------
; Key-to-Address/dataline decoding subroutine
;
; This table returns the address line (A8-A15) encoded in the LSB 3 bits then data line
; (D0 - D4) in the 5 MSB bits. Useful for joystick programming
;
; To be called with KeySCAN value in E; returns result in A.
;
; Destroys AF, DE, HL
;------------------------------------------------------------------------------------------------
key_joycode
	ld a,e
	and 7			; keynumber 0-7 can be directly mapped to address number
	ld d,a
	ld a,7			; address line is equal to 7 - (keynumber & 7)
	sub d
	ld d,a			; saves address number (3 LSB bits of result)
	ld a,e
	and $F8
	rrca
	rrca
	rrca			; A = DATA LINE number (4 - 0)
	inc a
	ld c,$80		; bit set for D4 on key matrix (D7 on register, since 3 LSB bits = address line)
find_datl
	dec a
	jr z,kk_found
	rrc c
	jr find_datl
kk_found
	ld a,c
	or d
	ret


;	ld a,e
;	cp 40			; la tabella contiene 40 elementi
;	jr nc,over_j
;	ld hl,k_joytable
;	ld d,0
;	add hl,de
;	ld a,(hl)		; legge il codice
;	ret
;over_j
;	xor a			; no-key
;	ret
;
;k_joytable
;	defb	$87, $86, $85, $84, $83, $82, $81, $80	; D4 --> A15, A14, A13, A12, A11, A10, A9, A8
;	defb	$47, $46, $45, $44, $43, $42, $41, $40	; D3 --> A15, A14, A13, A12, A11, A10, A9, A8
;	defb	$27, $26, $25, $24, $23, $22, $21, $20	; D2 --> A15, A14, A13, A12, A11, A10, A9, A8
;	defb	$17, $16, $15, $14, $13, $12, $11, $10	; D1 --> A15, A14, A13, A12, A11, A10, A9, A8
;	defb	$0F, $0E, $0D, $0C, $0B, $0A, 9, 8	; D0 --> A15, A14, A13, A12, A11, A10, A9, A8


;
;------------------------------------------------------------------------------------------------
; Address/dataline to KEY encoding subroutine
;
; This table returns A with the keyscan code (0 - $27) of the address/data line joystick code
; passed in E. Returns A = $FF if key is not valid
;
; Destroys AF, DE
;------------------------------------------------------------------------------------------------
joy_keycode
	ld a,e
	and $F8			; zero on all 5 MSB bits means NO KEY
	jr nz,validkey
	ld a,$FF
	ret

validkey
	ld a,e
	and 7			; keynumber 0-7 can be directly mapped to address number
	ld d,a
	ld a,7			; address line is equal to 7 - (keynumber & 7)
	sub d

find_datk
	sla e
	ret c
	add 8			; key displacement for each DATA BIT position
	jr find_datk


;
;------------------------------------------------------------------------------------------------
; Subroutine che attende il rilascio di tutti i tasti. Distrugge AF.
;------------------------------------------------------------------------------------------------
wait_keyrelease
	push bc
	push de
	push hl
l_wrel
	call key_scan
	jr nz,l_wrel
	ld a,e
	cp $FF
	jr z,no_key
	cp 24			; symbol shift non viene considerato un tasto, rientriamo
	jr z,no_key
	cp 39			; stessa cosa per il caps shift
	jr nz,l_wrel
no_key
	pop hl
	pop de
	pop bc
	ret

;
;------------------------------------------------------------------------------------------------
; Key table subroutine. Returns a single, decoded character in A register based on
; keyscan code in DE register.
;
; NEVER WRITES to any ram variable, since we may have been called in NMI for snapshot purposes
;
; RETURNS:
; 0 = no valid keycode
; 1 = caps shift pressed (alone)
; 2 = symbol shift pressed (alone)
; 8 = BACKSPACE (caps_shift + '0')
; 13 = ENTER
; ASCII code (letters and numbers, capitals if caps shift pressed as well)
; Symbol ascii codes if symbol shift is pressed as well
;
; Destroys AF, D, HL
;------------------------------------------------------------------------------------------------
key_decode
	ld a,d
	cp 24			; D = 24 = symbol shift premuto insieme ad un altro tasto
	jr z,symbolshift	; (se premuto da solo, il suo codice viene letto da tabella ascii)

	ld a,e
	cp 40
	jr c,validk		; salta se E e` entro 0-39 (tasto premuto)
	ld a,d
	cp 39			; D = 39 = caps shift premuto
	jr z,capsshift
	xor a
	ret
capsshift
	ld a,1			; rientra con il codice del caps shift
	ret

validk
	push de			; salva lo stato del caps shift
	ld hl,asciitable
	ld d,0
	add hl,de
	ld a,(hl)		; legge il codice ascii
	pop de			; notare che se il caps shift e` premuto da solo, il suo codice
	push af			; verra` letto dalla tabella
	ld a,d
	cp 39			; D = 39 = caps shift premuto
	jr nz,nocaps		; non e` premuto: rientra con il codice letto da tabella
	pop af
	cp '0'
	jr nz,nobackspa
	ld a,8			; 0 + CAPS LOCK = BACKSPACE
	ret
nobackspa
	cp 'a'
	ret c			; rientra se non e` un carattere alfabetico
	sub 'a'-'A'		; altrimenti lo trasforma in maiuscolo
	ret

nocaps
	pop af
	ret

symbolshift
	ld a,e			; c'e' un altro tasto premuto contemporaneamente?
	cp 40
	jr c,altrota		; si, salta (siamo entro la tabella, quindi non e` $FF)
	ld a,2			; altrimenti rientra con codice ASCII $02 (lo usiamo per SYMBOL SHIFT)
	ret
altrota
	ld hl,symboltable	; trova il codice ascii del SIMBOLO sul tasto premuto
	ld d,0
	add hl,de
	ld a,(hl)
	ret


asciitable
	defb	'b','h','y','6','5','t','g','v'	; tasti relativi ai codici da 0 a 7
	defb	'n','j','u','7','4','r','f','c'	; 8 - 15
	defb	'm','k','i','8','3','e','d','x'	; 16 - 23
	defb	2,'l','o','9','2','w','s','z'	; 24 - 31 (0 = symbol shift)
	defb	' ',13,'p','0','1','q','a',1	; 32 - 39 (0 = caps shift), 13 = enter

symboltable
	defb	'*','^','y','&','%','>','g','/'	; tasti relativi ai codici da 0 a 7
	defb	',','-','u',39,'$','<','f','?'	; 8 - 15
	defb	'.','+','i','(','#','e','d','x'	; 16 - 23
	defb	0,'=',';',')','@','w','s',':'	; 24 - 31 (0 = symbol shift)
	defb	' ',13,'"','_','!','q','a',0	; 32 - 39 (0 = caps shift, 13 = enter)


;
;------------------------------------------------------------------------------------------------
; Subroutine that returns the KEY NUMBER (1 - 40) of the pressed key, or 0 for no key.
;
; Destroys AF
;------------------------------------------------------------------------------------------------
key_number
	ld a,e
	cp 40			; la tabella contiene 40 elementi
	jr nc,over
	push de
	push hl
	ld hl,keyposition
	ld d,0
	add hl,de
	ld a,(hl)
	pop hl
	pop de
	ret
over
	xor a			; no-key
	ret

keyposition
	defb	36, 26, 16, 6, 5, 15, 25, 35
	defb	37, 27, 17, 7, 4, 14, 24, 34
	defb	38, 28, 18, 8, 3, 13, 23, 33
	defb	39, 29, 19, 9, 2, 12, 22, 32
	defb	40, 30, 20, 10, 1, 11, 21, 31



;
;------------------------------------------------------------------------------------------------
; Subroutine di visualizzazione della stringa puntata da HL
;
; INPUT:
; HL = puntatore alla stringa ASCIIZ
; X_POS = coordinata X di visualizzazione, espressa in PIXEL (0 - 255)
; Y_POS = coordinata Y di visualizzazione, espressa in CARATTERI (0 - 23)
;
; Chiama COMPUTE_ADD per calcolare il contenuto di WR_VIDEO in base a X_POS e Y_POS
; Chiama DISP_CH per visualizzare ciascun carattare
;
; OUTPUT:
; X_POS, Y_POS aggiornati alla coordinata del prossimo carattere da visualizzare
;
; DISTRUGGE AF, DE, HL.
;------------------------------------------------------------------------------------------------
disp_text
	push iy
	push bc
	call compute_add	; calcola l'indirizzo di inizio scrittura in ram video in base a X_POS, Y_POS
	ld a,(hl)
	ld (attribute),a	; primo byte = attributi
	inc hl
l_text
	ld a,(hl)
	cp 0
	jr z,f_disptext		; fine lavoro
	call disp_ch		; visualizza il carattere
	inc hl
	jr l_text
f_disptext
	pop bc
	pop iy
	ret			; rientra con HL che punta al NULL byte


;------------------------------------------------------------------------------------------------
; Subroutine che calcola WR_VIDEO in base al contenuto di X_POS e Y_POS
;
; INPUT:
; X_POS = coordinata X di visualizzazione, espressa in PIXEL (0 - 255)
; Y_POS = coordinata Y di visualizzazione, espressa in CARATTERI (0 - 23)
;
; OUTPUT:
; WR_VIDEO puntera` al primo byte in cui scrivere; il numero di bit gia` occupati in questo
; byte e` (era gia`) espresso nei 3 bit LSB di X_POS, che infatti non prendono parte alla
; costituzione di WR_VIDEO
;
; DISTRUGGE AF, DE
; PRESERVA  HL
;------------------------------------------------------------------------------------------------
compute_add
	push hl
	ld hl,$4000		; indirizzo base VIDEO RAM

	ld a,(y_pos)		; l'offset per ciascun blocco di 8 righe e` $800
	and $F8
	ld d,a			; lo somma quindi direttamente ad HL, sul byte MSB
	ld e,0			; e dopo aver mascherato i 3 bit LSB
	add hl,de

	ld a,(y_pos)
	and 7			; i 3 bit LSB valgono offset +32 ciascuno
	rrca			; moltiplica *5
	rrca
	rrca
	ld e,a
	ld d,0
	add hl,de		; somma

	ld a,(x_pos)		; i 5 bit MSB costituiscono l'offset di COLONNA in questa riga
	srl a
	srl a			; shifta a destra 3 volte (offset 0 - 31)
	srl a
	ld e,a			; D e` gia` ZERO
	add hl,de		; somma offset
	ld (wr_video),hl


	ld hl,$5800		; attribute base address
	ld a,(y_pos)		; il valore unitario di Y_POS corrisponde ad un offset di +32 bytes
	ld e,a			; negli attributi
	ld d,0
	ld b,5
l_muldd
	sla e			; moltiplica *5 a 16 bit
	rl d
	djnz l_muldd
	add hl,de		; somma l'offset

	ld a,(x_pos)		; i 5 bit MSB costituiscono l'offset di COLONNA in questa riga
	srl a
	srl a			; shifta a destra 3 volte (offset 0 - 31)
	srl a
	ld e,a
	ld d,0
	add hl,de		; somma offset
	ld (wr_attribute),hl

	pop hl
	ret


;------------------------------------------------------------------------------------------------
; Subroutine di visualizzazione del carattere in 'A'
;
; INPUT:
; WR_VIDEO deve puntare al primo byte in ram video in cui scrivere;
; X_POS, che contiene la coordinata X espressa IN PIXEL (ovvero 0 - 255) viene utilizzato come
; indicatore di bit OCCUPATI nel byte puntato da WR_VIDEO, dato espresso dai suoi 3 bit LSB;
; (i 5 MSB sono gia` stati integrati in WR_VIDEO ad opera di COMPUTE_ADD)
;
; Y_POS (qui non utilizzata) contiene la coordinata Y espressa in CARATTERI (0-23).
; Viene usata da COMPUTE_ADD per calcolare WR_VIDEO.
;
; OUTPUT:
; Il carattere viene scritto nella ram video; X_POS incrementato del n. di bit costituenti la
; larghezza del carattere. Se necessario, ovvero se siamo andati oltre il byte correntemente
; puntato da WR_VIDEO, esso viene incrementato in modo da non richiedere altri calcoli in caso di
; chiamate consecutive (per visualizzare una stringa).
;
; DISTRUGGE AF, BC, DE, IY.
;------------------------------------------------------------------------------------------------
disp_ch
	push hl
	push ix
	sub ' '			; SPAZIO = carattere 0 nel font
	ld e,a			; carattere in DE
	ld d,0
	sla e			; moltiplica per 8 (dimensioni MAX di un carattere nel font)
	rl d
	sla e
	rl d
	sla e
	rl d
	ld iy,font_block	; indirizzo base FONT
	add iy,de

	ld d,$FF		; carica in DE una maschera di AND 16 bit utile ad azzerare i pixel
	ld b,CHAR_X		; sui quali verra` caricato il bitmap del font. Per un font largo 6,
ldefinemask			; DE conterra` $03FF (ovvero 6 bit MSB azzerati, gli altri tutto SET).
	srl d
	djnz ldefinemask
	ld e,$FF

	ld a,(x_pos)		; posizione X di scrittura (in pixel)
	and 7			; 3 bit LSB = bit da non utilizzare nel presente byte
	cp 0
	jr z,mask_done
	ld b,a
l_andmask
	scf			; sposta la maschera AND in modo corrispondente al pattern da caricare
	rr d
	rr e
	djnz l_andmask
mask_done
	push de
	pop ix			; IX = maschera AND

	ld b,8			; n. linee da caricare
	ld hl,(wr_video)	; indirizzo prima linea da caricare

next_row
	push af			; salva n. bit gia` occupati nel byte corrente
	ld d,(iy)		; DE = bitmap della riga di pixel corrente per questo carattere
	ld e,0

l_sh1
	cp 0			; A contiene il numero di pixel gia` usati nel byte corrente
	jr z,shiftdone
	srl d
	rr e			; shift a 16 bit verso destra del pattern da caricare

	dec a			; alla fine la maschera a 16 bit conterra` una "finestra" di bit RESET
	jr l_sh1		; nella posizione in cui il nuovo pattern andra` caricato

shiftdone
	ld a,(hl)		; byte in ram video
	and ixh			; maschera AND, azzera i bit da modificare
	or d			; inserisce i bit dal bitmap font
	ld (hl),a

	ld a,ixl
	cp 255			; se LSB maschera != 255 abbiamo una parte eccedente da scrivere
	jr z,nosecondo		; nel byte successivo, altrimenti salta.

	inc hl			; scrive parte eccedente nel byte alla destra di quello precedente
	ld a,(hl)
	and ixl
	or e
	ld (hl),a
	dec hl

nosecondo
	inc h			; la riga successiva si trova ad un offset +256
	inc iy			; nuova riga nel font
	pop af
	djnz next_row		; procede per 8 righe

	ld a,(attribute)
	ld hl,(wr_attribute)
	ld (hl),a
	ld c,a
	ld a,ixl
	cp 255			; se LSB maschera != 255 dobbiamo settare anche l'attributo per il byte successivo
	jr z,nos_attr
	ld a,c
	inc hl
	ld (hl),a

nos_attr
	ld a,(x_pos)		; caricamento completato:
	add CHAR_X		; aggiunge larghezza del carattere appena visualizzato alla coordinata
	ld (x_pos),a		; X del prossimo carattere, i cui 3 bit LSB indicano il n. di bit occupati
	cp 255-CHAR_X		; non c'e' posto per un ulteriore carattere in questa riga?
	jr c,sispace		; si, salta

	xor a			; altrimenti azzera coordinata X
	ld (x_pos),a
	ld a,(y_pos)
	inc a			; passa alla riga sottostante (inserire un controllo di SCROLL!)
	ld (y_pos),a
	call compute_add	; occorre ricalcolare l'indirizzo
	jr norepp

sispace
	and 7			; se necessario aggiorna WR_VIDEO, in modo da evitare la necessita`
	cp CHAR_X		; di richiamare COMPUTE_ADD
	jr nc,norepp
	ld hl,wr_video
	inc (hl)
	ld hl,wr_attribute
	inc (hl)
norepp
	pop ix
	pop hl
	ret	



;*************************************************************************************************
;*************************************************************************************************
; MMC READ/WRITE ROUTINES that can operate @21.25MHz processor clock.
;
; RD/WR are on 512 bytes boundary ONLY.
; MULTIPLE_BLOCK READ is supported, while by now MULTIPLE_BLOCK WR is not
;
;*************************************************************************************************
;*************************************************************************************************

;-------------------------------------------------------------------------------------------
; This subroutine should be called at power-on or when a MMC has been inserted.
; It tries to get the MMC CID (writing at the provided HL pointer) and, in case of failure,
; it calls the INIT procedure then tries again.
;
; Returns A = 0 if OK, or 1 = error (no MMC found or MMC error)
;
; iF != 0 AND != $ff, the mmc_get_cid returned error code is displayed on screen.
;-------------------------------------------------------------------------------------------
mmc_getcid_poweron
	call mmc_get_cid		; try to read the MMC CID INFO --> (HL)
	cp 0				; if it fails, then che SD/MMC needs to be initialized
	ret z				; to SPI mode (MMC_INIT) communications (once after power-on)

	cp 255				; card is probably not in SPI mode
	jr z,need_init

	bit 0,a				; IDLE/still initializing bit
	jr nz,need_init

	ld c,a				; unknown response: displays the code and exits
	push bc
	ld hl,mmcreturned
	xor a
	ld (x_pos),a
	ld a,23
	ld (y_pos),a
	call disp_text
	pop bc
	call disp_hex			; displays HEX code in C

	ld a,1				; error
	ret

need_init
	call mmc_init
	cp 0
	ret nz				; INIT error: MMC not detected.

	ld de,BLOCKSIZE
	call mmc_send_blocksize
	jr mmc_getcid_poweron

mmcreturned
	db	2,"MMC send_cmd returned  $",0

;
;-----------------------------------------------------------------------------------------
; READ MULTIPLE BLOCK OF DATA TEST subroutine
;
; This routine only works for blocksize = 512 (two INI sequence).
;
; HL, DE= MSB, LSB of 32bit address in MMC memory
; B	= number of 512 bytes blocks to be read
; IX	= ram buffer address
;
; RETURN code in A:
; 0 = OK
; 1 = read_block command error
; 2 = no wait_data token from MMC
;
; DESTROYS AF, B
;-----------------------------------------------------------------------------------------
mmc_read_multidata
	ld a,MMC_READ_MULTIPLE_BLOCK	; Command code for multiple block read
	call mmc_send_command
	cp 0
	ret nz				; ERRORE
	push hl				; HL should be saved

	push ix
	pop hl				; INI usa HL come puntatore
jrhere
	call mmc_waitdata_token
	cp $FE
	jr z,read_mmc_multiblock	; OK
	ld a,2				; no data token from MMC
	pop hl
	ret

read_mmc_multiblock
	push bc
	ld bc,SPI_PORT			; B = 0 = 256 bytes for the first INI; C = I/O port


	if USE_INI = 0
	inir
	nop
	inir
	endif

	if USE_INI = 1
ini_loop1
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	jr nz,ini_loop1
ini_loop2
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	ini
	jr nz,ini_loop2
	endif

	nop
	nop

	in a,(SPI_PORT)			; CRC
	nop
	nop
	in a,(SPI_PORT)			; CRC

	pop bc
	djnz jrhere

	ld a,MCC_TERMINATE_MULTI_READ
	call mmc_write_command

	in a,(SPI_PORT)			; CRC?
	nop
	nop
	in a,(SPI_PORT)

	call mmc_wait_response		; waits for the MMC to reply "0"
	ld b,a				; error code

	call cs_high			; set cs high
	call clock32			; 32 more clock cycles
	call clock32			; 32 more clock cycles
	call clock32			; 32 more clock cycles
	call clock32			; 32 more clock cycles

	pop hl
	ld a,b
	ret


;
;-----------------------------------------------------------------------------------------
; READ A SINGLE BLOCK OF DATA subroutine
;
; This routine only works for blocksize = 512 (two INIR sequence).
;
; HL, DE= MSB, LSB of 32bit address in MMC memory
; IX	= ram buffer address
;
; RETURN code in A:
; 0 = OK
; 1 = read_block command error
; 2 = no wait_data token from MMC
;
; DESTROYS AF, HL
;-----------------------------------------------------------------------------------------
mmc_read_data

	ld a,MMC_READ_SINGLE_BLOCK	; Command code for block read
	call mmc_send_command
	cp 0
	ret nz				; ERRORE

	call mmc_waitdata_token
	cp $FE
	jr z,read_mmc_block		; OK
	ld a,2				; no data token from MMC
	ret

read_mmc_block
	push ix
	pop hl				; HL = INIR write pointer

	push bc
	ld bc,SPI_PORT			; B = 0 = 256 bytes for the first INIR; C = I/O port

	inir
	nop
	inir

	nop
	nop

	in a,(SPI_PORT)			; CRC
	nop
	nop
	in a,(SPI_PORT)			; CRC

	call cs_high			; set cs high
	call clock32			; 32 more clock cycles

	pop bc

	xor a				; no error
	ret


clock32
	push bc
	ld b,4
l_4bytes
	in a,(SPI_PORT)			; some more clock cycles
	djnz l_4bytes
	pop bc
	ret


;-----------------------------------------------------------------------------------------
; READ FIRST 2 BYTES OF SD/MMC CARD and copies them to DE.
;
; RETURN code in A:
; 0 = OK
; 1 = read_block command error
; 2 = no wait_data token from MMC
;
; DESTROYS AF, DE
;-----------------------------------------------------------------------------------------
mmc_get_firstword
	push hl
	ld a,(card_offset)
	ld h,a
	ld l,0
	ld de,0
	ld a,MMC_READ_SINGLE_BLOCK	; Command code for block read
	call mmc_send_command
	pop hl
	cp 0
	ret nz				; ERRORE
	call mmc_waitdata_token
	cp $FE
	jr z,read_mmc_first		; OK
	ld a,2				; no data token from MMC
	ret

read_mmc_first
	in a,(SPI_PORT)
	ld e,a
	nop
	in a,(SPI_PORT)
	ld d,a

	push bc
	ld bc,BLOCKSIZE-2		; reads the remaining 510 bytes and throws them
read_firstword
	in a,(SPI_PORT)
	dec bc
	ld a,b
	or c
	jr nz,read_firstword
	pop bc

	in a,(SPI_PORT)			; CRC
	nop
	nop
	in a,(SPI_PORT)			; CRC

	call cs_high			; set cs high
	call clock32			; 32 more clock cycles

	xor a
	ret

;
;-----------------------------------------------------------------------------------------
; WRITE BLOCK OF DATA subroutine. By now, we don't use the MULTIPLE_BLOCK transfer.
;
; This routine only works for blocksize = 512 (two OUTI sequence).
;
; HL, DE= MSB, LSB of 32bit address in MMC memory
; B	= number of 512 bytes blocks to write
; IX	= ram buffer address
;
; RETURN code in A:
; 0 = OK
; 1 = read_block command error
; 2 = write error (no "5" response from MMC)
;
; Destroys AF, B, DE, HL, IX.
;-----------------------------------------------------------------------------------------
mmc_write_data
	ld a,MMC_WRITE_SINGLE_BLOCK	; Command code for block read
	call mmc_send_command
	cp 0
	ret nz				; ERRORE

	ld a,$FE
	out (SPI_PORT),a		; first byte to be sent = DATA TOKEN

write_mmc_block
	push bc
	ld bc,SPI_PORT			; C = PORT, B = 0 for 256 bytes on first OTIR
	push hl
	push ix
	pop hl

	otir
	nop				; without this NOP, the 2nd half of the 512 bytes block would miss some write cycles
	otir				; on the CPLD if Z80 runs at 21MHz

	push hl
	pop ix
	pop hl

	ld a,$95			; CRC... (!)
	out (SPI_PORT),a
	nop
	nop
	out (SPI_PORT),a

	call mmc_wait_response

	pop bc

	and $1F				; masks useful response bits
	cp 5
	jr nz,write_error

	ld a,d				; aggiorna il puntatore al blocco da leggere
	add BLOCKSIZE/256		; 2 se blocksize = 512
	jr nc,no_overw
	inc hl
no_overw
	ld d,a

wait_busy
	call mmc_wait_response		; MMC will report "00" until busy
	cp 0
	jr z,wait_busy

	call cs_high
	call clock32			; 32 more clock cycles

	djnz mmc_write_data		; next block

	xor a
	ret

write_error
	ld a,2				; write error code
	ret


;
;-----------------------------------------------------------------------------------------
; MMC SPI MODE initialization. RETURNS ERROR CODE IN A register:
;
; 0 = OK
; 1 = Card RESET ERROR
; 2 = Card INIT ERROR
;
; Destroys AF, B.
;-----------------------------------------------------------------------------------------
mmc_init
	call	cs_high			; set cs high
	ld	b,10			; sends 80 clocks
	ld a,$FF
l_init
	out (SPI_PORT),a
	djnz l_init
	nop
	
	call	cs_low			; set cs low
	
	ld a,MMC_GO_IDLE_STATE
	call mmc_write_command

	call mmc_wait_response
	cp $01				; MMC should respond 01 to this command
	jr nz,mmc_reset_failed		; fail to reset

	ld bc,120			; retry counter*256 (corrisponde a circa 5 secondi @3.5MHz)
mmc_reset_ok
	call cs_high			; set cs high
	ld a,$FF
	out (SPI_PORT),a		; 8 extra clock cycles
	nop
	nop
	
	call	cs_low			; set cs low
	
	ld a,MMC_SEND_OP_COND		; Sends OP_COND command
	call mmc_write_command

	call mmc_wait_response		; MMC_WAIT_RESPONSE tries to receive a response reading an SPI

	bit 0,a				; D0 SET = initialization still in progress...
	jr z,mmc_init_ok

	djnz mmc_reset_ok		; if no response, tries to send the entire block 254 more times
	dec c
	jr nz,mmc_reset_ok

	ld a,2				; error code for INIT ERROR
	jr mmc_errorx

mmc_init_ok
	call cs_high			; set cs high
	in a,(SPI_PORT)			; some extra clock cycles
	call pause1
	xor a
	ret
	
mmc_reset_failed			; MMC Reset error
	ld	a,1
mmc_errorx
	call	cs_high
	ret


;debug_display				; TEST: displays the code in A at current cursor position
;	push af				; ('A' based): A = 0x00, B = 0x01, etc.
;	push bc
;	push de
;	push hl
;	add 'A'
;	call disp_ch
;	ld bc,0
;lll
;	djnz lll
;	dec c
;	jr nz,lll
;	pop hl
;	pop de
;	pop bc
;	pop af
;	ret


;
;-----------------------------------------------------------------------------------------
; READ MMC CID subroutine. Data is stored at address (HL).
;
; Returns error code in A register:
; 0 = no error
; 1 = Read CID command error
; 2 = no wait_data token from MMC
;
; Destroys AF, HL.
;-----------------------------------------------------------------------------------------
mmc_get_cid
	push bc
	push de
	push hl
	ld hl,0
	ld de,0
	ld a,MMC_READ_CID
	call mmc_send_command
	pop hl
	cp 0
	jr z,cmd_cidok			; MMC_SEND_COMMAND reports 0 (ok) or MMC error code
	push af				; error code on STACK
	jr cid_exit
cmd_cidok
	call mmc_waitdata_token
	cp $FE
	jr z,waitda_cid_ok
	ld a,2
	push af
	jr cid_exit

waitda_cid_ok
	ld b,18				; 16 bytes + CRC?
	ld c,SPI_PORT

	inir

mmc_cidl
	xor a
	push af
cid_exit
	call cs_high			; set cs high
	in a,(SPI_PORT)
	pop af				; error code
	pop de
	pop bc
	ret


;
;-----------------------------------------------------------------------------------------
; SEND COMMAND TO MMC subroutine
;
; A = COMMAND CODE;
; H, L, D, E = 32 bit parameter (MSB ... LSB);
;
; Sends a $FF fake checksum
;
; RETURNS: 0 = OK; != 0 = MMC error code
;
; On OK, the CHIP SELECT will be LOW on exit
; On error, the CHIP SELECT will be deasserted on exit
;
; Destroys AF.
;-----------------------------------------------------------------------------------------
mmc_send_command
	push bc
	ld c,a
	call cs_high			; cs high
	call clock32
	nop
	call cs_low			; cs low
	ld a,c				; command code is the first byte to be sent
	out (SPI_PORT),a
	ld a,h
	nop
	out (SPI_PORT),a
	ld a,l
	nop
	out (SPI_PORT),a
	ld a,d
	nop
	out (SPI_PORT),a
	ld a,e
	nop
	out (SPI_PORT),a
	ld a,$ff
	nop
	out (SPI_PORT),a

	call mmc_wait_response		; waits for the MMC to reply != $FF
	cp 0
	jr nz,mmc_commande
	pop bc
	ret				; 0 = no error

mmc_commande
	push af				; saves the error code
	call cs_high			; set cs high
	in a,(SPI_PORT)
	pop af
	pop bc
	ret				; returns the error code got from MMC


;
;-----------------------------------------------------------------------------------------
; WAIT FOR DATA TOKEN ($FE) subroutine (calls MMC_WAIT_RESPONSE up to 256 times)
; Returns with A = code read from MMC (ok if $FE). Destroys AF.
;-----------------------------------------------------------------------------------------
mmc_waitdata_token
	push bc
	ld b,10				; retry counter
mmc_waitl
	call mmc_wait_response
	cp $FE				; waits for the MMC to reply $FE (DATA TOKEN)
	jr z,exit_wda
	cp $FF				; but if not $FF, exits immediately (error code from MMC)
	jr nz,exit_wda
	djnz mmc_waitl
exit_wda
	pop bc
	ret


;
;-----------------------------------------------------------------------------------------
; SENDS BLOCK_SIZE COMMAND TO MMC.
;
; Parameters: DE = nbytes. Destroys AF.
;-----------------------------------------------------------------------------------------
mmc_send_blocksize
	call	cs_low			; set cs low
	ld a,MMC_SET_BLOCK_SIZE
	out (SPI_PORT),a
	xor a
	nop
	out (SPI_PORT),a
	nop
	nop
	out (SPI_PORT),a
	ld a,d
	nop
	out (SPI_PORT),a
	ld a,e
	nop
	out (SPI_PORT),a
	ld a,$ff			; fake checksum
	nop
	out (SPI_PORT),a
	nop
	nop

	in a,(SPI_PORT)
	nop
	nop
	in a,(SPI_PORT)
	nop
	nop
	call	cs_high
	
	ret


;
;-----------------------------------------------------------------------------------------
; Sends a command with parameters = 00 and checksum = $95. Destroys AF.
;-----------------------------------------------------------------------------------------
mmc_write_command
	push bc
	out (SPI_PORT),a	; sends the command
	ld b,4
	xor a
l_sendc0
	out (SPI_PORT),a	; then sends four "00" bytes (parameters = NULL)
	djnz l_sendc0
	ld a,$95		; $95 is only needed when the CARD INIT is being performed,
	nop
	out (SPI_PORT),a	; then this byte is ignored.
	pop bc
	ret


;
;-----------------------------------------------------------------------------------------
; Waits for the MMC to respond. Returns with A = response code; $FF = NO RESPONSE
;
; Responses from CARD are in R1 format for all command except SEND_STATUS.
; When SET, a bit indicates:
;
; D0 = Idle state / init not completed yet
; D1 = Erase Reset
; D2 = Illegal Command
; D3 = Com CRC Error
; D4 = Erase Sequence Error
; D5 = Address Error
; D6 = Parameter Error
; D7 = ALWAYS LOW
;
; Destroys AF.
;-----------------------------------------------------------------------------------------
mmc_wait_response
	push bc
	ld bc,50		; retry counter
l_response
	in a,(SPI_PORT)		; reads a byte from MMC
	cp $FF			; $FF = no card data line activity
	jr nz,resp_ok
	djnz l_response
	dec c
	jr nz,l_response
resp_ok
	pop bc
	ret


;
;------------------------------------------------------------------------------------
; ALL CHIP_SELECT HIGH subroutine. Destroys no registers.
; Keeps RS-232 output (D2) HIGH too.
;------------------------------------------------------------------------------------
cs_high
	push af
	ld a,7
	out (OUT_PORT),a
	pop af
	ret

	
;
;------------------------------------------------------------------------------------
; CHIP_SELECT 0/1 (D0/D1) LOW subroutine (based on MMCID content). Destroys no registers.
; Keeps RS-232 output (D2) HIGH
;------------------------------------------------------------------------------------
cs_low
	push af
	ld a,(cs_outreg)
	out (OUT_PORT),a
	pop af
	ret


pause1
	push hl
	ld hl,$8000		; OK for 3.5MHz ONLY (7MHz requires two calls or ld hl,0 and so on)
loop3	dec hl
	ld a,h
	or l
	jr nz,loop3
	pop hl
	ret

;
;------------------------------------------------------------------------------------
; RTC data reception from PIC microcontroller. INT are disabled and frequency is
; set to 14MHz. Frequency and INT status is restored on exit.
;
; HL should point to a 7 bytes buffer that will be filled with data.
; An error code is returned in A. 0 = no error.
;
; Destroys AF, BC, DE, HL
;------------------------------------------------------------------------------------
rtc_enquire

rtc_getdata
	xor a
	ld (hl),a		;seconds
	inc hl
	ld (hl),a		;minutes
	inc hl
        ld a,22
	ld (hl),a		;hour
	inc hl
	ld a,7
	ld (hl),a		;day of week
	inc hl
        ld a,28                 ;date
	ld (hl),a		
	inc hl
        ld a,08                 ;month
	ld (hl),a
	inc hl
        ld a,11                 ;year
	ld (hl),a
	inc hl
	xor a
	ret


;
;------------------------------------------------------------------------------------
; Subroutine that cleans the entire joystick KEYMAPPING
;------------------------------------------------------------------------------------
joystick_reset
	ld hl,tmp_str
	xor a
	ld b,10
l_makebuff
	ld (hl),a
	inc hl
	djnz l_makebuff
	ld hl,tmp_str


; continues in JOY_PROGRAM
;
;------------------------------------------------------------------------------------
; Subroutine that sends 10 bytes of joystick programming data to the FPGA.
; HL should point to the data buffer
;
; After enabling the joy_write bit in SPECTR_MODE register, joystick keymapping
; FPGA registers (16 inputs *2 bytes = 32 bytes) are mapped in Z80 MEMORY
; using only A4:A0 address lines. This means that they can be written at
; any place on 32 bytes boundaries. This allows keeping a copy of joystick
; programming into a properly aligned buffer.
;
; This routine does not keep any copy of data. To save memory space, the
; FPGA registers are accessed on $4000 (video memory), which will be erased
; during snapshot restore (to be done later!)
;
; Destroys AF, BC, DE, HL
;------------------------------------------------------------------------------------
joy_program
;        di
        ld a,$B7                ;'B' on D7:D4 enables the J_WRITE signal
        out (OUT_PORT),a

        ld e,PJOY_SIZE          ;number of mapped keys
        ld b,h
        ld c,l
        ld hl,16384
l_joydecode
        ld a,(bc)               ;D2:D0 = address line to be used (A15-A8)
        and 7
        ld d,1
l_bitdecode
        cp 0
        jr z,bitfound
        sla d                   ;turns it into a single SET bit (D0 for A8)
        dec a
        jr l_bitdecode
bitfound
        ld (hl),d               ;address (COLUMN) write to FPGA
        inc hl
        ld a,(bc)               ;D7:D3 = ROW (five key rows)
        rrca
        rrca
        rrca
        and $1F                 ;right aligned
        ld (hl),a               ;row write to FPGA
        inc hl
        inc bc
        dec e
        jr nz,l_joydecode
        ld a,7                  ;disables the J_WRITE signal
        out (OUT_PORT),a
	ret


;
;------------------------------------------------------------------------------------
; BOOTROM SYSTEM CALLS FROM BASIC (or any other ROM).
;
; ULA3 will switch to CONTEXT MODE when code is executed from location $25 (that
; contains a 'JP bootrom_services').
;
; The caller should create a buffer containing the code of the requested operation and
; any needed parameter. The address of this buffer should then be specified in ram
; locations 23728-23729. The service routine will concatenate any response code or
; data to this buffer.
;
; Functions that should be called with minimum effort from basic programs are defined
; with 23729 location = 0, which could not be a valid ram address for the data buffer.
;
; Functions with "0" in byte 23729:
; 23728 = 0: NOP. Border is turned BLUE.
; 23728 = 1: Load next stapshot: the snapshot that follows the last restored will be loaded.
;
; For "regular" functions, the following rules will apply:
; The first byte [0] of the buffer is the FUNCTION CODE, as follows:
;
; Supported functions:
;
; 00	NOP: turns blue the border, then exit. Same if 23729 = $0000
;	Returns: 0
;
; 01	JUMP to BOOTROM (as with reset)
;
; 02	Transmit a string to to the PIC microcontroller
;	[1-2] = address of string
;	[3-4] = bytes to be sent (1-65535)
;	Returns: 0
;
; 03	Receive N bytes from the PIC microcontroller
;	[1-2] = address of buffer
;	[3-4] = bytes to wait for (1-65535)
;	[5]   = timeout between bytes. 0 = no timeout; 5 = 1 second timeout
;	Returns: [3-4] = received bytes number (0 = all); [5] = error code
;
; 04	Transmit a string to to the main RS-232 115K2baud port
;	[1-2] = address of string
;	[3-4] = bytes to be sent (1-65535)
;	Returns: 0
;
; 05	Receive N bytes from the main RS-232 115K2baud port
;	[1-2] = buffer address
;	[3-4] = bytes to wait for (1-65535)
;	[5] = timeout between bytes. 0 = no timeout; 5 = 1 second timeout
;	Returns: [3-4] = received bytes number (0 = all); [5] = error code;
;	Error code: D0 = overrun; D1 = framing error; D2 = timeout
;
; 06	Ask RTC DATE and TIME from PIC microcontroller
;	Returns: [0] 0 = OK; 6 = some error occurred
;	[1-7] = data from RTC chip in binary format (SS:MM:HH, GG DD:MM:YY)
;
; 07	Reads 64 bytes of FAT entry for a SD/MMC snapshot (or raw block of data)
;	[1-2] = buffer address
;	[3-4] = snapshot number
;	Returns: 0 (in BC) if successful
;	This function also sets the internal variable SNAP_TORESTORE, which is used
;	by functions 09 and 10 to access a 128KBytes block in raw mode
;
; 08	Writes 64 bytes of FAT entry for a SD/MMC snapshot (or raw block of data)
;	[1-2] = buffer address
;	[3-4] = snapshot number
;	Returns: 0 (in BC) if successful, SD card updated.
;	This function also sets the internal variable SNAP_TORESTORE, which is used
;	by functions 09 and 10 to access a 128KBytes block in raw mode
;
; 09	Reads a block of RAW DATA in the 128KB SD/MMC memory space pointed by the
;	internal variable SNAP_TORESTORE, which is set by functions 06 and 07.
;	[1-2] = buffer address
;	[3]   = read offset from SD/MMC card, in 512 bytes chunks (1 = 512 bytes)
;	[4]   = number of 512-bytes blocks to be read and returned into buffer
;	Returns: 0 (in BC) if successful
;
; 10	Writes a block of RAW DATA in the 128K SD/MMC memory space pointed by the
;	internal variable SNAP_TORESTORE, which is set by functions 06 and 07.
;	It's advisable to mark the snapshot 128K block with "CODE 03", which means
;	"RAW DATA", in order to prevent the BOOTROM firmware from trying to restore
;	it as a snapshot. Raw data blocks are displayed in white in main menu`.
;	[1-2] = buffer address
;	[3] = write offset in the SD/MMC card, in 512 bytes chunks (1 = 512 bytes)
;	[4] = number of 512-bytes blocks to write
;	Returns: 0 (in BC) if successful; $FF if block is not of proper type (3)
;
; 11	Reads one or more blocks of HI-RES video RAW DATA in the scan_converter RAM,
;	reading from snapshot number specified in the internal variable SNAP_TORESTORE,
;	which is set by functions 06 and 07.
;	Returns: nothing, but increments the internal current snapload pointers.
;	[1] = number of subsequent images to load.
;
;------------------------------------------------------------------------------------
bootrom_services
	ld hl,(23728)
	ld a,h
	cp 0
	jr nz,command_decode

	ld a,l			; simple 2-bytes commands: L = command's code
	cp 0
	jr z,no_operation
	cp 1			; 1 = restore next snapshot
	jr nz,no_cp1
	ld hl,(snap_torestore)	; HL = snapshot currently restored
	inc hl			; we will load next one
	jp read_snap		; performs the operation

no_cp1
no_operation
	ld a,1			; funzione ZERO: border BLU e rientro immediato (test)
	out ($FE),a
	ld bc,0			; restituisce zero
	jp $14


command_decode
	ld a,(hl)		; requested operation code
	inc hl			; HL = first parameter byte
	cp 0
	jr nz,no_ff0


;------------------------------------------------------------------------------------
no_ff0
	cp 1			; 1: salto in BOOTROM (il vecchio 'sandy_in')
	jr nz,no_ff1
	jp 0			; il context mode e` gia` attivo



;------------------------------------------------------------------------------------
no_ff1
;	cp 2			; 2: trasmissione di una stringa al PIC NOT TESTED!!
;	jr nz,no_ff2
;	call pre_serial		; Destroys AF, BC. proper clock speed, no INT
;	ld e,(hl)		; indirizzo del buffer
;	inc hl
;	ld d,(hl)
;	inc hl
;	ld c,(hl)		; nbytes da trasmettere
;	inc hl
;	ld b,(hl)
;	ex hl,de
;l_txpicbytes
;	ld d,(hl)		; data to be sent
;	push bc
;	call soft_232_tx	; destroys AF, BC, D
;	pop bc
;	inc hl
;	dec bc
;	ld a,c
;	or b
;	jr nz,l_txpicbytes
;	call post_serial	; restores previous clock and INT status
;	ld bc,0			; returns zero to BASIC
;	jp $14
	


;------------------------------------------------------------------------------------
no_ff2
;	cp 3			; 3: ricezione di una stringa dal PIC
;	jr nz,no_ff3
;
;	call pre_serial		; Destroys AF, BC. proper clock speed, no INT
;
;	ld e,(hl)		; buffer address
;	inc hl
;	ld d,(hl)
;	inc hl
;	push hl			; saves the address for return values
;	push de
;	ld e,(hl)		; bytes to wait for
;	inc hl
;	ld d,(hl)
;	inc hl
;	ld b,(hl)		; requested timeout (0 = no timeout), 180ms/unit (55 = 10 seconds)
;	pop hl			; receive buffer pointer
;l_rxpicbytes
;	push de
;	call soft_232_rx	; Dstroys AF, DE. If D = 0: E = received byte. D = 1 = framing error; D = 2 = timeout
;	ld (hl),e
;	ld a,d
;	pop de
;	cp 0
;	jr nz,error_occurp
;	inc hl
;	dec de
;	ld a,d
;	or e
;	jr nz,l_rxpicbytes
;error_occurp
;	pop hl			; A = 0 if loop completed
;	ld (hl),e		; remaining bytes
;	inc hl
;	ld (hl),d
;	inc hl
;	ld (hl),a		; error code
;
;	call post_serial	; Destroys AF, BC. restores speed and INT
;
;	ld b,0
;	ld c,(hl)		; error code in BC too
;	jp $14



;------------------------------------------------------------------------------------
no_ff3
	cp 4			; 4: Trasmissione di una stringa sulla RS-232 principale (115K2)
	jr nz,no_ff4
	ld e,(hl)		; indirizzo del buffer
	inc hl
	ld d,(hl)
	inc hl
	ld c,(hl)		; bytes da trasmettere
	inc hl
	ld b,(hl)
	ex hl,de
l_txmainbytes
	ld a,(hl)		; data to be sent
	call rs_tx
	inc hl
	dec bc
	ld a,c
	or b
	jr nz,l_txmainbytes
	ld bc,0			; returns zero to BASIC
	jp $14



;------------------------------------------------------------------------------------
no_ff4
	cp 5			; 5: Ricezione di una stringa dalla RS-232 principale (115K2)
	jr nz,no_ff5
	xor a
	out (NMIREG),a		; disabilita la generazione di NMI su ricezione di bytes

	ld e,(hl)		; buffer address
	inc hl
	ld d,(hl)
	inc hl
	push hl			; saves the address for return values
	push de
	ld e,(hl)		; bytes to wait for
	inc hl
	ld d,(hl)
	inc hl
	ld b,(hl)		; requested timeout (0 = no timeout), 180ms/unit (55 = 10 seconds)
	ld c,0			; fast compare byte
	pop hl			; receive buffer pointer
hardw_rx2
	push de
	push bc
hardw_rx3
	ld de,0x2000		; LSB timeout counter
hardw_rx
	in a,(STATUS)		; 11T
	and 7			; D2 = framing error, D1 = overrun, D0 = byte available
	cp 2
	jr nc,hardw_err
        bit 0,a			; 8T ATTESA PRESENZA DI UN BYTE
        jr z,no_bytepres	; 7T

	pop bc			; restores timeout counter
	in a,(RXREG)
	ld (hl),a
	inc hl
	pop de
	dec de			; bytes counter
	ld a,d
	or e
	jr nz,hardw_rx2

	ld c,0			; error code
mainrx_complete
	pop hl			; return code address
	ld (hl),e		; remaining bytes
	inc hl
	ld (hl),d
	inc hl
	ld (hl),c
	ld b,0
	ld a,1
	out (NMIREG),a		; riabilita gli NMI
	jp $14			; exit ok
	
no_bytepres
	ld a,b			; 4T
	cp c			; 4T no timeout?
	jr z,hardw_rx

	dec e
	jr nz,hardw_rx
	dec d
	jr nz,hardw_rx
	djnz hardw_rx3
	pop bc
	pop de
	ld c,4			; D2 = timeout error
	jr mainrx_complete

hardw_err
	pop bc
	pop de
	and 6
	rrca
	ld c,a			; C = error code: D1 = framing error, D0 = overrun
end_rxhardw
	in a,(RXREG)		; clears the error (and saves the byte)
	ld (hl),a
	jr mainrx_complete




;------------------------------------------------------------------------------------
no_ff5
	cp 6			; 6: Richiede data e ora al PIC (RTC a bordo)
	jr nz,no_ff6
	push hl
	call rtc_enquire	; data is returned at address (HL)
	pop hl
	dec hl
	ld (hl),a		; error code overwrites command code
	ld c,a
	ld b,0
	jp $14




;------------------------------------------------------------------------------------
no_ff6				; 7: Lettura 64 bytes entry di uno snapshot da SD/MMC (FAT).
	cp 7			; La variabile SNAP_TORESTORE resta impostata con l'entry
	jr nz,no_ff7		; richiesta. Su di essa si basa la LETTURA DI UN BLOCCO DATI 0-128K

	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	push de			; salva l'indirizzo buffer passato dal basic
	call common_fatread	; calcola gli indirizzi e legge il blocco da 512 bytes che contiene
	pop de			; i 64 richiesti
	cp 0
	jr z,allok_1
	ld c,a
	ld b,0
	jp $14
allok_1
	ld bc,64		; HL punta all'entry richiesta, DE al buffer in cui scrivere
	ldir			; copia 64 bytes
	jp $14			; rientra con BC = 0




;------------------------------------------------------------------------------------
no_ff7				; 8: Scrittura 64 bytes entry di uno snapshot su SD/MMC (FAT).
	cp 8
	jr nz,no_ff8

	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	push de			; salva l'indirizzo buffer passato dal basic
	call common_fatread	; per prima cosa legge i 512 bytes che contengono i 64 da modificare
	pop de
	cp 0
	jr z,allok_2
	ld c,a
	ld b,0
	jp $14
allok_2
	ld bc,64		; HL punta all'entry richiesta, DE al buffer in cui scrivere
	ex hl,de		; sovrascrive 64 bytes con quelli forniti dal BASIC
	ldir
	ld hl,(snap_torestore)
	ex hl,de
	call calc_hldec		; ricalcola l'indirizzo del blocco
	ld ix,SHADOW_RAM	; scratchpad area contains source data
	ld b,1			; blocks to write
	call mmc_write_data	; performs the single block write
	ld b,0
	ld c,a			; error code
	jp $14


common_fatread			; calcola HL:DE = indirizzo in SD/MMC e legge 512 bytes che contengono
	ld e,(hl)		; i 64 richiesti. Aggiorna SNAP_TORESTORE. Restituisce il blocco in -> SHADOW_RAM
	inc hl
	ld d,(hl)		; numero snapshot richiesto
	ex hl,de
	ld (snap_torestore),hl	; salva il numero dello snapshot richiesto
	ex hl,de

	call calc_hldec		; calcola l'indirizzo del blocco --> HL, DE e l'offset 64 bytes in C

	ld ix,SHADOW_RAM	; scratchpad area
	call mmc_read_data	; performs the single block read
	cp 0
	ret nz			; ERROR (code in A)

	ld b,c			; BC = offset 64 bytes entry da leggere
	ld c,0
	srl b
	rr c
	srl b
	rr c

	ld hl,SHADOW_RAM	; scratchpad area
	add hl,bc		; aggiunge l'offset relativo all'entry da leggere (ce ne sono 8 in 512 bytes)
	xor a			; no error
	ret




;------------------------------------------------------------------------------------
no_ff8				; 09: reads a block of RAW DATA from the 128KBytes area pointed
	cp 9			; by the internal variable SNAP_TORESTORE, which is set by functions
	jr nz,no_ff9		; 07 and 08.

	call common_rawrdwr
	call mmc_read_multidata
	ld c,a
	ld b,0
	jp $14			



;------------------------------------------------------------------------------------
no_ff9				; 10: writes a block of RAW DATA to the 128KBytes area pointed
	cp 10			; by the internal variable SNAP_TORESTORE, which is set by functions
	jr nz,no_ff10		; 07 and 08.

	call common_rawrdwr
	call mmc_write_data
	ld c,a
	ld b,0
	jp $14			


;------------------------------------------------------------------------------------
no_ff10
	jp $14


common_rawrdwr
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	push bc
	pop ix			; buffer address
	ld c,(hl)		; offset lettura in SD/MMC espresso in blocchi da 512bytes
	inc hl
	ld b,(hl)		; numero di blocchi da 512 bytes da leggere

	ld hl,(snap_torestore)
	ex hl,de		; DE = n. entry (da moltiplicare *2 perche` ognuna occupa 128K e non 64)
	sla c			; ne approfitta per scalare dentro il bit MSB offset blocco da 512 bytes
	rl e
	rl d
	ld hl,FATSIZE		; somma n. blocchi da 64KB che costituiscono la FAT (offset per la word MSB)
	add hl,de		; HL = word MSB scrittura in SD/MMC
	ld d,c			; word LSB = 7 bit da numero blocco da leggere, restanti 9 bit = "0" (512 bytes)
	ld e,0
	ret

font_block
        end 


; Original history: in bottom of file.
; Version 4.94c 27/01/2007
;
; New I/O ports for I/O CPLD: $1F-$7F instead of $0F-$3F.
; FLASH routine for ST chips no longer works at 7MHz when receiving data: there are problems at
; this speed (to be investigated). 7MHz are now only used by rom_64kcopy, then 3.5MHz speed is
; selected before receiving data to be flashed.
; 
;
;
; Version 4.94b 24/01/2007
;
; Multi-frame loading capability for "video" playback in hi-res mode. Framerate is 20/sec
; @21MHz.
;
;
;
; Version 4.94 23/01/2007
;
; Experimental 'I' keyboard command for HI_RES screen load from MMC card as RAW DATA
;
;
;
; Version 4.93c 17/01/2007
;
; New "residos_speed" shadow_ram variable, used by ResiDOS to determine the maximum allowable
; speed to be selected into CLKREG register.
;
;
;
; Version 4.93b 06/01/2007
;
; Two EQUs added to easily choose the desired frequency during snapshot RD and WR
; between 14 and 21 MHz.
;
;
; NOTE: After 512K 45ns ram chip installation, 21MHz operation are not reliable.
; For example, trying to read a snapshot at 21MHz when SHADOW_RD is enabled (firmware debug,
; running in RAM) will crash the machine. This happens with nonvolatile chip installed.
; Behaviour without this chip may differ, but the chip is necessary...
; (ISSUE SOLVED by ULA3 v4.8d 16/01/2007 (wait state on opcode fetch in RAM at 21MHz))
;
;
;
; Version 4.93 06/01/2007
;
; Added the rs-232 'I' command, that allows reading any RAM/ROM bank through the FASTPAGE
; register. The dualport ram is not accessed; bank 5 and 7 (shadow rom 8 and 9) are accessed
; instead. Up to 1 MByte (512K RAM/ 512K ROM) can be read.
;
; The WRITE loop is not present yet.
;
;
;
; Version 4.92 01/01/2007
;
; New 5-bytes variable space reserved (shared by) ResiDOS by Garry Lancaster, at beginning of
; shadow_rom rd/wr space (last 768 bytes of the BOOTROM space, at address $3D00).
;
; $3D00 (2 bytes) = ResiDOS signature (should contain "RD")
; $3D02 (2 bytes) = ResiDOS version (BCD)
; $3D04 (1 byte) = RAM Bank to be activated (by fastpage register) before jumping into ResiDOS
;
; When the bootrom firmware detects a version of ResiDOS in the nonvolatile ram, a corresponding
; user-selectable option is displayed. When selected, this option will bank-in the required
; 16KBytes RAM bank over rom space (0 - $3FFF) then jump at address 0, activating the existing
; ResiDOS installation. ResiDOS will then resume from the point it was halted before system
; shutdown.
;
;
;
; Version 4.91 30/10/2006
;
; Tests for 21.25MHz UART CLK: is now possible the use of OTIR/INIR even at 21.25MHz Z80 speed
;
;
;
; Version 4.90 28/10/2006
;
; Some register's LOCK-related bugfixes for snapshots.
; MMC snapshot save/restore now correctly handles both locks: they are removed prior to
; operate, then restored if necessary.
;
; When taking a snapshot while special paging mode (+2A/+3) is active (D0 $1FFD = '1'),
; the special paging is disabled and a full 128K snapshot is forced (data is read by means
; of $7FFD register).
;
;
;
; Version 4.88 22/10/2006
;
; Added support to D6 of $34DF register (all non-standard registers LOCK) to all routines
;
;
;
; Version 4.87 21/10/2006
;
; Added support to dual-card SPI interface in power-on initialization.
; Card is searched on MMC0 slot (D0 chip select). If not found, or found but no SNAPSHOT-Ready,
; then the MMC1 slot (D1 chip select) is searched.
;
; Inserting a +3e-formatted card on slot MMC0 and a snapshot-ready card on MMC1 slot allows
; loading +3e software from one card while keeping snapshot capabilities on the other one.
;
;
;
; Version 4.86 20/10/2006 (my wife's birthday)
;
; Added support for two SD/MMC card chip selects in CS_LOW routine:
; OUT_PORT: D0 = CS0 (as before); D1 = CS1 (new - was TX rs-232 to PIC).
; TX pin for software rs-232 to PIC microcontroller moved to D2 of OUT_PORT.
;
;
;
; Version 4.85b 17/10/2006
;
; "Battery LOW" detection routine added (for DS1210 non-volatile controller on 128K ram chip)
;
;
;
; Version 4.85 16/10/2006
;
; ST M29F010B 1Mbit FlasRom now supported, with 16Kbytes-block erase.
; A new 'FLASH' equ is now present, to include different program sections:
; 0 = no flash
; 1 = WINBOND W29C020 2Mbit
; 2 = ST M29F010B 1Mbit
;
;
;
; Version 4.84b 14/10/2006
;
; Snapshot parameter's edit window is now compatible with new clock settings.
; MMC and RS-232 snapshot now saves/restores the $54DF register too (FASTPAGE register) as long
; as the CONTEXT_SWITCH DISABLE and INT DOUBLER bits (D5 and D1) from/to $34DF register.
;
; $AA is now sent at power-on through the main RS-232 port. This will force ZX-Com to logout,
; avoiding strange behaviour after flashing a ROM: zx-com does not logs-off by default since
; more than one rom can be flashed in one session. If zx-badaloc is reset, however, zx-com
; should be notified.
;
;
;
; Version 4.84 11/10/2006
;
; Nonstandard registers of main cpld $2FFD, $3FFD $44DF, $54DF has been moved to x4DF.
;
; Added support to the new clock setting 14MHz/50Hz via NMI button:
;
; D2:D0 bits of register $24DF now works as follows:
;
; D2 = doubler (as before)
; D2:D1:D0
;  0  0  0		3.5MHz
;  0  0  1		7MHz
;  0  1  0		14MHz
;  0  1  1		21MHz
;  1  0  0		7MHz / 100Hz
;  1  0  1		14MHz /100Hz
;  1  1  0		Invalid (would be 28MHz/100Hz)
;  1  1  1		Invalid (would be 42MHz/100Hz)
;
;
;
;
; Version 4.83 07/10/2006
;
; COMMON_GETROM subroutine, used by both the 'E' and 'F' commands (Receive 16/32K rom into
; shadow_rom then jump in it and Receive ROM data to be flashed) now works on BANKED_RAM AREA
; ($C000) and is able to receive up to 64Kbytes of data, increasing the RAM_BANK on each
; overflow. This will allow easier implementation of Flash routine for chip other than Winbond,
; which is the only one that does not need to be erased on a 64Kbytes basis.
; It also takes now advantage of the new $44DF register, for D5 $7FFD unlocking.
;
; The 'F' command (flash) can now accept buffers up to 64Kbytes. Flash occurs starting from
; the requested ROM BANK then goes ahead until received data is completely written.
;
;
;
; Version 4.82 02/10/2006
;
; A SD/MMC card that is not recognized as 'snapshot card' is no longer initialized by default.
; All snapshot-related keys will not be displayed and an alternate menu` appears, allowing
; more rom selection and a specific command that let initialize the card for snapshot use.
;
; This allows different sd/mmc card filesystems to be used with (new installed roms).
;
;
;
; Version 4.81 01/10/2006
;
; The FLASH ROM routine takes now advantage of the new FAST PAGING $54DF register.
;
;
;
; Version 4.80 28/09/2006
;
; Added support to the new $1FFD register. Since a new ROM address bit appeared in the middle
; of ROM_MSB banks, several changes were made in order to keep things working.
; New commands from RS-232 added, for $1FFD register manipulation.
; SD/MMC snapshot will now save this register too. RS-232 snapshot will now send/restore
; this register too.
;
;
;
; Version 4.70 25/09/2006
;
; Support to the new ULA3 CPLD FLASHROM control capabilities. New 'F' command that allows
; ZX-Com to send a ROM_BANK, that will be programmed into the flash chip.
;
; Since many FLASH EPROM chips are 4Mbit wide, one more address line has been added (A18)
; and is managed by the flashing routine. Support to all other functions (rom selection,
; rom number into snapshot data, etc) will be added soon.
;
;
;
; Version 4.60 17/09/2006
;
; RANDOMIZE USR 37 (or jp $25) functions can now deal with any amount of data instead of
; single characters. The word at 23728, 23729 acts now as a pointer to the buffer containing
; the requested operation's code and all needed parameters.
;
; Added 4 functions that allow BASIC programs to manipulate data into SD/MMC card:
; two to deal with FAT area (the whole SD available space is break into 128KBytes chunks that
; are also used for single snapshots and any chunk has a 64 bytes fat with data type, filename,
; etc.) and two to exchange raw data blocks within the selected 128KB chunk.
; A new datatype (raw data) has been defined, which is not accepted as snapshot by the restore
; routine.
;
; Added a function that lets load the "next snapshot" from SD/MMC card.
; In order to allow BASIC programs to call this function with minimum effort, there is no need
; to create a buffer in memory: when the MSB address byte (23729) contains ZERO, then the LSB
; becomes the direct function's code.
;
; A program that loads the "next block" from tape can then be easily modified to load "next
; snapshot". Of course, the next snapshot saved into SD/MMC should be the next program block :-)
;
;
;
; Version 4.55 16/09/2006
;
; New RS-232 command 'E' lets ZX-Com upload a 16/32K binary file containing a Sinclair ROM,
; to be used (evaluated) in place of those available into EPROM chip.
; All bank and context switch is handled automatically.
; If the received block is 16KB long, then only SHADOW_ROM BANK 0 is used. A copy of BOOTROM
; is then created in BANK1 for continued bootrom functions such as snapshots, etc.
; (If bootrom is already running on SHADOW_ROM, then no copy takes place).
; If the received block is 32KB long, both BANK 0 and 1 are used and no BOOTROM firmware will
; be ever available (until a CPLD reset). In this case, the NMI from I/O CPLD is not re-enabled
; on exit, to prevent the system from crashing should ZX-Com try to communicate.
;
;
;
; Version 4.54 05/09/2006
;
; Time/date format of local variables changed to binary; the PIC microcontroller will accept
; /return binary data only, and will translate it to BCD before dealing with the RTC chip.
;
; Date and time of sd/mmc snapshot is now saved in the snap file information, after filename
; (7 bytes in the same sequence as for RTC chip, but in binary format)
;
; Date and time of snapshot are now shown while browsing, in the main menu screen
;
; Default processor speed raised to 7MHz to improve browsing. Speed is set back to 3.5MHz
; before jumping in any ROM, or to the desired (saved) speed when restoring a snapshot
;
;
;
; Version 4.53 03/09/2006
;
; 3.5MHz software rs-232 tx routine removed (now every communication is performed at 21MHz).
; Two new subroutines (pre_serial and post_serial) provide easy clock/INT setup before/after using
; serial RS-232. The first sets 21MHz and disabled INT, while the second restores both as they was.
;
; Added a function to send/receive date/time informations from/to PC <--> RTC chip, that allows
; the ZX-Com to sinchronize the PC clock with RTC clock in both directions.
;
;
; Version 4.52 02/09/2006
;
; Bugfix:
; The 'loadthis' subroutine, which jumps into the user-selected ROM from boot menu, used to
; disable the SHADOW_RD bit in CLKREG register without relocating itself into system ram
; (when debugging a firmware: only under these circumstances the SHADOW_RD is enabled).
; Although the CONTEXT_SWITCH protects code execution by keeping the BOOTROM active regardless
; of ROM_BANK content, changing the SHADOW_ROM READ ENABLE BIT will cause a rom bank switch
; because RAM/ROM selection for shadowing is not affected by context_switch (as it should:
; this way it's possible to debug the firmware loading it into shadow_rom).
;
; The bug was hard to discover because the 'loadthis' subroutine is located in the middle of the
; code space. Until the phisical eprom contained the same code at those locations, disabling the
; shadow_rd bit without relocating into system ram did not show any problem since program execution
; could have continued in EPROM. The code then used to jump into the selected rom, clearing
; everything.
;
; However, after a few changes at the beginning of the bootrom code block (that moved the 'loadthis'
; subroutine position), jumping into a basic ROM failed. The new 'loadthis' version checks if
; code is running into shadow_rom. If so, it also checks if the selected rom is a physical rom
; (not the shadow_rom bank0, where the loader copied the 48K basic rom). If both are true, then
; the code is relocated into ram before writing to the SHADOW_RD bit in the CLKREG register.
;
; New features:
; Since the PIC microcontroller has a lot of work to do at power-on (PS/2 devices initialization
; and RTC data read), the Z80 firmware waits for an 'alive' response at a special enquire byte,
; sent through the software RS-232 line. The PIC will respond only after the main loop is reached,
; which means that the device is ready.
;
; Time/date informations are requested from PIC at power on and stored in local variables (the
; previously described syncing procedure ensures that the PIC is ready to handle the request).
; The PIC, in turn, just asked these informations from the RTC chip and stored into local variables.
; Both PIC and Z80 then update these variables thanks to the 1Hz signal delivered by the RTC chip.
; At midnight, the PIC will read the RTC again in order to update day/date/month/year bytes.
; The same is performed by the Z80, who will receive data from PIC.
;
; The bootrom main menu shows date and time. Please note that date/time local variables are stored
; into SHADOW_RAM_BANK. This bank is writable only when CONTEXT_SWITCH is active. In case a standard
; rom will be modified to keep these variables updated, the code should better call a BOOTROM function
; in order to increment variables. Only the increment phase needs context_switch active; the 1Hz
; signal can be checked at any time by reading the STATUS ($2F) port.
;
; When a NMI is generated while the BOOTROM firmware is running (for example entering the ZX-Com program
; on the RS-232 port or pressing the NMI button while the poweron menu is active), context_switch is no
; longer deasserted on exit.
; The test is made on the active ROM_BANK. If the number is 3 (bootrom firmware) then the handler is
; quit by a direct RETN instruction, instead of jumping into address $14 (which would be trapped by
; CPLD, disabling context switch mode). This allows RTC local variables to be updated even after a
; NMI operation.
;
;
;
; Version 4.51 01/09/2006
;
; Added a function that asks the PIC to send the RTC content (current date and time), that is
; then received through the Z80 software RS-232 link to the PIC itself.
;
;
;
; Version 4.50 26/08/2006
;
; The new ULA3 V4.0c now switches the CONTEXT MODE on opcode fetch at $25 (37) too.
; There are 3 spare locations on both 48/128K roms; due to the context mode, our bootrom takes
; control and jumps to an handler that performs special tasks. The code of the desired function
; should be placed at RAM location 23728 before calling USR 37.
;
; Two RS-232 TX routines are provided: one running at 3.5Mhz, the other at 21.
; A frontend selects between them. If none is suitable, the frequency is changed to 21MHz
; and the second routine is called; then, the previous clock speed is restored.
;
; A new RS-232 RX routine is now provided, running at 21MHz. This speed is necessary in order
; to handle a timeout while waiting for the start bit. The caller should ensure (or switch)
; the 21MHz frequency.
;
; Both RX and TX can be invoked from BASIC, by USR 37 function.
;
;
;
; Version 4.40 25/08/2006
;
; Test: RS-232 transmission to new PIC16F877 with hardware USART
;
; Bugfix from previous release:
; The joystick_reset subroutine was called @3.5MHz Z80 speed, which prevented the table from
; being received by PIC (the rs-232 TX routine is software based on Z80 and should be run
; at 21MHz for proper 38K4baud timings)
; Now the SET_CPLDREGS sets 21MHz, then the main routine brings the clock back to 3.5MHz after
; the joystick_reset has finished it's job.
; 
; Bugfix from previous release:
; The software rs-232 TX routine (soft_232_tx) used to lower the SD/MMC chip select when placing
; the start bit on the RS-232 line
;
; Changes from previous release:
; The software-rs232 now deals with an hardware receiver, so the pre-activation to get PIC
; attention has been removed. Also, the first character to be sent is now the header code ($55).
; The rs-232 line should now be NORMALLY HIGH, so the CS_LOW and CS_HIGH (sd/mmc) will keep this
; line at '1' instead of '0'.
;
;
;
; Version 4.31b 28/07/2006
;
; Bugfix: Solved a bug in MMC_INIT that could have failed to initialize slower cards.
; Now the routine waits for the idle/init_in_progress bit to go inactive before proceeding.
;
;
;
; Version 4.30 27/07/2006
;
; Starting from ULA3 Rev4.0, RET and RETN are no longer needed in original roms since the
; context switch is terminated AFTER the opcode fetch at addresses $14 or $16.
; Being a two-byte instruction, RETN M1 cycle should be detected on second byte ($15).
; Unfortunately the SANDY ROM does not have any opcode fetch at $15 and this prevents the
; context switch from being disabled when the sandy rom start it's execution (thanks to the
; ROMCS = '1' that disables all badaloc's memories). This means that on return from SANDY rom,
; the system would switch to the BOOTROM instead of the (necessary) spectrum 48K rom.
;
; For this reason, instructions RETN and RET have been swapped so that trigger addresses remain
; the same:
;
; $14	C9	RET		<-- trigger 1 (needed by the SANDY ROM)
; $15	ED	RETN byte 1
; $16	45	RETN byte 2	<-- trigger 2 on last opcode fetch
;
; This frees the original roms from containing these opcodes. From now on, anything could be
; present at that locations (their execution would just reset the context, that in that case
; is reset already).
; 
;
;
; Version 4.20 22/07/2006
;
; Added support for bit D1 on $34DF register. When SET, this bit doubles the INT frequency
; by inserting a second INT signal in the middle of the screen generation. Hence, the
; INT speed will be 100/200Hz instead of 50/100.
;
; This bit is set/reset on snapshot restore from both SD/MMC and RS-232. The take-snapshot
; routines will default it to ZERO. It can be SET only by user (sd/mmc edit snapshot parameters
; menu` of BOOTROM FIRMWARE or zx-com VC++ program for RS-232 snapshots).
;
; Command 'S' (ask CPLD registers) now sends the $34DF content as well.
;
; Command 'C' (change cpu clock) no longer exits the handler: thanks to the new fixed-frequency
; UART, there is no baudrate shift.
;
; New command "D" for $34DF register setting
;
; When trying to take a sd/mmc 128K snapshot, the operation is no longer aborted if the $7FFD
; register's lock bit is found SET. A 48K snapshot will be performed instead, with border turned
; RED (as for 16K snapshot turned into 48K because of SP above the first 16K of memory) to
; indicate that something strange happened. If the program was running fine, then a 48K snap
; is probably all we need.
;
; 'M' command bugfix
;
;
;
; Version 4.11 07/05/2006
;
; IM_2 test vector expanded to 1-page size. This allows the IM_2 vector fetch to work even if
; databus is not floating to $FF.
;
; As an example, the SD/MMC take-snapshot did not work above 14MHz of Z80 clock if snapping an
; IM_2 mode program (Galaxian) because pullup resistors were not fast enough to provide the
; $00FF interrupt vector (0x00.. loaded in I register, ..FF by floating databus).
;
; Now the entire page 1 (0x100 - 0x201 = 258 bytes) is filled with '2', then the service
; routine is located at $202. 257 bytes would have been enough for any value fetched from
; databus; byte n. 258 is not used (the routine should start at $202).
;
; TEST: the SD/MMC snapshot now works with Galaxian even at 21MHz Z80 Clock. The clock speed,
; however, has been kept at it's previous value (14MHz) for safety reasons.
;
;
;
; Version 4.10 01/05/2006
;
; Magic Fields (Zack June 1996) used to crash after snapshot-take (and, of course, after snapshot
; restore). This was due to the dumb interrupt re-enabling procedure of bootrom up to V4.03b:
; if interrupt = enabled AND IM MODE = 2 AND the program's interrupt handler rely on BASIC ROM,
; then the program would crash because at least ONE interrupt were shooted BEFORE we restored
; the basic rom (this happens only on exit from our handler).
;
; New way:
;
; IM MODE is set to IM1, so the processor will execute our "fake" interrupt handler (EI, RETI);
; If interrupt have to be enabled, then EI is executed and followed by a couple of HALT instructions:
; this will ensure that, after execution of the last HALT, we have about 20ms (10 if interrupt = 100Hz)
; before a new interrupt will occur.
; The real IM MODE is then restored. Thanks to syncing to the INT signal, we exit the routine (replacing
; the rom where it's supposed to be) before the host's interrupt handler is executed.
;
;
;
; Versione 4.03b del 20/03/2006
;
; Due nuovi tasti [N], [M] consentono di muoversi di 128 passi alla volta nel menu` di
; selezione dello snapshot SD/MMC da caricare.
;
; Introdotto controllo sul limite di spazio disponibile in SD/MMC, che tiene conto anche
; della FAT ed impedisce di far avanzare i puntatori di snapshot oltre al dovuto.
; Al momento utilizza solo dele EQU relative a quella da 1GB.
;
; NOTA: e` possibile che 1GB sia inteso come "1 miliardo di bytes", infatti lo snapshot non
; funziona a partire dall'indice 7856 (il limite dovrebbe essere 8188, ovvero 8192 - 4 dovuto
; ai 512K di fat). Il software non tiene conto di questo fatto ed arriva fino a 8187.
; In futuro verificare se nel CID c'e' una informazione precisa in merito.
;
; Timeout di MMC_WAIT_RESPONSE allungato, anche se in realta` funzionava anche prima.
;
;
;
; Versione 4.02 del 18/03/2006
;
; La ULA2 V3.2 mappa automaticamente 1KB di SHADOW_ROM RD/WR sull'ultimo KB di rom space quando
; il context switch e` attivo.
;
; Questo consente di semplificare l'uso di spazio ram per le variabili. Non e` piu` necessario
; copiare l'intera bootrom in shadow_rom per poter eseguire uno switch manualmente.
; E` stata tuttavia conservata una versione modificata di enable_shadow e disable_shadow, ora
; chiamate "ENABLE_WRSHADOW" e "DISABLE_WRSHADOW": la prima abilita lo shadow_rom WR se la
; modalita` RD viene trovata attiva; la seconda disabilita il WR incondizionatamente.
; Servono in sessioni di debug, nelle quali il firmware e` stato uploadato sulla SHADOW_ROM
; e quindi il context switch non e` attivo: chiamate prima di accedere all'area variabili messa
; a disposizione sull'ultimo KB, consentono di portare a termine le scritture con successo.
;
;
;
; Versione 4.01 del 18/03/2006
;
; I bytes non utilizzati in ciascuna entry di snapshot su SD/MMC vengono inizializzati a zero.
; Questo comprende anche i 10 bytes di impostazioni del joystick programmabile, in tutto
; vengono azzerati gli ultimi 19 bytes del blocco di 64 che compone una entry.
;
; La velocita` dell'operazione di scrittura snapshot su SD/MMC e` stata ridotta da 21 a 14MHz,
; perche` diversamente la routine di rilevamento dell'IM MODE va in crash se la CPU e` impostata
; in IM2. E` possibile che i pullup sul databus non abbiano il tempo di portarlo a $FF.
;
; Il clock Z80 viene sempre settato a 3.5MHz all'inizio dell'esecuzione della BOOTROM.
;
; In caso di ricezioni da seriale con errori di framing o overrun, la routine NMI esce subito
; senza trasmettere codici di alcun tipo. Trasmette invece '!' in caso di ricezione corretta
; di un byte diverso da quello atteso (come prima).
;
;
;
; Versione 4.00 del 18/03/2006
;
; Inserita gestione joystick programmabile per snapshot su SD/MMC. Vengono utilizzati 10 bytes
; liberi nell'entry per memorizzare il codice del tasto corrispondente a ciascuno dei 10
; ingressi di cui l'interfaccia dispone.
;
; Il primo byte rappresenta il codice a 8 bit da inviare alla CPLD quando viene attivato
; l'ingresso n.1, e cosi` via. Questi 10 bytes vengono trasmessi al PIC16F873 mediante una
; seriale software funzionante a 38K4 bit/sec.
; Il PIC e` collegato al joystick ed all'occorrenza scrive i dati nel registro della CPLD di
; gestione della tastiera, la quale simulera` la pressione del tasto.
;
; Un byte e` cosi` composto: 5 bit MSB = stato DATA LINE (D4 - D0 da piazzare sul databus in
; caso di lettura su $FE per simulare il tasto desiderato); 3 bit LSB = 0-7 numero di linea
; di indirizzo (A8 = A15) che deve essere attivata per indurre la CPLD a presentare quel
; dato sul databus. La CPLD contiene quindi 8 registri da 5 bit ciascuno, indirizzabili mediante
; i 3 bit LSB del byte che contiene il codice del tasto.
;
; La gestione del joystick programmabile verra` inclusa anche nelgli snapshot via UART.
;
;
; Versione 3.31 del 15/03/2006
;
; Inserita routine di trasmissione rs-232 (in software) per la trasmissione dei dati di
; programmazione del joystick al PIC16F873.
; Viene utilizzata la linea D1 sulla porta OUT_PORT
;
;
;
; Versione 3.30b del 11/03/2006
;
; Inserito temporaneamente OUT ($FC),$87 all'inizio dell'handler NMI. Serve ad abilitare
; con certezza l'hardware UART-SD/MMC nel caso il firmware SANDY l'avesse disabilitato.
; In questo modo e` possibile eseguire lo snapshot anche delle utilities sandy, come il
; FORMAT, che durante l'esecuzione tengono il proprio hardware attivo (disattivando l'altro).
;
;
;
; Versione 3.30 del 06/03/2006
;
; Nuova subroutine mmc_read_multidata, che esegue la lettura di piu` blocchi in modalita`
; READ_MULTIPLE_BLOCK.
; La subroutine mmc_read_data e` rimasta e legge ora solo un singolo blocco.
;
; Tempi di caricamento:
;
;  16K	 23,20ms
;  48K	 67,65ms
; 128K	180,00ms
;
;
;
; Versione 3.20 del 05/03/2006
;
; OTIR/INIR sostituite da OUTI/INI per rallentare leggermente il data-rate: questo consente
; di aumentare la frequenza del processore a 21.25MHz, portando il throughput a 758,9 KB/sec.
; I test di questa versione hanno portato ad un aggiornamento della CPLD I/O, che ha
; evidenziato problemi con la procedura di start ricezione SPI triggerata da ciclo di IORQ+RD
; se la frequenza del processore e` 21MHz.
;
; Tempi di caricamento:
;
;  16K	 28,65ms
;  48K	 84,30ms
; 128K	223,00ms
; Singolo blocco da 512 bytes: 850us
;
;
;
; Versione 3.10c del 03/03/2006
;
; Porta SPI hardware implementata nella cpld I/O, routines di comunicazione semplificate.
;
; Tempi di caricamento:
;
;  16K	 32,25ms
;  48K	 94,75ms
; 128K	251,60ms
; Singolo blocco da 512 bytes: 955us
;
; Aggiunta attesa rilascio tasto quando viene premuto [E] (edit filename) o [R] (edit snapshot
; parameters)
;
;
;
; Versione 3.02 del 01/03/2006
;
; In caso di SD/MMC card vergine, il numero della prossima entry per snapshot viene
; inizializzato a zero. Tale numero e` conservato nei primi 2 bytes della prima entry
; (ciascuna di esse e` costituita da 64 bytes in cui i primi due hanno senso solo per la
; prima entry). Il codice che stabilisce se la card e` gia` inizializzata viene salvato
; negli stessi due bytes, ma nel secondo blocco (entry n. 9) ed e` pari a $AA55.
;
;
;
; Versione 3.01 del 01/03/2006
;
; L'IM_MODE in cui il processore si trova al momento di uno snapshot viene ora determinato
; mediante un interrupt di test con vettore $00FF. Se e` attivo l'IM_2, la routine verra`
; eseguita e carichera` 2 in una variabile; diversamente, la variabile restera` al suo default
; (che e` 1). Non e` possibile identificare IM_0 in quanto con pullup sul databus esegue un
; RST $38 come l'IM_1.
;
; Modificato anche lo snapshot rs-232 per sfruttare questa capacita`. Dato che il detect
; dell'IM_MODE resetta IFF2 (backup dello stato di abilitazione dell'INT) puo` darsi che
; il programma "snappato" cessi di funzionare, in quanto gli INT non vengono riabilitati
; all'uscita dall'handler NMI. Dato che tale uscita non si verifica fino a quando l'utente
; non preme il tasto LogOUT su zx-com, non sarebbe saggio abilitare gli INT subito dopo il
; completamento dello snapshot (non c'e' modo di settare solo IFF2 senza abilitare l'INT).
; Per questo motivo, la versione 3.1 di zx-com "ricorda" se l'ultimo snapshot aveva gli INT
; abilitati e, quando l'utente termina la sessione, invia un codice modificato che provoca
; anche l'esecuzione di una EI.
; Lo snapshot su SD/MMC, invece, provvede a ripristinare l'EI immediatamente dopo l'esecuzione
; perche` esce subito al programma HOST.
;
; La routine rs-232 di lettura/scrittura di un blocco della SD/MMC utilizza ora la SHADOW_ROM
; come area schatchpad e non altera quindi alcun byte della memoria spectrum.
;
;
;
; Versione 3.00 del 28/02/2006
;
; Utilizzo degli ultimi 768 bytes (3 pagine) dello spazio di indirizzamento della BOOTROM
; come area RAM, grazie all'abilitazione della SHADOW_ROM.
;
; Al momento solo le routines SnapShot SD/MMC attivano questa funzionalita`, mentre il resto
; del programma continua a funzionare come prima. Potendo contare su di un'area STACK dedicata,
; non e` piu` necessario in nessun caso rilocare lo SP in ram video (nemmeno durante uno
; snapshot da 128K con il SP che punta all'area BANKED).
;
; Principio di funzionamento:
; Al power-on tutta la ROM viene copiata nel corrispondente banco di SHADOW_ROM (che e` il n.1)
; per prepararne l'uso in caso di bisogno. NOTARE che se la SHADOW viene trovata gia` attiva
; in lettura, significa che siamo in debug del firmware e quindi NON ESEGUE LA COPIA DALLA EPROM.
; Quando una routine ha bisogno di ram esclusiva per le proprie operazioni (es. snapshot su SD)
; deve chiamare SHADOW_ENABLE. Essa attiva la shadow_rom in RD/WR. Da quel momento, l'intero
; banco da 16K $0000 - $3FFF diventa accessibile anche in scrittura. Sono state definite delle
; equ che riservano gli ultimi 512 bytes come scratchpad per la lettura/scrittura di settori
; della SD/MMC piu` una ulteriore pagina (256 bytes) usata per alcune variabili nonche` lo STACK,
; che inizia sul lato opposto della pagina stessa.
; Al termine delle operazioni, e` bene disattivare la SHADOW_ROM chiamando SHADOW_DISABLE.
; Questa subroutine disattiva il WR ENABLE incondizionatamente, mentre ripristina lo stato del RD
; ENABLE in base allo stato che aveva allo start-up (informazione che e` stata salvata fra le
; variabili in shadow_rom stessa).
;
; Lo SnapShot di SD/MMC non deve piu` sporcare il video ad uso STACK e contemporaneamente ha
; la maggior compatibilita` possibile, perche` non usa piu` lo stack del programma host se non
; per il doveroso salvataggio dei registri da ripristinare in fase di restore.
;
;
;
; Versione 2.30 del 27/02/2006
;
; I tasti Q,W,E attuano uno snapshot "normale", mentre A,S,D forzano la rilocazione di SP in
; ram video (vedi descrizione della versione 2.21c).
;
;
;
; Versione 2.21c del 26/02/2006
;
; Lo snapshot su SD/MMC imposta il bit SPECTRUM_MODE non piu` basandosi sul numero della rom
; attiva, che potrebbe essere la 1 sia per 128K (che si e` spostato su quella rom) che per un
; vero 48K mode, ma bensi` (semplicemente) sul bottone premuto dall'utente, che determina le
; dimensioni dello snapshot stesso.
;
; Lo snapshot su SD/MMC sposta SEMPRE lo SP in ram video, invece di condizionare questa azione
; agli snap da 128K con SP in area banked. Questo consente di limitare l'uso dello stack del
; programma HOST al solo salvataggio dei registri (che gia` richiede un certo spazio) in modo
; da ridurre il rischio di uno stack overflow, che si verificava con lunarjetman (48K).
;
; Snapshot da 128K vengono abortiti se il registro $7FFD ha il bit 5 SET (register lock).
;
;
;
; Versione 2.21b del 25/02/2006
;
; Corretto bug: mancata inizializzazione della variabile "FLAGS"
;
;
;
; Versione 2.21 del 24/02/2006
;
; Funzionamento dello snapshot 128K anche con SP che punta all'area BANKED:
; In quel caso, SP viene caricato a $5800 (primo byte attributi) ed il primo dato salvato
; su questo nuovo stack e` il valore del vero SP (verra` scritto alle locazioni $57FF e $57FE).
; In fase di ripristino, se lo SP conterra` $57FE, il suo valore verra` recuperato da li`
; prima di procedere al recupero dei registri, che vengono sempre salvati sul vero stack.
; Questa cosa serve perche` le routines di takesnap/restoresnap DEVONO usare lo stack e questi
; non puo` sparire. Inoltre avere lo SP in area banked significherebbe sporcare qualche byte
; in TUTTI I BANCHI man mano che vengono indirizzati.
;
;
;
; Versione 2.20 del 24/02/2006
;
; E` ora possibile editare il nome dello snapshot nonche` i parametri principali, ovvero
; Z80 Clock, IM_MODE, INT Enable, Spectrum Type.
;
;
;
; Versione 2.10 del 23/02/2006
;
; Menu di scelta rom, snapshot e cambiamento prossima entry usata per la memorizzazione funzionanti
; sia con tastiera che joystick.
;
;
;
; Versione 2.01 del 21/02/2006
;
; SNAPSHOT 16K/48K/128K funzionante con sistema di allocazione della SD/MMC progressivo.
; Ogni snapshot occupa sempre 128K ed ha un blocco di riferimento di 64 bytes ad inizio card.
; In pratica e` una specie di FAT che contiene 8192 entries, corrispondenti ad altrettanti
; blocchi da 128K (qualcuno in meno perche` la fat stessa occupa i primi 512KBytes).
;
;
;
; Versione 2.00 del 19/02/2006
;
; SNAPSHOT sperimentale FUNZIONANTE 16/48K in area fissa della SD/MMC, ovvero a partire dal
; blocco ZERO.
;
;
;
; Versione 1.04c del 18/02/2006
;
; Al power-on, i tre bit LSB dell'output-port generico $3F vengono posti a '1', si tratta dei
; tre bit di gestione della MMC/SD tra cui vi e` il _CS
; Il bit di INPUT e` collegato a D7 della porta KEMPSTON ($1F).
;
;
;
; Versione 1.04 del 15/02/2006
;
; A seconda della rom scelta da menu` (48/128) avviene una scrittura al registro $34DF:
; D0 LOW = 48K mode; D1 HIGH = 128K mode.
;
; Stessa cosa al ripristino di uno SnapShot, in base al contenuto del ROM_BANK appena
; ripristinato. In questo modo, i programmi troveranno gia` il registro settato correttamente.
; La struttura delle prime 4 rom non deve quindi piu` essere modificata.
; Sara` inoltre necessario aggiornare la subroutine SPECTRUM_SELECT nel caso altre rom di tipo
; 128K dovessero essere aggiunte in seguito: al momento solo la ROM 0 provoca il 128K-mode.
;
; Al momento la sola differenza consiste nel numero di pixel per riga (448/456 rispettivamente).
;
; PROVA: non utilizziamo il JP $14 (RET) neanche dopo un ripristino di snapshot: potrebbe
; essere inutile in quando EI o DI setta entrambi i bit, quindi RETN non sovrascrive il settaggio.
; Successivamente si potra` pensare di eliminare il terzo byte da tutte le rom.
;
;
;
; Versione 1.03c del 05/02/2006
;
; Il gioco GALAXIAN (ATARISOFT) non funzionava a causa dell'uso della modalita` interrupt IM2.
; Ora il byte F (flag in cui D2 indica lo stato di IFF (interrupt enable FLIP-FLOP)) ha
; tutti i bit mascherati a ZERO tranne D2; D1-D0 indicano IM MODE: 00 = IM0, 01 = IM1, 10 = IM2.
; Dato che la routine di SNAP non e` in grado di stabilire lo stato IM MODE, carichiamo IM1
; per default. Il software VC++ e` in grado di modificare questi bit per fare dei tentativi.
; Nella precedente versione, tutto il contenuto del registro dei flag era salvato pari-pari.
;
;
;
; Versione 1.03b del 04/02/2006
;
; Il pulsante NMI congela il programma in esecuzione ed attende la pressione dei tasti da 0 a 4,
; che consentono di selezionare la frequenza del processore a 3.5, 7, 7+, 14+, 21MHz rispettivamente.
; La '+' indica che anche la frequenza del video viene raddoppiata (da 50 a 100Hz)
;
;
;
; Versione 1.03a del 04/02/2006
;
; Corretti due aspetti riguardanti la funzione SnapShot:
;
; 1)
; Rimosso handler INTERRUPT MASCHERABILE, che non puo` esistere. Il motivo e` che al termine di
; un Upload SnapShot, in base allo stato salvato del flag di abilitazione degli interrupt, potrebbe
; essere eseguita una istruzione EI. Dato che un INT e` sicuramente pendente, si avra` almeno una
; chiamata mentre e` ancora attiva la nostra ROM. In quel momento, pero`, nemmeno un bit di memoria
; o dei registri puo` essere variato senza rischiare di compromettere il sistema.
; In particolare, il precedente handler (che incrementava una variabile FRAMES e leggeva la tastiera)
; mandava sistematicamente in crash il ripristino di SnapShot 128K.
;
; 2)
; Durante l'upload di uno SnapShot, vengono chiamate subroutines per almeno 2 livelli di stack, il che`
; comporta la modifica di almeno 4 bytes di memoria.
; E` quindi opportuno settare lo SP PRIMA di iniziare l'upload, in modo che esso scriva dati OLTRE
; il punto in cui si trovava al momento dello SNAP, ovvero in zona regolarmente utilizzabile dato che
; si presume che il software che stava girando abbia spazio di stack opportuno. E` altresi` bene
; ripristinare subito anche il RAM_BANK dello snapshot, qualora lo stack sia stato messo in ram banked.
;
; E` possibile cambiare lo SP e continuare l'esecuzione in quanto le routines NON USANO LO STACK
; per le chiamate delle varie funzioni da seriale, ma semplicemente dei JP.
;
; Al momento del primo NMI con conseguente entrata nel loop di attesa comandi, sullo stack vengono
; salvati diversi registri, dopodiche` ci si ferma ad uno stadio "PIATTO", dal quale non ci si muove
; fino all'uscita dell'handler se non per l'utilizzo di questi 2 ULTERIORI livelli di stack, che pero`
; vengono scritti/letti (e quindi funzionano indipendentemente dal punto in RAM in cui cio` avviene).
;
; Cambiare l'SP non inficia il funzionamento dell'handler, perche` fino a quando non dovra` uscire
; non verra` recuperato NULLA dallo stack.
;
; Dato che l'esecuzione delle routines di ricezione dei blocchi e dei registri sporcano questi 2
; livelli di stack, e` opportuno che esso contenga la posizione CHE AVEVA AL MOMENTO DELLO SNAP
; FIN DALL'INIZIO. Diversamente sporcheremmo almeno 4 bytes in un punto pressoche` casuale, che dipende
; dal contenuto dello SP (a sua volta dipendente da cio` che e` stato fatto in precedenza).
;
; Dopo il cambio di SP, l'handler continuera` il proprio lavoro e, dopo il trasferimento delle RAM
; (upload SnapShot) sara` magicamente comparsa la catena di stack salvata al momento dello snapshot,
; proprio dove SP sta` puntanto. Questo consentira` la felice uscita dall'handler e lancio del
; programma snappato.
;
; E` pero` ovvio che se viene cambiato lo stack ma non viene ripristinato il contenuto della RAM,
; il sistema crashera` al momento dell'uscita dall'handler stesso.
;
; Dopo il trasferimento della RAM si potra` chiamare l'ultima funzione ('P') che invia lo stato
; degli interrupt mascherabili ed il registro $24DF della CPLD (ROM_BANK e CLOCK SPEED).
;
;
;
; Questa ROM deve essere programmata nel quarto banco della EPROM ($C000) e contiene la gestione della
; scelta del banco da attivare e la routine HANDLER NMI, che non deve piu` essere programmata in tutte
; le rom.
; Per ottenere un FAST_CONTEXT_SWITCH la CPLD versione 2.5 rileva un ciclo M1+MREQ all'indirizzo $66
; (NMI vector) e manda automaticamente a "0011" i 4 bit di selezione del ROM_BANK, a prescindere dal
; contenuto dei registri che li controllano. In questo modo, gia` la prima istruzione all'indirizzo
; $66 verra` letta dalla EPROM $C000 invece di quella corrente.
;
; NOTA: se e` attivo lo SHADOW_ROM, lo switch attivera` il banco 1. E` quindi necessario che durante le
; prove, la rom principale di trovi nel BANK 0 della SHADOW. Il resto dell'indirizzo ROM_BANK verra`
; ignorato (solo commutazione fra i banchi 0 e 1 della SHADOW).
;
; Uscita dalla ROM: la CPLD rileva un ciclo M1+MREQ all'indirizzo $14 OPPURE $16 e provvede a
; ripristinare l'output dei registri sui pin ROM_BANK.
; Nessuna modifica e` necessaria alle ROM originali, perche` in questo caso l'azione e` DOPO LA FINE
; del ciclo M1 che triggera il meccanismo.
;
; Nella nostra BOOTROM, a quegli indirizzi, sono presenti le seguenti istruzioni:
;
; $14	C9	opcode RET (1 byte)
; $15	ED
; $16	45	opcode RETN (2 bytes)
;
; NOTA IMPORTANTE: Dato che il RETN e` costituito da 2 bytes, il detect del ciclo M1 viene fatto sul
; secondo ($16) altrimenti la commutazione avverrebbe dopo che soltanto il primo byte (ED) sara`
; stato letto.
;
; Al termine delle sue operazioni, l'handler NMI eseguira` un salto a $15 invece di RETN, rientrando
; di fatto nella ROM che era stata interrotta.
; Scopo del RET alla locazione $14: in caso di upload SnapShot, rientriamo con un RET invece di RETN
; per non sovrascrivere lo stato del flip-flop di abilitazione degli INT, che sara` appena stato
; impostato in base ai dati di snapshot mediante una EI o DI. E` comunque utile in tutti i casi in cui
; si deve uscire dallo stato di CONTEXT attivo senza manipolare il flip flop di interrupt.
;
;
; Questa ROM deve essere lanciata di default al POWER-ON e consente di selezionare la ROM da attivare
; (oltre ovviamente a contenere l'handler NMI descritto prima).
;
; L'history relativa alle versioni precedenti puo` essere trovata nell'archivio "zx-com 2.1 stabile".
;
;


